@@ -9,16 +9,9 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: cl_read.c,v 10.29 2001/08/18 21:51:59 skimo Exp $ (Berkeley) $Date: 2001/08/18 21:51:59 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
-#ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
-#endif
-#include <sys/time.h>
 
 #include <bitstring.h>
 #include <errno.h>
@@ -38,15 +31,15 @@ static const char sccsid[] = "$Id: cl_read.c,v 10.29 2001/08/18 21:51:59 skimo E
 #undef columns
 #undef lines  
 
-static input_t	cl_read __P((SCR *,
-    u_int32_t, char *, size_t, int *, struct timeval *));
-static int	cl_resize __P((SCR *, size_t, size_t));
+static input_t	cl_read(SCR *,
+    u_int32_t, char *, size_t, int *, struct timeval *);
+static int	cl_resize(SCR *, size_t, size_t);
 
 /*
  * cl_event --
  *	Return a single event.
  *
- * PUBLIC: int cl_event __P((SCR *, EVENT *, u_int32_t, int));
+ * PUBLIC: int cl_event(SCR *, EVENT *, u_int32_t, int);
  */
 int
 cl_event(SCR *sp, EVENT *evp, u_int32_t flags, int ms)
@@ -54,7 +47,7 @@ cl_event(SCR *sp, EVENT *evp, u_int32_t flags, int ms)
 	struct timeval t, *tp;
 	CL_PRIVATE *clp;
 	size_t lines, columns;
-	int changed, nr;
+	int changed, nr = 0;
 	CHAR_T *wp;
 	size_t wlen;
 	int rc;
@@ -124,12 +117,12 @@ read:
 		else
 		    msgq(sp, M_ERR, "323|Invalid input. Truncated.");
 		break;
-	case INP_ERR:
-		evp->e_event = E_ERR;
-		break;
 	case INP_EOF:
 		evp->e_event = E_EOF;
 		break;
+	case INP_ERR:
+		evp->e_event = E_ERR;
+		break;
 	case INP_INTR:
 		goto retest;
 	case INP_TIMEOUT:
@@ -146,13 +139,12 @@ read:
  *	Read characters from the input.
  */
 static input_t
-cl_read(SCR *sp, u_int32_t flags, char *bp, size_t blen, int *nrp, struct timeval *tp)
+cl_read(SCR *sp, u_int32_t flags, char *bp, size_t blen, int *nrp,
+    struct timeval *tp)
 {
 	struct termios term1, term2;
-	struct timeval poll;
 	CL_PRIVATE *clp;
 	GS *gp;
-	SCR *tsp;
 	fd_set rdfd;
 	input_t rval;
 	int maxfd, nr, term_reset;
@@ -184,13 +176,10 @@ cl_read(SCR *sp, u_int32_t flags, char *bp, size_t blen, int *nrp, struct timeva
 	 * 2: A read with an associated timeout, e.g., trying to complete
 	 *    a map sequence.  If input exists, we fall into #3.
 	 */
-	FD_ZERO(&rdfd);
-	poll.tv_sec = 0;
-	poll.tv_usec = 0;
 	if (tp != NULL) {
+		FD_ZERO(&rdfd);
 		FD_SET(STDIN_FILENO, &rdfd);
-		switch (select(STDIN_FILENO + 1,
-		    &rdfd, NULL, NULL, tp == NULL ? &poll : tp)) {
+		switch (select(STDIN_FILENO + 1, &rdfd, NULL, NULL, tp)) {
 		case 0:
 			return (INP_TIMEOUT);
 		case -1:
@@ -231,11 +220,27 @@ cl_read(SCR *sp, u_int32_t flags, char *bp, size_t blen, int *nrp, struct timeva
 	 * the only way to keep from locking out scripting windows.
 	 */
 	if (F_ISSET(gp, G_SCRWIN)) {
-		FD_ZERO(&rdfd);
+loop:		FD_ZERO(&rdfd);
 		FD_SET(STDIN_FILENO, &rdfd);
 		maxfd = STDIN_FILENO;
-		if (sscr_check_input(sp, &rdfd, maxfd))
+		if (F_ISSET(sp, SC_SCRIPT)) {
+			FD_SET(sp->script->sh_master, &rdfd);
+			if (sp->script->sh_master > maxfd)
+				maxfd = sp->script->sh_master;
+		}
+		switch (select(maxfd + 1, &rdfd, NULL, NULL, NULL)) {
+		case 0:
+			abort();
+		case -1:
 			goto err;
+		default:
+			break;
+		}
+		if (!FD_ISSET(STDIN_FILENO, &rdfd)) {
+			if (sscr_input(sp))
+				return (INP_ERR);
+			goto loop;
+		}
 	}
 
 	/*
@@ -302,10 +307,20 @@ err:		if (errno == EINTR)
 static int
 cl_resize(SCR *sp, size_t lines, size_t columns)
 {
-	int rval;
+	ARGS *argv[2], a, b;
+	CHAR_T b1[1024];
 
-	rval = api_opts_set(sp, L("lines"), NULL, lines, 0);
-	if (api_opts_set(sp, L("columns"), NULL, columns, 0))
-		rval = 1;
-	return (rval);
+	a.bp = b1;
+	b.bp = NULL;
+	a.len = b.len = 0;
+	argv[0] = &a;
+	argv[1] = &b;
+
+	a.len = SPRINTF(b1, sizeof(b1), L("lines=%lu"), (u_long)lines);
+	if (opts_set(sp, argv, NULL))
+		return (1);
+	a.len = SPRINTF(b1, sizeof(b1), L("columns=%lu"), (u_long)columns);
+	if (opts_set(sp, argv, NULL))
+		return (1);
+	return (0);
 }
