@@ -5,29 +5,10 @@
  *	Keith Bostic.  All rights reserved.
  *
  * See the LICENSE file for redistribution information.
- *
- *	$Id: common.h,v 10.20 2002/03/02 23:36:22 skimo Exp $ (Berkeley) $Date: 2002/03/02 23:36:22 $
  */
 
-/*
- * Avoid include sys/types.h after definition of pgno_t
- */
-#include <sys/types.h>
-#include <sys/queue.h>
-#include <bitstring.h>
-/*
- * Porting information built at configuration time.  Included before
- * any of nvi's include files.
- */
-#include "port.h"
-
-/*
- * Pseudo-local includes.  These are files that are unlikely to exist
- * on most machines to which we're porting vi, and we want to include
- * them in a very specific order, regardless.
- */
-#include "db.h"
-#include <regex.h>
+#include "/usr/include/db.h"	/* Only include db1. */
+#include <regex.h>		/* May refer to the bundled regex. */
 
 /*
  * Forward structure declarations.  Not pretty, but the include files
@@ -35,8 +16,8 @@
  */
 typedef struct _cb		CB;
 typedef struct _csc		CSC;
-typedef struct _conv	    	CONV;
-typedef struct _conv_win    	CONVWIN;
+typedef struct _conv		CONV;
+typedef struct _conv_win	CONVWIN;
 typedef struct _event		EVENT;
 typedef struct _excmd		EXCMD;
 typedef struct _exf		EXF;
@@ -54,10 +35,9 @@ typedef struct _tag		TAG;
 typedef struct _tagf		TAGF;
 typedef struct _tagq		TAGQ;
 typedef struct _text		TEXT;
-typedef struct _win		WIN;
 
 /* Autoindent state. */
-typedef enum { C_NOTSET, C_CARATSET, C_NOCHANGE, C_ZEROSET } carat_t;
+typedef enum { C_NOTSET, C_CARATSET, C_ZEROSET } carat_t;
 
 /* Busy message types. */
 typedef enum { BUSY_ON = 1, BUSY_OFF, BUSY_UPDATE } busy_t;
@@ -98,10 +78,9 @@ typedef enum { SEQ_ABBREV, SEQ_COMMAND, SEQ_INPUT } seq_t;
 #include "conv.h"		/* Required by ex.h and screen.h */
 #include "../ex/ex.h"		/* Required by gs.h. */
 #include "gs.h"			/* Required by screen.h. */
-#include "log.h"		/* Required by screen.h */
 #include "screen.h"		/* Required by exf.h. */
 #include "exf.h"
+#include "log.h"
 #include "mem.h"
-#include "vi_auto.h"
 
 #include "extern.h"
