@@ -9,14 +9,10 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: exf.c,v 10.72 2003/08/10 09:44:01 skimo Exp $ (Berkeley) $Date: 2003/08/10 09:44:01 $";
-#endif /* not lint */
-
-#include <sys/param.h>
-#include <sys/types.h>		/* XXX: param.h may not have included types.h */
+#include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/stat.h>
+#include <sys/time.h>
 
 /*
  * We include <sys/file.h>, because the flock(2) and open(2) #defines
@@ -34,15 +30,14 @@ static const char sccsid[] = "$Id: exf.c,v 10.72 2003/08/10 09:44:01 skimo Exp $
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#include <time.h>
 
 #include "common.h"
 
-static int	file_backup __P((SCR *, char *, char *));
-static void	file_cinit __P((SCR *));
-static void	file_comment __P((SCR *));
-static int	file_spath __P((SCR *, FREF *, struct stat *, int *));
-static int   	db_setup __P((SCR *sp, EXF *ep));
+static int	file_backup(SCR *, char *, char *);
+static void	file_cinit(SCR *);
+static void	file_encinit(SCR *);
+static void	file_comment(SCR *);
+static int	file_spath(SCR *, FREF *, struct stat *, int *);
 
 /*
  * file_add --
@@ -57,7 +52,7 @@ static int   	db_setup __P((SCR *sp, EXF *ep));
  * vi now remembers the last location in any file that it has ever edited,
  * not just the previously edited file.
  *
- * PUBLIC: FREF *file_add __P((SCR *, char *));
+ * PUBLIC: FREF *file_add(SCR *, char *);
  */
 FREF *
 file_add(SCR *sp, char *name)
@@ -76,15 +71,11 @@ file_add(SCR *sp, char *name)
 	 */
 	gp = sp->gp;
 	if (name != NULL)
-		for (frp = gp->frefq.cqh_first;
-		    frp != (FREF *)&gp->frefq; frp = frp->q.cqe_next) {
+		TAILQ_FOREACH_SAFE(frp, gp->frefq, q, tfrp) {
 			if (frp->name == NULL) {
-				tfrp = frp->q.cqe_next;
-				CIRCLEQ_REMOVE(&gp->frefq, frp, q);
-				if (frp->name != NULL)
-					free(frp->name);
+				TAILQ_REMOVE(gp->frefq, frp, q);
+				free(frp->name);
 				free(frp);
-				frp = tfrp;
 				continue;
 			}
 			if (!strcmp(frp->name, name))
@@ -92,7 +83,7 @@ file_add(SCR *sp, char *name)
 		}
 
 	/* Allocate and initialize the FREF structure. */
-	CALLOC(sp, frp, FREF *, 1, sizeof(FREF));
+	CALLOC(sp, frp, 1, sizeof(FREF));
 	if (frp == NULL)
 		return (NULL);
 
@@ -109,7 +100,7 @@ file_add(SCR *sp, char *name)
 	}
 
 	/* Append into the chain of file names. */
-	CIRCLEQ_INSERT_TAIL(&gp->frefq, frp, q);
+	TAILQ_INSERT_TAIL(gp->frefq, frp, q);
 
 	return (frp);
 }
@@ -120,18 +111,19 @@ file_add(SCR *sp, char *name)
  *	let go of any previous file.  Don't release the previous file until
  *	absolutely sure we have the new one.
  *
- * PUBLIC: int file_init __P((SCR *, FREF *, char *, int));
+ * PUBLIC: int file_init(SCR *, FREF *, char *, int);
  */
 int
 file_init(SCR *sp, FREF *frp, char *rcv_name, int flags)
 {
 	EXF *ep;
+	RECNOINFO oinfo = { 0 };
 	struct stat sb;
 	size_t psize;
-	int fd, exists, open_err, readonly, stolen;
-	char *oname, tname[MAXPATHLEN];
+	int fd, exists, open_err, readonly;
+	char *oname, *tname;
 
-	stolen = open_err = readonly = 0;
+	open_err = readonly = 0;
 
 	/*
 	 * If the file is a recovery file, let the recovery code handle it.
@@ -151,42 +143,24 @@ file_init(SCR *sp, FREF *frp, char *rcv_name, int flags)
 	 */
 	F_CLR(frp, ~FR_CURSORSET);
 
-	/*
-	 * Scan the user's path to find the file that we're going to
-	 * try and open.
-	 */
-	if (file_spath(sp, frp, &sb, &exists))
-		return (1);
-
-	/*
-	 * Check whether we already have this file opened in some
-	 * other screen.
-	 */
-	if (exists) {
-		EXF *exfp;
-		for (exfp = sp->gp->exfq.cqh_first;
-		    exfp != (EXF *)&sp->gp->exfq; exfp = exfp->q.cqe_next) {
-			if (exfp->mdev == sb.st_dev &&
-			    exfp->minode == sb.st_ino && 
-			    (exfp != sp->ep || exfp->refcnt > 1)) {
-				ep = exfp;
-				goto postinit;
-			}
-		}
-	}
-
 	/*
 	 * Required EXF initialization:
 	 *	Flush the line caches.
 	 *	Default recover mail file fd to -1.
 	 *	Set initial EXF flag bits.
 	 */
-	CALLOC_RET(sp, ep, EXF *, 1, sizeof(EXF));
-	CIRCLEQ_INIT(&ep->scrq);
-	sp->c_lno = ep->c_nlines = OOBLNO;
-	ep->rcv_fd = ep->fcntl_fd = -1;
+	CALLOC_RET(sp, ep, 1, sizeof(EXF));
+	ep->c_lno = ep->c_nlines = OOBLNO;
+	ep->rcv_fd = -1;
 	F_SET(ep, F_FIRSTMODIFY);
 
+	/*
+	 * Scan the user's path to find the file that we're going to
+	 * try and open.
+	 */
+	if (file_spath(sp, frp, &sb, &exists))
+		return (1);
+
 	/*
 	 * If no name or backing file, for whatever reason, create a backing
 	 * temporary file, saving the temp file name so we can later unlink
@@ -195,54 +169,56 @@ file_init(SCR *sp, FREF *frp, char *rcv_name, int flags)
 	 */
 	oname = frp->name;
 	if (LF_ISSET(FS_OPENERR) || oname == NULL || !exists) {
-		if (opts_empty(sp, O_TMP_DIRECTORY, 0))
+		struct stat sb;
+
+		if (opts_empty(sp, O_TMPDIR, 0))
 			goto err;
-		(void)snprintf(tname, sizeof(tname),
-		    "%s/vi.XXXXXX", O_STR(sp, O_TMP_DIRECTORY));
-		if ((fd = mkstemp(tname)) == -1) {
+		if ((tname =
+		    join(O_STR(sp, O_TMPDIR), "vi.XXXXXXXXXX")) == NULL) {
+			msgq(sp, M_SYSERR, NULL);
+			goto err;
+		}
+		if ((fd = mkstemp(tname)) == -1 || fstat(fd, &sb)) {
+			free(tname);
 			msgq(sp, M_SYSERR,
 			    "237|Unable to create temporary file");
 			goto err;
 		}
 		(void)close(fd);
 
-		if (frp->name == NULL)
+		frp->tname = tname;
+		if (frp->name == NULL) {
 			F_SET(frp, FR_TMPFILE);
-		if ((frp->tname = strdup(tname)) == NULL ||
-		    (frp->name == NULL && 
-		     (frp->name = strdup(tname)) == NULL)) {
-			if (frp->tname != NULL) {
-				free(frp->tname);
+			if ((frp->name = strdup(tname)) == NULL) {
+				msgq(sp, M_SYSERR, NULL);
+				goto err;
 			}
-			msgq(sp, M_SYSERR, NULL);
-			(void)unlink(tname);
-			goto err;
 		}
 		oname = frp->tname;
 		psize = 1024;
 		if (!LF_ISSET(FS_OPENERR))
 			F_SET(frp, FR_NEWFILE);
 
-		time(&ep->mtime);
+		ep->mtim = sb.st_mtimespec;
 	} else {
 		/*
 		 * XXX
 		 * A seat of the pants calculation: try to keep the file in
-		 * 15 pages or less.  Don't use a page size larger than 10K
+		 * 15 pages or less.  Don't use a page size larger than 16K
 		 * (vi should have good locality) or smaller than 1K.
 		 */
 		psize = ((sb.st_size / 15) + 1023) / 1024;
-		if (psize > 10)
-			psize = 10;
+		if (psize > 16)
+			psize = 16;
 		if (psize == 0)
 			psize = 1;
-		psize *= 1024;
+		psize = p2roundup(psize) << 10;
 
 		F_SET(ep, F_DEVSET);
 		ep->mdev = sb.st_dev;
 		ep->minode = sb.st_ino;
 
-		ep->mtime = sb.st_mtime;
+		ep->mtim = sb.st_mtimespec;
 
 		if (!S_ISREG(sb.st_mode))
 			msgq_str(sp, M_ERR, oname,
@@ -250,48 +226,30 @@ file_init(SCR *sp, FREF *frp, char *rcv_name, int flags)
 	}
 
 	/* Set up recovery. */
+	oinfo.bval = '\n';			/* Always set. */
+	oinfo.psize = psize;
+	oinfo.flags = F_ISSET(sp->gp, G_SNAPSHOT) ? R_SNAPSHOT : 0;
 	if (rcv_name == NULL) {
-		/* ep->rcv_path NULL if rcv_tmp fails */
-		rcv_tmp(sp, ep, frp->name);
+		if (!rcv_tmp(sp, ep, frp->name))
+			oinfo.bfname = ep->rcv_path;
 	} else {
 		if ((ep->rcv_path = strdup(rcv_name)) == NULL) {
 			msgq(sp, M_SYSERR, NULL);
 			goto err;
 		}
+		oinfo.bfname = ep->rcv_path;
 		F_SET(ep, F_MODIFIED);
 	}
 
-	if (db_setup(sp, ep))
-		goto err;
-
 	/* Open a db structure. */
-	if ((sp->db_error = db_create(&ep->db, 0, 0)) != 0) {
-		msgq(sp, M_DBERR, "db_create");
-		goto err;
-	}
-
-	ep->db->set_re_delim(ep->db, '\n');		/* Always set. */
-	ep->db->set_pagesize(ep->db, psize);
-	ep->db->set_flags(ep->db, DB_RENUMBER | DB_SNAPSHOT);
-	if (rcv_name == NULL)
-		ep->db->set_re_source(ep->db, oname);
-
-/* 
- * Don't let db use mmap when using fcntl for locking
- */
-#ifdef HAVE_LOCK_FCNTL
-#define NOMMAPIFFCNTL DB_NOMMAP
-#else
-#define NOMMAPIFFCNTL 0
-#endif
-
-#define _DB_OPEN_MODE	S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH
-
-	if ((sp->db_error = db_open(ep->db, ep->rcv_path, DB_RECNO,
-	    ((rcv_name == 0) ? DB_TRUNCATE : 0) | VI_DB_THREAD | NOMMAPIFFCNTL,
-	    _DB_OPEN_MODE)) != 0) {
+	if ((ep->db = dbopen(rcv_name == NULL ? oname : NULL,
+	    O_NONBLOCK | O_RDONLY,
+	    S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH,
+	    DB_RECNO, &oinfo)) == NULL) {
 		msgq_str(sp,
-		    M_DBERR, rcv_name == NULL ? oname : rcv_name, "%s");
+		    M_SYSERR, rcv_name == NULL ? oname : rcv_name, "%s");
+		if (F_ISSET(frp, FR_NEWFILE))
+			goto err;
 		/*
 		 * !!!
 		 * Historically, vi permitted users to edit files that couldn't
@@ -299,33 +257,10 @@ file_init(SCR *sp, FREF *frp, char *rcv_name, int flags)
 		 * line, but it's quite useful for "vi *.c", since you can skip
 		 * past files that you can't read.
 		 */ 
-		ep->db = NULL; /* Don't close it; it wasn't opened */
-
-		if (LF_ISSET(FS_OPENERR)) 
-		    goto err;
-
 		open_err = 1;
 		goto oerr;
 	}
 
-	/* re_source is loaded into the database.
-	 * Close it and reopen it in the environment. 
-	 */
-	if ((sp->db_error = ep->db->close(ep->db, 0))) {
-		msgq(sp, M_DBERR, "close");
-		goto err;
-	}
-	if ((sp->db_error = db_create(&ep->db, ep->env, 0)) != 0) {
-		msgq(sp, M_DBERR, "db_create 2");
-		goto err;
-	}
-	if ((sp->db_error = db_open(ep->db, ep->rcv_path, DB_RECNO,
-	    VI_DB_THREAD | NOMMAPIFFCNTL, _DB_OPEN_MODE)) != 0) {
-		msgq_str(sp,
-		    M_DBERR, ep->rcv_path, "%s");
-		goto err;
-	}
-
 	/*
 	 * Do the remaining things that can cause failure of the new file,
 	 * mark and logging initialization.
@@ -333,7 +268,6 @@ file_init(SCR *sp, FREF *frp, char *rcv_name, int flags)
 	if (mark_init(sp, ep) || log_init(sp, ep))
 		goto err;
 
-postinit:
 	/*
 	 * Set the alternate file name to be the file we're discarding.
 	 *
@@ -368,7 +302,6 @@ postinit:
 			(void)file_end(sp, ep, 1);
 			goto err;
 		}
-		sp->ep = NULL;
 		F_CLR(frp, FR_DONTDELETE);
 	}
 
@@ -391,33 +324,30 @@ postinit:
 	 * when locking is a little more reliable, this should change to be
 	 * an error.
 	 */
-	if (rcv_name == NULL && ep->refcnt == 0) {
-		if ((ep->fd = open(oname, O_RDWR)) == -1)
-		    goto no_lock;
-
-		switch (file_lock(sp, oname, &ep->fcntl_fd, ep->fd, 1)) {
+	if (rcv_name == NULL)
+		switch (file_lock(sp, oname, ep->db->fd(ep->db), 0)) {
 		case LOCK_FAILED:
-no_lock:
 			F_SET(frp, FR_UNLOCKED);
 			break;
 		case LOCK_UNAVAIL:
 			readonly = 1;
+			if (F_ISSET(sp, SC_READONLY))
+				break;
 			msgq_str(sp, M_INFO, oname,
 			    "239|%s already locked, session is read-only");
 			break;
 		case LOCK_SUCCESS:
 			break;
 		}
-	}
 
 	/*
          * Historically, the readonly edit option was set per edit buffer in
          * vi, unless the -R command-line option was specified or the program
          * was executed as "view".  (Well, to be truthful, if the letter 'w'
          * occurred anywhere in the program name, but let's not get into that.)
-	 * So, the persistant readonly state has to be stored in the screen
+	 * So, the persistent readonly state has to be stored in the screen
 	 * structure, and the edit option value toggles with the contents of
-	 * the edit buffer.  If the persistant readonly flag is set, set the
+	 * the edit buffer.  If the persistent readonly flag is set, set the
 	 * readonly edit option.
 	 *
 	 * Otherwise, try and figure out if a file is readonly.  This is a
@@ -463,32 +393,22 @@ no_lock:
 
 	/* Switch... */
 	++ep->refcnt;
-	CIRCLEQ_INSERT_HEAD(&ep->scrq, sp, eq);
 	sp->ep = ep;
 	sp->frp = frp;
 
+	/* Detect and set the file encoding */
+	file_encinit(sp);
+
 	/* Set the initial cursor position, queue initial command. */
 	file_cinit(sp);
 
-	/* Report conversion errors again. */
-	F_CLR(sp, SC_CONV_ERROR);
-
 	/* Redraw the screen from scratch, schedule a welcome message. */
 	F_SET(sp, SC_SCR_REFORMAT | SC_STATUS);
 
-	if (frp->lno == OOBLNO)
-		F_SET(sp, SC_SCR_TOP);
-
-	/* Append into the chain of file structures. */
-	if (ep->refcnt == 1)
-		CIRCLEQ_INSERT_TAIL(&sp->gp->exfq, ep, q);
-
 	return (0);
 
-err:	if (frp->name != NULL) {
-		free(frp->name);
-		frp->name = NULL;
-	}
+err:	free(frp->name);
+	frp->name = NULL;
 	if (frp->tname != NULL) {
 		(void)unlink(frp->tname);
 		free(frp->tname);
@@ -497,17 +417,14 @@ err:	if (frp->name != NULL) {
 
 oerr:	if (F_ISSET(ep, F_RCV_ON))
 		(void)unlink(ep->rcv_path);
-	if (ep->rcv_path != NULL) {
-		free(ep->rcv_path);
-		ep->rcv_path = NULL;
-	}
-	if (ep->db != NULL) {
-		(void)ep->db->close(ep->db, DB_NOSYNC);
-		ep->db = NULL;
-	}
+	free(ep->rcv_path);
+	ep->rcv_path = NULL;
+
+	if (ep->db != NULL)
+		(void)ep->db->close(ep->db);
 	free(ep);
 
-	return (open_err && !LF_ISSET(FS_OPENERR) ?
+	return (open_err ?
 	    file_init(sp, frp, rcv_name, flags | FS_OPENERR) : 1);
 }
 
@@ -519,10 +436,10 @@ oerr:	if (F_ISSET(ep, F_RCV_ON))
 static int
 file_spath(SCR *sp, FREF *frp, struct stat *sbp, int *existsp)
 {
-	CHAR_T savech;
+	int savech;
 	size_t len;
 	int found;
-	char *name, *p, *t, path[MAXPATHLEN];
+	char *name, *p, *t, *path;
 
 	/*
 	 * If the name is NULL or an explicit reference (i.e., the first
@@ -548,16 +465,24 @@ file_spath(SCR *sp, FREF *frp, struct stat *sbp, int *existsp)
 	/* Try the O_PATH option values. */
 	for (found = 0, p = t = O_STR(sp, O_PATH);; ++p)
 		if (*p == ':' || *p == '\0') {
-			if (t < p - 1) {
+			/*
+			 * Ignore the empty strings and ".", since we've already
+			 * tried the current directory.
+			 */
+			if (t < p && (p - t != 1 || *t != '.')) {
 				savech = *p;
 				*p = '\0';
-				len = snprintf(path,
-				    sizeof(path), "%s/%s", t, name);
+				if ((path = join(t, name)) == NULL) {
+					msgq(sp, M_SYSERR, NULL);
+					break;
+				}
+				len = strlen(path);
 				*p = savech;
 				if (!stat(path, sbp)) {
 					found = 1;
 					break;
 				}
+				free(path);
 			}
 			t = p + 1;
 			if (*p == '\0')
@@ -566,10 +491,8 @@ file_spath(SCR *sp, FREF *frp, struct stat *sbp, int *existsp)
 
 	/* If we found it, build a new pathname and discard the old one. */
 	if (found) {
-		MALLOC_RET(sp, p, char *, len + 1);
-		memcpy(p, path, len + 1);
 		free(frp->name);
-		frp->name = p;
+		frp->name = path;
 	}
 	*existsp = found;
 	return (0);
@@ -623,7 +546,7 @@ file_cinit(SCR *sp)
 		}
 		CHAR2INT(sp, gp->c_option, strlen(gp->c_option) + 1,
 			 wp, wlen);
-		if (ex_run_str(sp, "-c option", wp, wlen - 1, 1, 1))
+		if (ex_run_str(sp, "-c option", wp, wlen - 1, 1, 0))
 			return;
 		gp->c_option = NULL;
 	} else if (F_ISSET(sp, SC_EX)) {
@@ -688,7 +611,7 @@ file_cinit(SCR *sp)
  * file_end --
  *	Stop editing a file.
  *
- * PUBLIC: int file_end __P((SCR *, EXF *, int));
+ * PUBLIC: int file_end(SCR *, EXF *, int);
  */
 int
 file_end(SCR *sp, EXF *ep, int force)
@@ -704,7 +627,6 @@ file_end(SCR *sp, EXF *ep, int force)
 	 */
 	if (ep == NULL)
 		ep = sp->ep;
-	CIRCLEQ_REMOVE(&ep->scrq, sp, eq);
 	if (--ep->refcnt != 0)
 		return (0);
 
@@ -737,9 +659,8 @@ file_end(SCR *sp, EXF *ep, int force)
 		free(frp->tname);
 		frp->tname = NULL;
 		if (F_ISSET(frp, FR_TMPFILE)) {
-			CIRCLEQ_REMOVE(&sp->gp->frefq, frp, q);
-			if (frp->name != NULL)
-				free(frp->name);
+			TAILQ_REMOVE(sp->gp->frefq, frp, q);
+			free(frp->name);
 			free(frp);
 		}
 		sp->frp = NULL;
@@ -750,15 +671,10 @@ file_end(SCR *sp, EXF *ep, int force)
 	 *
 	 * Close the db structure.
 	 */
-	if (ep->db->close != NULL) {
-		if ((sp->db_error = ep->db->close(ep->db, DB_NOSYNC)) != 0 && 
-		    !force) {
-			msgq_str(sp, M_DBERR, frp->name, "241|%s: close");
-			CIRCLEQ_INSERT_HEAD(&ep->scrq, sp, eq);
-			++ep->refcnt;
-			return (1);
-		}
-		ep->db = NULL;
+	if (ep->db->close != NULL && ep->db->close(ep->db) && !force) {
+		msgq_str(sp, M_SYSERR, frp->name, "241|%s: close");
+		++ep->refcnt;
+		return (1);
 	}
 
 	/* COMMITTED TO THE CLOSE.  THERE'S NO GOING BACK... */
@@ -769,19 +685,6 @@ file_end(SCR *sp, EXF *ep, int force)
 	/* Free up any marks. */
 	(void)mark_end(sp, ep);
 
-	if (ep->env) {
-		DB_ENV *env;
-
-		ep->env->close(ep->env, 0);
-		ep->env = 0;
-		if ((sp->db_error = db_env_create(&env, 0)))
-			msgq(sp, M_DBERR, "env_create");
-		if ((sp->db_error = db_env_remove(env, ep->env_path, 0)))
-			msgq(sp, M_DBERR, "env->remove");
-		if (ep->env_path != NULL && rmdir(ep->env_path))
-			msgq_str(sp, M_SYSERR, ep->env_path, "242|%s: remove");
-	}
-
 	/*
 	 * Delete recovery files, close the open descriptor, free recovery
 	 * memory.  See recover.c for a description of the protocol.
@@ -797,19 +700,12 @@ file_end(SCR *sp, EXF *ep, int force)
 		if (ep->rcv_mpath != NULL && unlink(ep->rcv_mpath))
 			msgq_str(sp, M_SYSERR, ep->rcv_mpath, "243|%s: remove");
 	}
-	CIRCLEQ_REMOVE(&sp->gp->exfq, ep, q);
-	if (ep->fd != -1)
-		(void)close(ep->fd);
-	if (ep->fcntl_fd != -1)
-		(void)close(ep->fcntl_fd);
 	if (ep->rcv_fd != -1)
 		(void)close(ep->rcv_fd);
-	if (ep->env_path != NULL)
-		free(ep->env_path);
-	if (ep->rcv_path != NULL)
-		free(ep->rcv_path);
-	if (ep->rcv_mpath != NULL)
-		free(ep->rcv_mpath);
+	free(ep->rcv_path);
+	free(ep->rcv_mpath);
+	if (ep->c_blen > 0)
+		free(ep->c_lp);
 
 	free(ep);
 	return (0);
@@ -821,7 +717,7 @@ file_end(SCR *sp, EXF *ep, int force)
  *	semantics for whether or not writes would happen.  That's
  *	why all the flags.
  *
- * PUBLIC: int file_write __P((SCR *, MARK *, MARK *, char *, int));
+ * PUBLIC: int file_write(SCR *, MARK *, MARK *, char *, int);
  */
 int
 file_write(SCR *sp, MARK *fm, MARK *tm, char *name, int flags)
@@ -835,7 +731,7 @@ file_write(SCR *sp, MARK *fm, MARK *tm, char *name, int flags)
 	size_t len;
 	u_long nlno, nch;
 	int fd, nf, noname, oflags, rval;
-	char *p, *s, *t, buf[MAXPATHLEN + 64];
+	char *p, *s, *t, buf[1024];
 	const char *msgstr;
 
 	ep = sp->ep;
@@ -900,7 +796,7 @@ file_write(SCR *sp, MARK *fm, MARK *tm, char *name, int flags)
 		if (noname && !LF_ISSET(FS_FORCE | FS_APPEND) &&
 		    ((F_ISSET(ep, F_DEVSET) &&
 		    (sb.st_dev != ep->mdev || sb.st_ino != ep->minode)) ||
-		    sb.st_mtime != ep->mtime)) {
+		    timespeccmp(&sb.st_mtimespec, &ep->mtim, !=))) {
 			msgq_str(sp, M_ERR, name, LF_ISSET(FS_POSSIBLE) ?
 "250|%s: file modified more recently than this copy; use ! to override" :
 "251|%s: file modified more recently than this copy");
@@ -920,35 +816,46 @@ file_write(SCR *sp, MARK *fm, MARK *tm, char *name, int flags)
 		return (1);
 
 	/* Open the file. */
-	SIGBLOCK;
 	if ((fd = open(name, oflags,
 	    S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) < 0) {
+		if (errno == EACCES && LF_ISSET(FS_FORCE)) {
+			/*
+			 * If the user owns the file but does not
+			 * have write permission on it, grant it
+			 * automatically for the duration of the
+			 * opening of the file, if possible.
+			 */
+			struct stat sb;
+			mode_t fmode;
+
+			if (stat(name, &sb) != 0)
+				goto fail_open;
+			fmode = sb.st_mode;
+			if (!(sb.st_mode & S_IWUSR) && sb.st_uid == getuid())
+				fmode |= S_IWUSR;
+			else
+				goto fail_open;
+			if (chmod(name, fmode) != 0)
+				goto fail_open;
+			fd = open(name, oflags, S_IRUSR | S_IWUSR |
+			    S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+			if (fd == -1)
+				goto fail_open;
+			(void)fchmod(fd, sb.st_mode);
+			goto success_open;
+		fail_open:
+			errno = EACCES;
+		}
 		msgq_str(sp, M_SYSERR, name, "%s");
-		SIGUNBLOCK;
 		return (1);
 	}
-	SIGUNBLOCK;
+success_open:
 
 	/* Try and get a lock. */
-	if (!noname && file_lock(sp, NULL, NULL, fd, 0) == LOCK_UNAVAIL)
+	if (!noname && file_lock(sp, NULL, fd, 0) == LOCK_UNAVAIL)
 		msgq_str(sp, M_ERR, name,
 		    "252|%s: write lock was unavailable");
 
-#if __linux__
-	/*
-	 * XXX
-	 * In libc 4.5.x, fdopen(fd, "w") clears the O_APPEND flag (if set).
-	 * This bug is fixed in libc 4.6.x.
-	 *
-	 * This code works around this problem for libc 4.5.x users.
-	 * Note that this code is harmless if you're using libc 4.6.x.
-	 */
-	if (LF_ISSET(FS_APPEND) && lseek(fd, (off_t)0, SEEK_END) < 0) {
-		msgq(sp, M_SYSERR, name);
-		return (1);
-	}
-#endif
-
 	/*
 	 * Use stdio for buffering.
 	 *
@@ -980,17 +887,16 @@ file_write(SCR *sp, MARK *fm, MARK *tm, char *name, int flags)
 	 * we re-init the time.  That way the user can clean up the disk
 	 * and rewrite without having to force it.
 	 */
-	if (noname) {
+	if (noname)
 		if (stat(name, &sb))
-			time(&ep->mtime);
+			timepoint_system(&ep->mtim);
 		else {
 			F_SET(ep, F_DEVSET);
 			ep->mdev = sb.st_dev;
 			ep->minode = sb.st_ino;
 
-			ep->mtime = sb.st_mtime;
+			ep->mtim = sb.st_mtimespec;
 		}
-	}
 
 	/*
 	 * If the write failed, complain loudly.  ex_writefp() has already
@@ -1019,12 +925,11 @@ file_write(SCR *sp, MARK *fm, MARK *tm, char *name, int flags)
 	 */
 	if (LF_ISSET(FS_ALL) && !LF_ISSET(FS_APPEND)) {
 		F_CLR(ep, F_MODIFIED);
-		if (F_ISSET(frp, FR_TMPFILE)) {
+		if (F_ISSET(frp, FR_TMPFILE))
 			if (noname)
 				F_SET(frp, FR_TMPEXIT);
 			else
 				F_CLR(frp, FR_TMPEXIT);
-		}
 	}
 
 	p = msg_print(sp, name, &nf);
@@ -1063,7 +968,7 @@ file_write(SCR *sp, MARK *fm, MARK *tm, char *name, int flags)
 			*--s = '.';
 		}
 	}
-	msgq(sp, M_INFO, s);
+	msgq(sp, M_INFO, "%s", s);
 	if (nf)
 		FREE_SPACE(sp, p, 0);
 	return (0);
@@ -1131,9 +1036,14 @@ file_backup(SCR *sp, char *name, char *bname)
 		++bname;
 	} else
 		version = 0;
-	CHAR2INT(sp, bname, strlen(bname) + 1, wp, wlen);
-	if (argv_exp2(sp, &cmd, wp, wlen - 1))
+	CHAR2INT(sp, bname, strlen(bname), wp, wlen);
+	if ((wp = v_wstrdup(sp, wp, wlen)) == NULL)
 		return (1);
+	if (argv_exp2(sp, &cmd, wp, wlen)) {
+		free(wp);
+		return (1);
+	}
+	free(wp);
 
 	/*
 	 *  0 args: impossible.
@@ -1157,7 +1067,7 @@ file_backup(SCR *sp, char *name, char *bname)
 	 */
 	if (version) {
 		GET_SPACE_GOTOC(sp, bp, blen, cmd.argv[0]->len * 2 + 50);
-		INT2SYS(sp, cmd.argv[0]->bp, cmd.argv[0]->len + 1,
+		INT2CHAR(sp, cmd.argv[0]->bp, cmd.argv[0]->len + 1,
 			 p, nlen); 
 		d = strdup(p);
 		p = d;
@@ -1183,7 +1093,7 @@ file_backup(SCR *sp, char *name, char *bname)
 			p = slash + 1;
 		}
 		if (dirp == NULL) {
-			INT2SYS(sp, cmd.argv[0]->bp, cmd.argv[0]->len + 1,
+			INT2CHAR(sp, cmd.argv[0]->bp, cmd.argv[0]->len + 1,
 				estr, nlen);
 			goto err;
 		}
@@ -1198,7 +1108,7 @@ file_backup(SCR *sp, char *name, char *bname)
 		wfname = bp;
 	} else {
 		bp = NULL;
-		INT2SYS(sp, cmd.argv[0]->bp, cmd.argv[0]->len + 1,
+		INT2CHAR(sp, cmd.argv[0]->bp, cmd.argv[0]->len + 1,
 			wfname, nlen);
 	}
 	
@@ -1246,6 +1156,7 @@ file_backup(SCR *sp, char *name, char *bname)
 		estr = wfname;
 		goto err;
 	}
+	free(d);
 	if (bp != NULL)
 		FREE_SPACE(sp, bp, blen);
 	return (0);
@@ -1259,13 +1170,58 @@ err:	if (rfd != -1)
 	}
 	if (estr)
 		msgq_str(sp, M_SYSERR, estr, "%s");
-	if (d != NULL)
-		free(d);
+	free(d);
 	if (bp != NULL)
 		FREE_SPACE(sp, bp, blen);
 	return (1);
 }
 
+/*
+ * file_encinit --
+ *	Read the first line and set the O_FILEENCODING.
+ */
+static void
+file_encinit(SCR *sp)
+{
+#if defined(USE_WIDECHAR) && defined(USE_ICONV)
+	size_t len;
+	char *p;
+	size_t blen = 0;
+	char buf[4096];	/* not need to be '\0'-terminated */
+	recno_t ln = 1;
+	EXF *ep;
+
+	ep = sp->ep;
+
+	while (!db_rget(sp, ln++, &p, &len)) {
+		if (blen + len > sizeof(buf))
+			len = sizeof(buf) - blen;
+		memcpy(buf + blen, p, len);
+		blen += len;
+		if (blen == sizeof(buf))
+			break;
+		else
+			buf[blen++] = '\n';
+	}
+
+	/*
+	 * Detect UTF-8 and fallback to the locale/preset encoding.
+	 *
+	 * XXX
+	 * A manually set O_FILEENCODING indicates the "fallback
+	 * encoding", but UTF-8, which can be safely detected, is not
+	 * inherited from the old screen.
+	 */
+	if (looks_utf8(buf, blen) > 1)
+		o_set(sp, O_FILEENCODING, OS_STRDUP, "utf-8", 0);
+	else if (!O_ISSET(sp, O_FILEENCODING) ||
+	    !strcasecmp(O_STR(sp, O_FILEENCODING), "utf-8"))
+		o_set(sp, O_FILEENCODING, OS_STRDUP, codeset(), 0);
+
+	conv_enc(sp, O_FILEENCODING, 0);
+#endif
+}
+
 /*
  * file_comment --
  *	Skip the first comment.
@@ -1273,7 +1229,7 @@ err:	if (rfd != -1)
 static void
 file_comment(SCR *sp)
 {
-	db_recno_t lno;
+	recno_t lno;
 	size_t len;
 	CHAR_T *p;
 
@@ -1298,11 +1254,15 @@ file_comment(SCR *sp)
 		} while (!db_get(sp, ++lno, 0, &p, &len));
 	} else if (len > 1 && p[0] == '/' && p[1] == '/') {
 		F_SET(sp, SC_SCR_TOP);
-		while (!db_get(sp, ++lno, 0, &p, &len))
-			if (len < 1 || p[0] != '/' || p[1] != '/') {
-				sp->lno = lno;
-				return;
-			}
+		p += 2;
+		len -= 2;
+		do {
+			for (; len > 1; --len, ++p)
+				if (p[0] == '/' && p[1] == '/') {
+					sp->lno = lno;
+					return;
+				}
+		} while (!db_get(sp, ++lno, 0, &p, &len));
 	}
 }
 
@@ -1311,7 +1271,7 @@ file_comment(SCR *sp)
  * 	First modification check routine.  The :next, :prev, :rewind, :tag,
  *	:tagpush, :tagpop, ^^ modifications check.
  *
- * PUBLIC: int file_m1 __P((SCR *, int, int));
+ * PUBLIC: int file_m1(SCR *, int, int);
  */
 int
 file_m1(SCR *sp, int force, int flags)
@@ -1330,7 +1290,7 @@ file_m1(SCR *sp, int force, int flags)
 	 * unless force is also set.  Otherwise, we fail unless forced or
 	 * there's another open screen on this file.
 	 */
-	if (F_ISSET(ep, F_MODIFIED)) {
+	if (F_ISSET(ep, F_MODIFIED))
 		if (O_ISSET(sp, O_AUTOWRITE)) {
 			if (!force && file_aw(sp, flags))
 				return (1);
@@ -1340,7 +1300,6 @@ file_m1(SCR *sp, int force, int flags)
 "263|File modified since last complete write; write or use :edit! to override");
 			return (1);
 		}
-	}
 
 	return (file_m3(sp, force));
 }
@@ -1350,7 +1309,7 @@ file_m1(SCR *sp, int force, int flags)
  * 	Second modification check routine.  The :edit, :quit, :recover
  *	modifications check.
  *
- * PUBLIC: int file_m2 __P((SCR *, int));
+ * PUBLIC: int file_m2(SCR *, int);
  */
 int
 file_m2(SCR *sp, int force)
@@ -1380,7 +1339,7 @@ file_m2(SCR *sp, int force)
  * file_m3 --
  * 	Third modification check routine.
  *
- * PUBLIC: int file_m3 __P((SCR *, int));
+ * PUBLIC: int file_m3(SCR *, int);
  */
 int
 file_m3(SCR *sp, int force)
@@ -1414,7 +1373,7 @@ file_m3(SCR *sp, int force)
  *	is not set, write the file.  A routine so there's a place to put the
  *	comment.
  *
- * PUBLIC: int file_aw __P((SCR *, int));
+ * PUBLIC: int file_aw(SCR *, int);
  */
 int
 file_aw(SCR *sp, int flags)
@@ -1473,13 +1432,12 @@ file_aw(SCR *sp, int flags)
  * If the user edits a temporary file, there may be times when there is no
  * alternative file name.  A name argument of NULL turns it off.
  *
- * PUBLIC: void set_alt_name __P((SCR *, char *));
+ * PUBLIC: void set_alt_name(SCR *, char *);
  */
 void
 set_alt_name(SCR *sp, char *name)
 {
-	if (sp->alt_name != NULL)
-		free(sp->alt_name);
+	free(sp->alt_name);
 	if (name == NULL)
 		sp->alt_name = NULL;
 	else if ((sp->alt_name = strdup(name)) == NULL)
@@ -1488,36 +1446,16 @@ set_alt_name(SCR *sp, char *name)
 
 /*
  * file_lock --
- *	Get an exclusive lock on a file and set close-on-exec flag
+ *	Get an exclusive lock on a file.
  *
- * XXX
- * The default locking is flock(2) style, not fcntl(2).  The latter is
- * known to fail badly on some systems, and its only advantage is that
- * it occasionally works over NFS.
- *
- * Furthermore, the semantics of fcntl(2) are wrong.  The problems are
- * two-fold: you can't close any file descriptor associated with the file
- * without losing all of the locks, and you can't get an exclusive lock
- * unless you have the file open for writing.  Someone ought to be shot,
- * but it's probably too late, they may already have reproduced.  To get
- * around these problems, nvi opens the files for writing when it can and
- * acquires a second file descriptor when it can't.  The recovery files
- * are examples of the former, they're always opened for writing.  The DB
- * files can't be opened for writing because the semantics of DB are that
- * files opened for writing are flushed back to disk when the DB session
- * is ended. So, in that case we have to acquire an extra file descriptor.
- *
- * PUBLIC: lockr_t file_lock __P((SCR *, char *, int *, int, int));
+ * PUBLIC: lockr_t file_lock(SCR *, char *, int, int);
  */
 lockr_t
-file_lock(SCR *sp, char *name, int *fdp, int fd, int iswrite)
+file_lock(SCR *sp, char *name, int fd, int iswrite)
 {
-	fcntl(fd, F_SETFD, 1);
-
 	if (!O_ISSET(sp, O_LOCKFILES))
 		return (LOCK_SUCCESS);
 	
-#ifdef HAVE_LOCK_FLOCK			/* Hurrah!  We've got flock(2). */
 	/*
 	 * !!!
 	 * We need to distinguish a lock not being available for the file
@@ -1526,117 +1464,13 @@ file_lock(SCR *sp, char *name, int *fdp, int fd, int iswrite)
 	 * they are the former.  There's no portable way to do this.
 	 */
 	errno = 0;
-	return (flock(fd, LOCK_EX | LOCK_NB) ? errno == EAGAIN
+	if (!flock(fd, LOCK_EX | LOCK_NB)) {
+		fcntl(fd, F_SETFD, 1);
+		return (LOCK_SUCCESS);
+	}
+	return (errno == EAGAIN
 #ifdef EWOULDBLOCK
 	    || errno == EWOULDBLOCK
 #endif
-	    ? LOCK_UNAVAIL : LOCK_FAILED : LOCK_SUCCESS);
-#endif
-#ifdef HAVE_LOCK_FCNTL			/* Gag me.  We've got fcntl(2). */
-{
-	struct flock arg;
-	int didopen, sverrno;
-
-	arg.l_type = F_WRLCK;
-	arg.l_whence = 0;		/* SEEK_SET */
-	arg.l_start = arg.l_len = 0;
-	arg.l_pid = 0;
-
-	/*
-	 * If the file descriptor isn't opened for writing, it must fail.
-	 * If we fail because we can't get a read/write file descriptor,
-	 * we return LOCK_SUCCESS, believing that the file is readonly
-	 * and that will be sufficient to warn the user.
-	 */
-	if (!iswrite) {
-		if (name == NULL || fdp == NULL)
-			return (LOCK_FAILED);
-		if ((fd = open(name, O_RDWR, 0)) == -1)
-			return (LOCK_SUCCESS);
-		*fdp = fd;
-		didopen = 1;
-	}
-
-	errno = 0;
-	if (!fcntl(fd, F_SETLK, &arg))
-		return (LOCK_SUCCESS);
-	if (didopen) {
-		sverrno = errno;
-		(void)close(fd);
-		errno = sverrno;
-	}
-
-	/*
-	 * !!!
-	 * We need to distinguish a lock not being available for the file
-	 * from the file system not supporting locking.  Fcntl is documented
-	 * as returning EACCESS and EAGAIN; add EWOULDBLOCK for good measure,
-	 * and assume they are the former.  There's no portable way to do this.
-	 */
-	return (errno == EACCES || errno == EAGAIN
-#ifdef EWOULDBLOCK
-	|| errno == EWOULDBLOCK
-#endif
-	?  LOCK_UNAVAIL : LOCK_FAILED);
-}
-#endif
-#if !defined(HAVE_LOCK_FLOCK) && !defined(HAVE_LOCK_FCNTL)
-	return (LOCK_SUCCESS);
-#endif
-}
-
-#ifdef USE_DB4_LOGGING
-#define VI_DB_INIT_LOG	DB_INIT_LOG
-#else
-#define VI_DB_INIT_LOG	0
-#endif
-
-static int
-db_setup(SCR *sp, EXF *ep)
-{
-	char path[MAXPATHLEN];
-	int fd;
-	DB_ENV	*env;
-
-	(void)snprintf(path, sizeof(path), "%s/vi.XXXXXX", O_STR(sp, O_RECDIR));
-	if ((fd = mkstemp(path)) == -1) {
-		msgq(sp, M_SYSERR, "%s", path);
-		goto err;
-	}
-	(void)close(fd);
-	(void)unlink(path);
-	if (mkdir(path, S_IRWXU)) {
-		msgq(sp, M_SYSERR, "%s", path);
-		goto err;
-	}
-	if (db_env_create(&env, 0)) {
-		msgq(sp, M_ERR, "env_create");
-		goto err;
-	}
-#ifdef USE_DB4_LOGGING
-	if ((sp->db_error = vi_db_init_recover(env))) {
-		msgq(sp, M_DBERR, "init_recover");
-		goto err;
-	}
-	if ((sp->db_error = __vi_init_recover(env))) {
-		msgq(sp, M_DBERR, "init_recover");
-		goto err;
-	}
-#endif
-	if ((sp->db_error = db_env_open(env, path, 
-	    DB_PRIVATE | DB_CREATE | DB_INIT_MPOOL | VI_DB_THREAD 
-	    | VI_DB_INIT_LOG, 0)) != 0) {
-		msgq(sp, M_DBERR, "env->open");
-		goto err;
-	}
-
-	if ((ep->env_path = strdup(path)) == NULL) {
-		msgq(sp, M_SYSERR, NULL);
-		(void)rmdir(path);
-		goto err;
-	}
-	ep->env = env;
-	return 0;
-err:
-	return 1;
+	    ? LOCK_UNAVAIL : LOCK_FAILED);
 }
