@@ -9,10 +9,6 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: key.c,v 10.48 2001/06/25 15:19:10 skimo Exp $ (Berkeley) $Date: 2001/06/25 15:19:10 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/time.h>
@@ -21,20 +17,20 @@ static const char sccsid[] = "$Id: key.c,v 10.48 2001/06/25 15:19:10 skimo Exp $
 #include <ctype.h>
 #include <errno.h>
 #include <limits.h>
-#include <locale.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
 #include <unistd.h>
 
 #include "common.h"
 #include "../vi/vi.h"
 
-static int	v_event_append __P((SCR *, EVENT *));
-static int	v_event_grow __P((SCR *, int));
-static int	v_key_cmp __P((const void *, const void *));
-static void	v_keyval __P((SCR *, int, scr_keyval_t));
-static void	v_sync __P((SCR *, int));
+static int	v_event_append(SCR *, EVENT *);
+static int	v_event_grow(SCR *, int);
+static int	v_key_cmp(const void *, const void *);
+static void	v_keyval(SCR *, int, scr_keyval_t);
+static void	v_sync(SCR *, int);
 
 /*
  * !!!
@@ -97,34 +93,18 @@ static int nkeylist =
  * v_key_init --
  *	Initialize the special key lookup table.
  *
- * PUBLIC: int v_key_init __P((SCR *));
+ * PUBLIC: int v_key_init(SCR *);
  */
 int
 v_key_init(SCR *sp)
 {
-	CHAR_T ch;
+	int ch;
 	GS *gp;
 	KEYLIST *kp;
 	int cnt;
 
 	gp = sp->gp;
 
-	/*
-	 * XXX
-	 * 8-bit only, for now.  Recompilation should get you any 8-bit
-	 * character set, as long as nul isn't a character.
-	 */
-	(void)setlocale(LC_ALL, "");
-#if __linux__
-	/*
-	 * In libc 4.5.26, setlocale(LC_ALL, ""), doesn't setup the table
-	 * for ctype(3c) correctly.  This bug is fixed in libc 4.6.x.
-	 *
-	 * This code works around this problem for libc 4.5.x users.
-	 * Note that this code is harmless if you're using libc 4.6.x.
-	 */
-	(void)setlocale(LC_CTYPE, "");
-#endif
 	v_key_ilookup(sp);
 
 	v_keyval(sp, K_CNTRLD, KEY_VEOF);
@@ -136,16 +116,12 @@ v_key_init(SCR *sp)
 	qsort(keylist, nkeylist, sizeof(keylist[0]), v_key_cmp);
 
 	/* Initialize the fast lookup table. */
-	for (gp->max_special = 0, kp = keylist, cnt = nkeylist; cnt--; ++kp) {
-		if (gp->max_special < kp->ch)
-			gp->max_special = kp->ch;
-		if (kp->ch <= MAX_FAST_KEY)
-			gp->special_key[kp->ch] = kp->value;
-	}
+	for (kp = keylist, cnt = nkeylist; cnt--; ++kp)
+		gp->special_key[kp->ch] = kp->value;
 
 	/* Find a non-printable character to use as a message separator. */
-	for (ch = 1; ch <= MAX_CHAR_T; ++ch)
-		if (!ISPRINT(ch)) {
+	for (ch = 1; ch <= UCHAR_MAX; ++ch)
+		if (!isprint(ch)) {
 			gp->noprint = ch;
 			break;
 		}
@@ -196,7 +172,7 @@ v_keyval(SCR *sp, int val, scr_keyval_t name)
  * v_key_ilookup --
  *	Build the fast-lookup key display array.
  *
- * PUBLIC: void v_key_ilookup __P((SCR *));
+ * PUBLIC: void v_key_ilookup(SCR *);
  */
 void
 v_key_ilookup(SCR *sp)
@@ -220,7 +196,7 @@ v_key_ilookup(SCR *sp)
  *	Return the length of the string that will display the key.
  *	This routine is the backup for the KEY_LEN() macro.
  *
- * PUBLIC: size_t v_key_len __P((SCR *, ARG_CHAR_T));
+ * PUBLIC: size_t v_key_len(SCR *, ARG_CHAR_T);
  */
 size_t
 v_key_len(SCR *sp, ARG_CHAR_T ch)
@@ -234,29 +210,41 @@ v_key_len(SCR *sp, ARG_CHAR_T ch)
  *	Return the string that will display the key.  This routine
  *	is the backup for the KEY_NAME() macro.
  *
- * PUBLIC: u_char *v_key_name __P((SCR *, ARG_CHAR_T));
+ * PUBLIC: char *v_key_name(SCR *, ARG_CHAR_T);
  */
-u_char *
+char *
 v_key_name(SCR *sp, ARG_CHAR_T ach)
 {
 	static const char hexdigit[] = "0123456789abcdef";
 	static const char octdigit[] = "01234567";
-	CHAR_T ch, mask;
+	int ch;
 	size_t len;
-	int cnt, shift;
 	char *chp;
 
-	ch = ach;
+	/*
+	 * Cache the last checked character.  It won't be a problem
+	 * since nvi will rescan the mapping when settings changed.
+	 */
+	if (ach && sp->lastc == ach)
+		return (sp->cname);
+	sp->lastc = ach;
+
+#ifdef USE_WIDECHAR
+	len = wctomb(sp->cname, ach);
+	if (len > MB_CUR_MAX)
+#endif
+		sp->cname[(len = 1)-1] = (u_char)ach;
+
+	ch = (u_char)sp->cname[0];
+	sp->cname[len] = '\0';
 
 	/* See if the character was explicitly declared printable or not. */
 	if ((chp = O_STR(sp, O_PRINT)) != NULL)
-		for (; *chp != '\0'; ++chp)
-			if (*chp == ch)
-				goto pr;
+		if (strstr(chp, sp->cname) != NULL)
+			goto done;
 	if ((chp = O_STR(sp, O_NOPRINT)) != NULL)
-		for (; *chp != '\0'; ++chp)
-			if (*chp == ch)
-				goto nopr;
+		if (strstr(chp, sp->cname) != NULL)
+			goto nopr;
 
 	/*
 	 * Historical (ARPA standard) mappings.  Printable characters are left
@@ -270,45 +258,55 @@ v_key_name(SCR *sp, ARG_CHAR_T ach)
 	 * told that this is a reasonable assumption...
 	 *
 	 * XXX
-	 * This code will only work with CHAR_T's that are multiples of 8-bit
-	 * bytes.
-	 *
-	 * XXX
-	 * NB: There's an assumption here that all printable characters take
-	 * up a single column on the screen.  This is not always correct.
+	 * The code prints non-printable wide characters in 4 or 5 digits
+	 * Unicode escape sequences, so only supports plane 0 to 15.
 	 */
-	if (ISPRINT(ch)) {
-pr:		sp->cname[0] = ch;
-		len = 1;
+	if (CAN_PRINT(sp, ach))
 		goto done;
-	}
-nopr:	if (ISCNTRL(ch) && (ch < 0x20 || ch == 0x7f)) {
+nopr:	if (iscntrl(ch) && (ch < 0x20 || ch == 0x7f)) {
 		sp->cname[0] = '^';
 		sp->cname[1] = ch == 0x7f ? '?' : '@' + ch;
 		len = 2;
-	} else if (O_ISSET(sp, O_OCTAL)) {
-#define	BITS	(sizeof(CHAR_T) * 8)
-#define	SHIFT	(BITS - BITS % 3)
-#define	TOPMASK	(BITS % 3 == 2 ? 3 : 1) << (BITS - BITS % 3)
+		goto done;
+	}
+#ifdef USE_WIDECHAR
+	if (INTISWIDE(ach)) {
+		int uc = -1;
+
+		if (!strcmp(codeset(), "UTF-8"))
+			uc = decode_utf8(sp->cname);
+#ifdef USE_ICONV
+		else {
+			char buf[sizeof(sp->cname)] = "";
+			size_t left = sizeof(sp->cname);
+			char *in = sp->cname;
+			char *out = buf;
+			iconv(sp->conv.id[IC_IE_TO_UTF16],
+			    (iconv_src_t)&in, &len, &out, &left);
+			iconv(sp->conv.id[IC_IE_TO_UTF16],
+			    NULL, NULL, NULL, NULL);
+			uc = decode_utf16(buf, 1);
+		}
+#endif
+		if (uc >= 0) {
+			len = snprintf(sp->cname, sizeof(sp->cname),
+			    uc < 0x10000 ? "\\u%04x" : "\\U%05X", uc);
+			goto done;
+		}
+	}
+#endif
+	if (O_ISSET(sp, O_OCTAL)) {
 		sp->cname[0] = '\\';
-		sp->cname[1] = octdigit[(ch & TOPMASK) >> SHIFT];
-		shift = SHIFT - 3;
-		for (len = 2, mask = 7 << (SHIFT - 3),
-		    cnt = BITS / 3; cnt-- > 0; mask >>= 3, shift -= 3)
-			sp->cname[len++] = octdigit[(ch & mask) >> shift];
+		sp->cname[1] = octdigit[(ch & 0300) >> 6];
+		sp->cname[2] = octdigit[(ch &  070) >> 3];
+		sp->cname[3] = octdigit[ ch &   07      ];
 	} else {
 		sp->cname[0] = '\\';
 		sp->cname[1] = 'x';
-		for (len = 2, chp = (u_int8_t *)&ch,
-		    /* sizeof(CHAR_T) conflict with MAX_CHARACTER_COLUMNS
-		     * and code depends on big endian
-		     * and might not be needed in the long run
-		     */
-		    cnt = /*sizeof(CHAR_T)*/1; cnt-- > 0; ++chp) {
-			sp->cname[len++] = hexdigit[(*chp & 0xf0) >> 4];
-			sp->cname[len++] = hexdigit[*chp & 0x0f];
-		}
+		sp->cname[2] = hexdigit[(ch & 0xf0) >> 4];
+		sp->cname[3] = hexdigit[ ch & 0x0f      ];
 	}
+	len = 4;
 done:	sp->cname[sp->clen = len] = '\0';
 	return (sp->cname);
 }
@@ -318,9 +316,9 @@ done:	sp->cname[sp->clen = len] = '\0';
  *	Fill in the value for a key.  This routine is the backup
  *	for the KEY_VAL() macro.
  *
- * PUBLIC: int v_key_val __P((SCR *, ARG_CHAR_T));
+ * PUBLIC: e_key_t v_key_val(SCR *, ARG_CHAR_T);
  */
-int
+e_key_t
 v_key_val(SCR *sp, ARG_CHAR_T ch)
 {
 	KEYLIST k, *kp;
@@ -340,28 +338,25 @@ v_key_val(SCR *sp, ARG_CHAR_T ch)
  * an associated flag value, which indicates if it has already been quoted,
  * and if it is the result of a mapping or an abbreviation.
  *
- * PUBLIC: int v_event_push __P((SCR *, EVENT *, CHAR_T *, size_t, u_int));
+ * PUBLIC: int v_event_push(SCR *, EVENT *, CHAR_T *, size_t, u_int);
  */
 int
-v_event_push(SCR *sp, EVENT *p_evp, CHAR_T *p_s, size_t nitems, u_int flags)
-	        
-	             			/* Push event. */
-	            			/* Push characters. */
-	              			/* Number of items to push. */
-	            			/* CH_* flags. */
+v_event_push(SCR *sp,
+	EVENT *p_evp,			/* Push event. */
+	CHAR_T *p_s,			/* Push characters. */
+	size_t nitems,			/* Number of items to push. */
+	u_int flags)			/* CH_* flags. */
 {
 	EVENT *evp;
 	GS *gp;
-	WIN *wp;
 	size_t total;
 
 	/* If we have room, stuff the items into the buffer. */
 	gp = sp->gp;
-	wp = sp->wp;
-	if (nitems <= wp->i_next ||
-	    (wp->i_event != NULL && wp->i_cnt == 0 && nitems <= wp->i_nelem)) {
-		if (wp->i_cnt != 0)
-			wp->i_next -= nitems;
+	if (nitems <= gp->i_next ||
+	    (gp->i_event != NULL && gp->i_cnt == 0 && nitems <= gp->i_nelem)) {
+		if (gp->i_cnt != 0)
+			gp->i_next -= nitems;
 		goto copy;
 	}
 
@@ -371,24 +366,24 @@ v_event_push(SCR *sp, EVENT *p_evp, CHAR_T *p_s, size_t nitems, u_int flags)
 	 * extra.
 	 */
 #define	TERM_PUSH_SHIFT	30
-	total = wp->i_cnt + wp->i_next + nitems + TERM_PUSH_SHIFT;
-	if (total >= wp->i_nelem && v_event_grow(sp, MAX(total, 64)))
+	total = gp->i_cnt + gp->i_next + nitems + TERM_PUSH_SHIFT;
+	if (total >= gp->i_nelem && v_event_grow(sp, MAX(total, 64)))
 		return (1);
-	if (wp->i_cnt)
-		MEMMOVE(wp->i_event + TERM_PUSH_SHIFT + nitems,
-		    wp->i_event + wp->i_next, wp->i_cnt);
-	wp->i_next = TERM_PUSH_SHIFT;
+	if (gp->i_cnt)
+		memmove(gp->i_event + TERM_PUSH_SHIFT + nitems,
+		    gp->i_event + gp->i_next, gp->i_cnt * sizeof(EVENT));
+	gp->i_next = TERM_PUSH_SHIFT;
 
 	/* Put the new items into the queue. */
-copy:	wp->i_cnt += nitems;
-	for (evp = wp->i_event + wp->i_next; nitems--; ++evp) {
+copy:	gp->i_cnt += nitems;
+	for (evp = gp->i_event + gp->i_next; nitems--; ++evp) {
 		if (p_evp != NULL)
 			*evp = *p_evp++;
 		else {
 			evp->e_event = E_CHARACTER;
 			evp->e_c = *p_s++;
 			evp->e_value = KEY_VAL(sp, evp->e_c);
-			FL_INIT(evp->e_flags, flags);
+			F_INIT(&evp->e_ch, flags);
 		}
 	}
 	return (0);
@@ -403,17 +398,17 @@ v_event_append(SCR *sp, EVENT *argp)
 {
 	CHAR_T *s;			/* Characters. */
 	EVENT *evp;
-	WIN *wp;
+	GS *gp;
 	size_t nevents;			/* Number of events. */
 
 	/* Grow the buffer as necessary. */
 	nevents = argp->e_event == E_STRING ? argp->e_len : 1;
-	wp = sp->wp;
-	if (wp->i_event == NULL ||
-	    nevents > wp->i_nelem - (wp->i_next + wp->i_cnt))
+	gp = sp->gp;
+	if (gp->i_event == NULL ||
+	    nevents > gp->i_nelem - (gp->i_next + gp->i_cnt))
 		v_event_grow(sp, MAX(nevents, 64));
-	evp = wp->i_event + wp->i_next + wp->i_cnt;
-	wp->i_cnt += nevents;
+	evp = gp->i_event + gp->i_next + gp->i_cnt;
+	gp->i_cnt += nevents;
 
 	/* Transform strings of characters into single events. */
 	if (argp->e_event == E_STRING)
@@ -430,10 +425,10 @@ v_event_append(SCR *sp, EVENT *argp)
 
 /* Remove events from the queue. */
 #define	QREM(len) {							\
-	if ((wp->i_cnt -= len) == 0)					\
-		wp->i_next = 0;						\
+	if ((gp->i_cnt -= len) == 0)					\
+		gp->i_next = 0;						\
 	else								\
-		wp->i_next += len;					\
+		gp->i_next += len;					\
 }
 
 /*
@@ -521,7 +516,7 @@ v_event_append(SCR *sp, EVENT *argp)
  * point.  Given that this might make the log grow unacceptably (consider that
  * cursor keys are done with maps), for now we leave any changes made in place.
  *
- * PUBLIC: int v_event_get __P((SCR *, EVENT *, int, u_int32_t));
+ * PUBLIC: int v_event_get(SCR *, EVENT *, int, u_int32_t);
  */
 int
 v_event_get(SCR *sp, EVENT *argp, int timeout, u_int32_t flags)
@@ -530,10 +525,8 @@ v_event_get(SCR *sp, EVENT *argp, int timeout, u_int32_t flags)
 	GS *gp;
 	SEQ *qp;
 	int init_nomap, ispartial, istimeout, remap_cnt;
-	WIN *wp;
 
 	gp = sp->gp;
-	wp = sp->wp;
 
 	/* If simply checking for interrupts, argp may be NULL. */
 	if (argp == NULL)
@@ -545,14 +538,14 @@ retry:	istimeout = remap_cnt = 0;
 	 * If the queue isn't empty and we're timing out for characters,
 	 * return immediately.
 	 */
-	if (wp->i_cnt != 0 && LF_ISSET(EC_TIMEOUT))
+	if (gp->i_cnt != 0 && LF_ISSET(EC_TIMEOUT))
 		return (0);
 
 	/*
 	 * If the queue is empty, we're checking for interrupts, or we're
 	 * timing out for characters, get more events.
 	 */
-	if (wp->i_cnt == 0 || LF_ISSET(EC_INTERRUPT | EC_TIMEOUT)) {
+	if (gp->i_cnt == 0 || LF_ISSET(EC_INTERRUPT | EC_TIMEOUT)) {
 		/*
 		 * If we're reading new characters, check any scripting
 		 * windows for input.
@@ -602,7 +595,7 @@ append:			if (v_event_append(sp, argp))
 	if (LF_ISSET(EC_INTERRUPT | EC_TIMEOUT))
 		return (0);
 	 
-newmap:	evp = &wp->i_event[wp->i_next];
+newmap:	evp = &gp->i_event[gp->i_next];
 
 	/* 
 	 * If the next event in the queue isn't a character event, return
@@ -624,13 +617,14 @@ newmap:	evp = &wp->i_event[wp->i_next];
 	 *
 	 * return it forthwith.
 	 */
-	if (istimeout || FL_ISSET(evp->e_flags, CH_NOMAP) ||
+	if (istimeout || F_ISSET(&evp->e_ch, CH_NOMAP) ||
 	    !LF_ISSET(EC_MAPCOMMAND | EC_MAPINPUT) ||
-	    evp->e_c < MAX_BIT_SEQ && !bit_test(gp->seqb, evp->e_c))
+	    ((evp->e_c & ~MAX_BIT_SEQ) == 0 &&
+	    !bit_test(gp->seqb, evp->e_c)))
 		goto nomap;
 
 	/* Search the map. */
-	qp = seq_find(sp, NULL, evp, NULL, wp->i_cnt,
+	qp = seq_find(sp, NULL, evp, NULL, gp->i_cnt,
 	    LF_ISSET(EC_MAPCOMMAND) ? SEQ_COMMAND : SEQ_INPUT, &ispartial);
 
 	/*
@@ -676,12 +670,12 @@ nomap:		if (!ISDIGIT(evp->e_c) && LF_ISSET(EC_MAPNODIGIT))
 not_digit:	argp->e_c = CH_NOT_DIGIT;
 		argp->e_value = K_NOTUSED;
 		argp->e_event = E_CHARACTER;
-		FL_INIT(argp->e_flags, 0);
+		F_INIT(&argp->e_ch, 0);
 		return (0);
 	}
 
 	/* Find out if the initial segments are identical. */
-	init_nomap = !e_memcmp(qp->output, &wp->i_event[wp->i_next], qp->ilen);
+	init_nomap = !e_memcmp(qp->output, &gp->i_event[gp->i_next], qp->ilen);
 
 	/* Delete the mapped characters from the queue. */
 	QREM(qp->ilen);
@@ -720,7 +714,7 @@ not_digit:	argp->e_c = CH_NOT_DIGIT;
 			if (v_event_push(sp, NULL,
 			    qp->output, qp->ilen, CH_NOMAP | CH_MAPPED))
 				return (1);
-			evp = &wp->i_event[wp->i_next];
+			evp = &gp->i_event[gp->i_next];
 			goto nomap;
 		}
 		if (v_event_push(sp, NULL, qp->output, qp->olen, CH_MAPPED))
@@ -743,15 +737,11 @@ static void
 v_sync(SCR *sp, int flags)
 {
 	GS *gp;
-	WIN *wp;
 
 	gp = sp->gp;
-	for (wp = gp->dq.cqh_first; wp != (void *)&gp->dq; 
-	    wp = wp->q.cqe_next)
-		for (sp = wp->scrq.cqh_first; sp != (void *)&wp->scrq; 
-		    sp = sp->q.cqe_next)
+	TAILQ_FOREACH(sp, gp->dq, q)
 		rcv_sync(sp, flags);
-	for (sp = gp->hq.cqh_first; sp != (void *)&gp->hq; sp = sp->q.cqe_next)
+	TAILQ_FOREACH(sp, gp->hq, q)
 		rcv_sync(sp, flags);
 }
 
@@ -759,7 +749,7 @@ v_sync(SCR *sp, int flags)
  * v_event_err --
  *	Unexpected event.
  *
- * PUBLIC: void v_event_err __P((SCR *, EVENT *));
+ * PUBLIC: void v_event_err(SCR *, EVENT *);
  */
 void
 v_event_err(SCR *sp, EVENT *evp)
@@ -774,9 +764,6 @@ v_event_err(SCR *sp, EVENT *evp)
 	case E_INTERRUPT:
 		msgq(sp, M_ERR, "279|Unexpected interrupt event");
 		break;
-	case E_IPCOMMAND:
-		msgq(sp, M_ERR, "318|Unexpected command or input");
-		break;
 	case E_REPAINT:
 		msgq(sp, M_ERR, "281|Unexpected repaint event");
 		break;
@@ -800,22 +787,25 @@ v_event_err(SCR *sp, EVENT *evp)
 	default:
 		abort();
 	}
+
+	/* Free any allocated memory. */
+	free(evp->e_asp);
 }
 
 /*
  * v_event_flush --
  *	Flush any flagged keys, returning if any keys were flushed.
  *
- * PUBLIC: int v_event_flush __P((SCR *, u_int));
+ * PUBLIC: int v_event_flush(SCR *, u_int);
  */
 int
 v_event_flush(SCR *sp, u_int flags)
 {
-	WIN *wp;
+	GS *gp;
 	int rval;
 
-	for (rval = 0, wp = sp->wp; wp->i_cnt != 0 &&
-	    FL_ISSET(wp->i_event[wp->i_next].e_flags, flags); rval = 1)
+	for (rval = 0, gp = sp->gp; gp->i_cnt != 0 &&
+	    F_ISSET(&gp->i_event[gp->i_next].e_ch, flags); rval = 1)
 		QREM(1);
 	return (rval);
 }
@@ -827,14 +817,14 @@ v_event_flush(SCR *sp, u_int flags)
 static int
 v_event_grow(SCR *sp, int add)
 {
-	WIN *wp;
+	GS *gp;
 	size_t new_nelem, olen;
 
-	wp = sp->wp;
-	new_nelem = wp->i_nelem + add;
-	olen = wp->i_nelem * sizeof(wp->i_event[0]);
-	BINC_RET(sp, EVENT, wp->i_event, olen, new_nelem * sizeof(EVENT));
-	wp->i_nelem = olen / sizeof(wp->i_event[0]);
+	gp = sp->gp;
+	new_nelem = gp->i_nelem + add;
+	olen = gp->i_nelem * sizeof(gp->i_event[0]);
+	BINC_RET(sp, EVENT, gp->i_event, olen, new_nelem * sizeof(gp->i_event[0]));
+	gp->i_nelem = olen / sizeof(gp->i_event[0]);
 	return (0);
 }
 
