@@ -9,10 +9,6 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: log.c,v 10.26 2002/03/02 23:12:13 skimo Exp $ (Berkeley) $Date: 2002/03/02 23:12:13 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/stat.h>
@@ -20,7 +16,9 @@ static const char sccsid[] = "$Id: log.c,v 10.26 2002/03/02 23:12:13 skimo Exp $
 #include <bitstring.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <libgen.h>
 #include <limits.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -33,12 +31,11 @@ static const char sccsid[] = "$Id: log.c,v 10.26 2002/03/02 23:12:13 skimo Exp $
  *
  *	LOG_CURSOR_INIT		MARK
  *	LOG_CURSOR_END		MARK
- *	LOG_LINE_APPEND_F 	db_recno_t		char *
- *	LOG_LINE_APPEND_B 	db_recno_t		char *
- *	LOG_LINE_DELETE_F	db_recno_t		char *
- *	LOG_LINE_DELETE_B	db_recno_t		char *
- *	LOG_LINE_RESET_F	db_recno_t		char *
- *	LOG_LINE_RESET_B	db_recno_t		char *
+ *	LOG_LINE_APPEND 	recno_t		char *
+ *	LOG_LINE_DELETE		recno_t		char *
+ *	LOG_LINE_INSERT		recno_t		char *
+ *	LOG_LINE_RESET_F	recno_t		char *
+ *	LOG_LINE_RESET_B	recno_t		char *
  *	LOG_MARK		LMARK
  *
  * We do before image physical logging.  This means that the editor layer
@@ -63,12 +60,13 @@ static const char sccsid[] = "$Id: log.c,v 10.26 2002/03/02 23:12:13 skimo Exp $
  * behaved that way.
  */
 
-static int	vi_log_get __P((SCR *sp, db_recno_t *lnop, size_t *size));
-static int	log_cursor1 __P((SCR *, int));
-static void	log_err __P((SCR *, char *, int));
+static int	log_cursor1(SCR *, int);
+static void	log_err(SCR *, char *, int);
 #if defined(DEBUG) && 0
-static void	log_trace __P((SCR *, char *, db_recno_t, u_char *));
+static void	log_trace(SCR *, char *, recno_t, u_char *);
 #endif
+static int	apply_with(int (*)(SCR *, recno_t, CHAR_T *, size_t),
+					SCR *, recno_t, u_char *, size_t);
 
 /* Try and restart the log on failure, i.e. if we run out of memory. */
 #define	LOG_ERR {							\
@@ -80,8 +78,8 @@ static void	log_trace __P((SCR *, char *, db_recno_t, u_char *));
  * because it is passed to db_set as a string
  */
 typedef struct {
-    char    data[sizeof(u_char) /* type */ + sizeof(db_recno_t)];
-    CHAR_T  str[1];
+	char    data[sizeof(u_char) /* type */ + sizeof(recno_t)];
+	CHAR_T  str[1];
 } log_t;
 #define CHAR_T_OFFSET ((char *)(((log_t*)0)->str) - (char *)0)
 
@@ -89,7 +87,7 @@ typedef struct {
  * log_init --
  *	Initialize the logging subsystem.
  *
- * PUBLIC: int log_init __P((SCR *, EXF *));
+ * PUBLIC: int log_init(SCR *, EXF *);
  */
 int
 log_init(SCR *sp, EXF *ep)
@@ -102,23 +100,20 @@ log_init(SCR *sp, EXF *ep)
 	 * buffers because the global ones are almost by definition
 	 * going to be in use when the log runs.
 	 */
-	sp->wp->l_lp = NULL;
-	sp->wp->l_len = 0;
+	ep->l_lp = NULL;
+	ep->l_len = 0;
 	ep->l_cursor.lno = 1;		/* XXX Any valid recno. */
 	ep->l_cursor.cno = 0;
 	ep->l_high = ep->l_cur = 1;
 
-	if (db_create(&ep->log, 0, 0) != 0 ||
-	    db_open(ep->log, NULL, DB_RECNO,
-			  DB_CREATE | VI_DB_THREAD, S_IRUSR | S_IWUSR) != 0) {
+	ep->log = dbopen(NULL, O_CREAT | O_NONBLOCK | O_RDWR,
+	    S_IRUSR | S_IWUSR, DB_RECNO, NULL);
+	if (ep->log == NULL) {
 		msgq(sp, M_SYSERR, "009|Log file");
 		F_SET(ep, F_NOLOG);
 		return (1);
 	}
 
-	ep->l_win = NULL;
-	/*LOCK_INIT(sp->wp, ep);*/
-
 	return (0);
 }
 
@@ -126,7 +121,7 @@ log_init(SCR *sp, EXF *ep)
  * log_end --
  *	Close the logging subsystem.
  *
- * PUBLIC: int log_end __P((SCR *, EXF *));
+ * PUBLIC: int log_end(SCR *, EXF *);
  */
 int
 log_end(SCR *sp, EXF *ep)
@@ -135,16 +130,13 @@ log_end(SCR *sp, EXF *ep)
 	 * !!!
 	 * ep MAY NOT BE THE SAME AS sp->ep, DON'T USE THE LATTER.
 	 */
-	/*LOCK_END(sp->wp, ep);*/
 	if (ep->log != NULL) {
-		(void)(ep->log->close)(ep->log,DB_NOSYNC);
+		(void)(ep->log->close)(ep->log);
 		ep->log = NULL;
 	}
-	if (sp->wp->l_lp != NULL) {
-		free(sp->wp->l_lp);
-		sp->wp->l_lp = NULL;
-	}
-	sp->wp->l_len = 0;
+	free(ep->l_lp);
+	ep->l_lp = NULL;
+	ep->l_len = 0;
 	ep->l_cursor.lno = 1;		/* XXX Any valid recno. */
 	ep->l_cursor.cno = 0;
 	ep->l_high = ep->l_cur = 1;
@@ -155,7 +147,7 @@ log_end(SCR *sp, EXF *ep)
  * log_cursor --
  *	Log the current cursor position, starting an event.
  *
- * PUBLIC: int log_cursor __P((SCR *));
+ * PUBLIC: int log_cursor(SCR *);
  */
 int
 log_cursor(SCR *sp)
@@ -171,11 +163,8 @@ log_cursor(SCR *sp)
 	 * put out the ending cursor record.
 	 */
 	if (ep->l_cursor.lno == OOBLNO) {
-		if (ep->l_win && ep->l_win != sp->wp)
-			return 0;
 		ep->l_cursor.lno = sp->lno;
 		ep->l_cursor.cno = sp->cno;
-		ep->l_win = NULL;
 		return (log_cursor1(sp, LOG_CURSOR_END));
 	}
 	ep->l_cursor.lno = sp->lno;
@@ -195,37 +184,25 @@ log_cursor1(SCR *sp, int type)
 
 	ep = sp->ep;
 
-	/*
-	if (type == LOG_CURSOR_INIT &&
-	    LOCK_TRY(sp->wp, ep))
-		return 1;
-	*/
+	BINC_RETC(sp, ep->l_lp, ep->l_len, sizeof(u_char) + sizeof(MARK));
+	ep->l_lp[0] = type;
+	memmove(ep->l_lp + sizeof(u_char), &ep->l_cursor, sizeof(MARK));
 
-	BINC_RETC(sp, sp->wp->l_lp, sp->wp->l_len, sizeof(u_char) + sizeof(MARK));
-	sp->wp->l_lp[0] = type;
-	memmove(sp->wp->l_lp + sizeof(u_char), &ep->l_cursor, sizeof(MARK));
-
-	memset(&key, 0, sizeof(key));
 	key.data = &ep->l_cur;
-	key.size = sizeof(db_recno_t);
-	memset(&data, 0, sizeof(data));
-	data.data = sp->wp->l_lp;
+	key.size = sizeof(recno_t);
+	data.data = ep->l_lp;
 	data.size = sizeof(u_char) + sizeof(MARK);
-	if (ep->log->put(ep->log, NULL, &key, &data, 0) == -1)
+	if (ep->log->put(ep->log, &key, &data, 0) == -1)
 		LOG_ERR;
 
 #if defined(DEBUG) && 0
-	vtrace(sp, "%lu: %s: %u/%u\n", ep->l_cur,
+	TRACE(sp, "%lu: %s: %u/%u\n", ep->l_cur,
 	    type == LOG_CURSOR_INIT ? "log_cursor_init" : "log_cursor_end",
 	    sp->lno, sp->cno);
 #endif
 	/* Reset high water mark. */
 	ep->l_high = ++ep->l_cur;
 
-	/*
-	if (type == LOG_CURSOR_END)
-		LOCK_UNLOCK(sp->wp, ep);
-	*/
 	return (0);
 }
 
@@ -233,16 +210,16 @@ log_cursor1(SCR *sp, int type)
  * log_line --
  *	Log a line change.
  *
- * PUBLIC: int log_line __P((SCR *, db_recno_t, u_int));
+ * PUBLIC: int log_line(SCR *, recno_t, u_int);
  */
 int
-log_line(SCR *sp, db_recno_t lno, u_int action)
+log_line(SCR *sp, recno_t lno, u_int action)
 {
 	DBT data, key;
 	EXF *ep;
 	size_t len;
 	CHAR_T *lp;
-	db_recno_t lcur;
+	recno_t lcur;
 
 	ep = sp->ep;
 	if (F_ISSET(ep, F_NOLOG))
@@ -261,17 +238,6 @@ log_line(SCR *sp, db_recno_t lno, u_int action)
 		if (log_cursor1(sp, LOG_CURSOR_INIT))
 			return (1);
 		ep->l_cursor.lno = OOBLNO;
-		ep->l_win = sp->wp;
-	} /*else if (ep->l_win != sp->wp) {
-		printf("log_line own: %p, this: %p\n", ep->l_win, sp->wp);
-		return 1;
-	}*/
-
-	switch (action) {
-	/* newly added for DB4 logging */
-	case LOG_LINE_APPEND_B:
-	case LOG_LINE_DELETE_F:
-		return 0;
 	}
 
 	/*
@@ -282,58 +248,51 @@ log_line(SCR *sp, db_recno_t lno, u_int action)
 	 */
 	if (action == LOG_LINE_RESET_B) {
 		if (db_get(sp, lno, DBG_NOCACHE, &lp, &len)) {
-			static CHAR_T nul = 0;
 			if (lno != 1) {
 				db_err(sp, lno);
 				return (1);
 			}
 			len = 0;
-			lp = &nul;
+			lp = L("");
 		}
 	} else
 		if (db_get(sp, lno, DBG_FATAL, &lp, &len))
 			return (1);
 	BINC_RETC(sp,
-	    sp->wp->l_lp, sp->wp->l_len, 
+	    ep->l_lp, ep->l_len,
 	    len * sizeof(CHAR_T) + CHAR_T_OFFSET);
-	sp->wp->l_lp[0] = action;
-	memmove(sp->wp->l_lp + sizeof(u_char), &lno, sizeof(db_recno_t));
-	MEMMOVEW(sp->wp->l_lp + CHAR_T_OFFSET, lp, len);
+	ep->l_lp[0] = action;
+	memmove(ep->l_lp + sizeof(u_char), &lno, sizeof(recno_t));
+	memmove(ep->l_lp + CHAR_T_OFFSET, lp, len * sizeof(CHAR_T));
 
 	lcur = ep->l_cur;
-	memset(&key, 0, sizeof(key));
 	key.data = &lcur;
-	key.size = sizeof(db_recno_t);
-	memset(&data, 0, sizeof(data));
-	data.data = sp->wp->l_lp;
+	key.size = sizeof(recno_t);
+	data.data = ep->l_lp;
 	data.size = len * sizeof(CHAR_T) + CHAR_T_OFFSET;
-	if (ep->log->put(ep->log, NULL, &key, &data, 0) == -1)
+	if (ep->log->put(ep->log, &key, &data, 0) == -1)
 		LOG_ERR;
 
 #if defined(DEBUG) && 0
 	switch (action) {
-	case LOG_LINE_APPEND_F:
-		vtrace(sp, "%u: log_line: append_f: %lu {%u}\n",
+	case LOG_LINE_APPEND:
+		TRACE(sp, "%lu: log_line: append: %lu {%u}\n",
 		    ep->l_cur, lno, len);
 		break;
-	case LOG_LINE_APPEND_B:
-		vtrace(sp, "%u: log_line: append_b: %lu {%u}\n",
+	case LOG_LINE_DELETE:
+		TRACE(sp, "%lu: log_line: delete: %lu {%u}\n",
 		    ep->l_cur, lno, len);
 		break;
-	case LOG_LINE_DELETE_F:
-		vtrace(sp, "%lu: log_line: delete_f: %lu {%u}\n",
-		    ep->l_cur, lno, len);
-		break;
-	case LOG_LINE_DELETE_B:
-		vtrace(sp, "%lu: log_line: delete_b: %lu {%u}\n",
+	case LOG_LINE_INSERT:
+		TRACE(sp, "%lu: log_line: insert: %lu {%u}\n",
 		    ep->l_cur, lno, len);
 		break;
 	case LOG_LINE_RESET_F:
-		vtrace(sp, "%lu: log_line: reset_f: %lu {%u}\n",
+		TRACE(sp, "%lu: log_line: reset_f: %lu {%u}\n",
 		    ep->l_cur, lno, len);
 		break;
 	case LOG_LINE_RESET_B:
-		vtrace(sp, "%lu: log_line: reset_b: %lu {%u}\n",
+		TRACE(sp, "%lu: log_line: reset_b: %lu {%u}\n",
 		    ep->l_cur, lno, len);
 		break;
 	}
@@ -351,7 +310,7 @@ log_line(SCR *sp, db_recno_t lno, u_int action)
  *	would mean that undo operations would only reset marks, and not
  *	cause any other change.
  *
- * PUBLIC: int log_mark __P((SCR *, LMARK *));
+ * PUBLIC: int log_mark(SCR *, LMARK *);
  */
 int
 log_mark(SCR *sp, LMARK *lmp)
@@ -368,25 +327,22 @@ log_mark(SCR *sp, LMARK *lmp)
 		if (log_cursor1(sp, LOG_CURSOR_INIT))
 			return (1);
 		ep->l_cursor.lno = OOBLNO;
-		ep->l_win = sp->wp;
 	}
 
-	BINC_RETC(sp, sp->wp->l_lp,
-	    sp->wp->l_len, sizeof(u_char) + sizeof(LMARK));
-	sp->wp->l_lp[0] = LOG_MARK;
-	memmove(sp->wp->l_lp + sizeof(u_char), lmp, sizeof(LMARK));
+	BINC_RETC(sp, ep->l_lp,
+	    ep->l_len, sizeof(u_char) + sizeof(LMARK));
+	ep->l_lp[0] = LOG_MARK;
+	memmove(ep->l_lp + sizeof(u_char), lmp, sizeof(LMARK));
 
-	memset(&key, 0, sizeof(key));
 	key.data = &ep->l_cur;
-	key.size = sizeof(db_recno_t);
-	memset(&data, 0, sizeof(data));
-	data.data = sp->wp->l_lp;
+	key.size = sizeof(recno_t);
+	data.data = ep->l_lp;
 	data.size = sizeof(u_char) + sizeof(LMARK);
-	if (ep->log->put(ep->log, NULL, &key, &data, 0) == -1)
+	if (ep->log->put(ep->log, &key, &data, 0) == -1)
 		LOG_ERR;
 
 #if defined(DEBUG) && 0
-	vtrace(sp, "%lu: mark %c: %lu/%u\n",
+	TRACE(sp, "%lu: mark %c: %lu/%u\n",
 	    ep->l_cur, lmp->name, lmp->lno, lmp->cno);
 #endif
 	/* Reset high water mark. */
@@ -394,58 +350,22 @@ log_mark(SCR *sp, LMARK *lmp)
 	return (0);
 }
 
-/*
- * vi_log_get --
- *	Get a line from the log in log buffer.
- */
-static int
-vi_log_get(SCR *sp, db_recno_t *lnop, size_t *size)
-{
-	DBT key, data;
-	size_t nlen;
-	EXF *ep;
-
-	ep = sp->ep;
-
-	nlen = 1024;
-retry:
-	BINC_RETC(sp, sp->wp->l_lp, sp->wp->l_len, nlen);
-
-	memset(&key, 0, sizeof(key));
-	key.data = lnop;		/* Initialize db request. */
-	key.size = sizeof(db_recno_t);
-	memset(&data, 0, sizeof(data));
-	data.data = sp->wp->l_lp;
-	data.ulen = sp->wp->l_len;
-	data.flags = DB_DBT_USERMEM;
-	switch (ep->log->get(ep->log, NULL, &key, &data, 0)) {
-	case ENOMEM:
-		nlen = data.size;
-		goto retry;
-	case 0:
-		*size = data.size;
-		return 0;
-	default:
-		return 1;
-	}
-}
-
 /*
  * Log_backward --
  *	Roll the log backward one operation.
  *
- * PUBLIC: int log_backward __P((SCR *, MARK *));
+ * PUBLIC: int log_backward(SCR *, MARK *);
  */
 int
 log_backward(SCR *sp, MARK *rp)
 {
+	DBT key, data;
 	EXF *ep;
 	LMARK lm;
 	MARK m;
-	db_recno_t lno;
+	recno_t lno;
 	int didop;
 	u_char *p;
-	size_t size;
 
 	ep = sp->ep;
 	if (F_ISSET(ep, F_NOLOG)) {
@@ -459,46 +379,40 @@ log_backward(SCR *sp, MARK *rp)
 		return (1);
 	}
 
-	if (ep->l_win && ep->l_win != sp->wp) {
-		ex_emsg(sp, NULL, EXM_LOCKED);
-		return 1;
-	}
-	ep->l_win = sp->wp;
-		
-
 	F_SET(ep, F_NOLOG);		/* Turn off logging. */
 
+	key.data = &ep->l_cur;		/* Initialize db request. */
+	key.size = sizeof(recno_t);
 	for (didop = 0;;) {
 		--ep->l_cur;
-		if (vi_log_get(sp, &ep->l_cur, &size))
+		if (ep->log->get(ep->log, &key, &data, 0))
 			LOG_ERR;
 #if defined(DEBUG) && 0
 		log_trace(sp, "log_backward", ep->l_cur, data.data);
 #endif
-		switch (*(p = (u_char *)sp->wp->l_lp)) {
+		switch (*(p = (u_char *)data.data)) {
 		case LOG_CURSOR_INIT:
 			if (didop) {
 				memmove(rp, p + sizeof(u_char), sizeof(MARK));
 				F_CLR(ep, F_NOLOG);
-				ep->l_win = NULL;
 				return (0);
 			}
 			break;
 		case LOG_CURSOR_END:
 			break;
-		case LOG_LINE_APPEND_F:
+		case LOG_LINE_APPEND:
+		case LOG_LINE_INSERT:
 			didop = 1;
-			memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
+			memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
 			if (db_delete(sp, lno))
 				goto err;
 			++sp->rptlines[L_DELETED];
 			break;
-		case LOG_LINE_DELETE_B:
+		case LOG_LINE_DELETE:
 			didop = 1;
-			memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-			if (db_insert(sp, lno, 
-			    (CHAR_T *)(p + CHAR_T_OFFSET),
-			    (size - CHAR_T_OFFSET) / sizeof(CHAR_T)))
+			memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
+			if (apply_with(db_insert, sp, lno,
+				p + CHAR_T_OFFSET, data.size - CHAR_T_OFFSET))
 				goto err;
 			++sp->rptlines[L_ADDED];
 			break;
@@ -506,10 +420,9 @@ log_backward(SCR *sp, MARK *rp)
 			break;
 		case LOG_LINE_RESET_B:
 			didop = 1;
-			memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-			if (db_set(sp, lno, 
-			    (CHAR_T *)(p + CHAR_T_OFFSET),
-			    (size - CHAR_T_OFFSET) / sizeof(CHAR_T)))
+			memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
+			if (apply_with(db_set, sp, lno,
+				p + CHAR_T_OFFSET, data.size - CHAR_T_OFFSET))
 				goto err;
 			if (sp->rptlchange != lno) {
 				sp->rptlchange = lno;
@@ -530,7 +443,6 @@ log_backward(SCR *sp, MARK *rp)
 	}
 
 err:	F_CLR(ep, F_NOLOG);
-	ep->l_win = NULL;
 	return (1);
 }
 
@@ -544,17 +456,17 @@ err:	F_CLR(ep, F_NOLOG);
  * then move back on and do a 'U', the line will be restored to the way
  * it was before the original change.
  *
- * PUBLIC: int log_setline __P((SCR *));
+ * PUBLIC: int log_setline(SCR *);
  */
 int
 log_setline(SCR *sp)
 {
+	DBT key, data;
 	EXF *ep;
 	LMARK lm;
 	MARK m;
-	db_recno_t lno;
+	recno_t lno;
 	u_char *p;
-	size_t size;
 
 	ep = sp->ep;
 	if (F_ISSET(ep, F_NOLOG)) {
@@ -566,27 +478,22 @@ log_setline(SCR *sp)
 	if (ep->l_cur == 1)
 		return (1);
 
-	if (ep->l_win && ep->l_win != sp->wp) {
-		ex_emsg(sp, NULL, EXM_LOCKED);
-		return 1;
-	}
-	ep->l_win = sp->wp;
-
 	F_SET(ep, F_NOLOG);		/* Turn off logging. */
 
+	key.data = &ep->l_cur;		/* Initialize db request. */
+	key.size = sizeof(recno_t);
 	for (;;) {
 		--ep->l_cur;
-		if (vi_log_get(sp, &ep->l_cur, &size))
+		if (ep->log->get(ep->log, &key, &data, 0))
 			LOG_ERR;
 #if defined(DEBUG) && 0
 		log_trace(sp, "log_setline", ep->l_cur, data.data);
 #endif
-		switch (*(p = (u_char *)sp->wp->l_lp)) {
+		switch (*(p = (u_char *)data.data)) {
 		case LOG_CURSOR_INIT:
 			memmove(&m, p + sizeof(u_char), sizeof(MARK));
 			if (m.lno != sp->lno || ep->l_cur == 1) {
 				F_CLR(ep, F_NOLOG);
-				ep->l_win = NULL;
 				return (0);
 			}
 			break;
@@ -595,19 +502,19 @@ log_setline(SCR *sp)
 			if (m.lno != sp->lno) {
 				++ep->l_cur;
 				F_CLR(ep, F_NOLOG);
-				ep->l_win = NULL;
 				return (0);
 			}
 			break;
-		case LOG_LINE_APPEND_F:
-		case LOG_LINE_DELETE_B:
+		case LOG_LINE_APPEND:
+		case LOG_LINE_INSERT:
+		case LOG_LINE_DELETE:
 		case LOG_LINE_RESET_F:
 			break;
 		case LOG_LINE_RESET_B:
-			memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
+			memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
 			if (lno == sp->lno &&
-			    db_set(sp, lno, (CHAR_T *)(p + CHAR_T_OFFSET),
-				(size - CHAR_T_OFFSET) / sizeof(CHAR_T)))
+				apply_with(db_set, sp, lno,
+				p + CHAR_T_OFFSET, data.size - CHAR_T_OFFSET))
 				goto err;
 			if (sp->rptlchange != lno) {
 				sp->rptlchange = lno;
@@ -626,7 +533,6 @@ log_setline(SCR *sp)
 	}
 
 err:	F_CLR(ep, F_NOLOG);
-	ep->l_win = NULL;
 	return (1);
 }
 
@@ -634,18 +540,18 @@ err:	F_CLR(ep, F_NOLOG);
  * Log_forward --
  *	Roll the log forward one operation.
  *
- * PUBLIC: int log_forward __P((SCR *, MARK *));
+ * PUBLIC: int log_forward(SCR *, MARK *);
  */
 int
 log_forward(SCR *sp, MARK *rp)
 {
+	DBT key, data;
 	EXF *ep;
 	LMARK lm;
 	MARK m;
-	db_recno_t lno;
+	recno_t lno;
 	int didop;
 	u_char *p;
-	size_t size;
 
 	ep = sp->ep;
 	if (F_ISSET(ep, F_NOLOG)) {
@@ -659,45 +565,40 @@ log_forward(SCR *sp, MARK *rp)
 		return (1);
 	}
 
-	if (ep->l_win && ep->l_win != sp->wp) {
-		ex_emsg(sp, NULL, EXM_LOCKED);
-		return 1;
-	}
-	ep->l_win = sp->wp;
-
 	F_SET(ep, F_NOLOG);		/* Turn off logging. */
 
+	key.data = &ep->l_cur;		/* Initialize db request. */
+	key.size = sizeof(recno_t);
 	for (didop = 0;;) {
 		++ep->l_cur;
-		if (vi_log_get(sp, &ep->l_cur, &size))
+		if (ep->log->get(ep->log, &key, &data, 0))
 			LOG_ERR;
 #if defined(DEBUG) && 0
 		log_trace(sp, "log_forward", ep->l_cur, data.data);
 #endif
-		switch (*(p = (u_char *)sp->wp->l_lp)) {
+		switch (*(p = (u_char *)data.data)) {
 		case LOG_CURSOR_END:
 			if (didop) {
 				++ep->l_cur;
 				memmove(rp, p + sizeof(u_char), sizeof(MARK));
 				F_CLR(ep, F_NOLOG);
-				ep->l_win = NULL;
 				return (0);
 			}
 			break;
 		case LOG_CURSOR_INIT:
 			break;
-		case LOG_LINE_APPEND_F:
+		case LOG_LINE_APPEND:
+		case LOG_LINE_INSERT:
 			didop = 1;
-			memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-			if (db_insert(sp, lno, 
-			    (CHAR_T *)(p + CHAR_T_OFFSET),
-			    (size - CHAR_T_OFFSET) / sizeof(CHAR_T)))
+			memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
+			if (apply_with(db_insert, sp, lno,
+				p + CHAR_T_OFFSET, data.size - CHAR_T_OFFSET))
 				goto err;
 			++sp->rptlines[L_ADDED];
 			break;
-		case LOG_LINE_DELETE_B:
+		case LOG_LINE_DELETE:
 			didop = 1;
-			memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
+			memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
 			if (db_delete(sp, lno))
 				goto err;
 			++sp->rptlines[L_DELETED];
@@ -706,10 +607,9 @@ log_forward(SCR *sp, MARK *rp)
 			break;
 		case LOG_LINE_RESET_F:
 			didop = 1;
-			memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-			if (db_set(sp, lno, 
-			    (CHAR_T *)(p + CHAR_T_OFFSET),
-			    (size - CHAR_T_OFFSET) / sizeof(CHAR_T)))
+			memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
+			if (apply_with(db_set, sp, lno,
+				p + CHAR_T_OFFSET, data.size - CHAR_T_OFFSET))
 				goto err;
 			if (sp->rptlchange != lno) {
 				sp->rptlchange = lno;
@@ -730,7 +630,6 @@ log_forward(SCR *sp, MARK *rp)
 	}
 
 err:	F_CLR(ep, F_NOLOG);
-	ep->l_win = NULL;
 	return (1);
 }
 
@@ -743,61 +642,53 @@ log_err(SCR *sp, char *file, int line)
 {
 	EXF *ep;
 
-	msgq(sp, M_SYSERR, "015|%s/%d: log put error", tail(file), line);
+	msgq(sp, M_SYSERR, "015|%s/%d: log put error", basename(file), line);
 	ep = sp->ep;
-	(void)ep->log->close(ep->log, DB_NOSYNC);
+	(void)ep->log->close(ep->log);
 	if (!log_init(sp, ep))
 		msgq(sp, M_ERR, "267|Log restarted");
 }
 
 #if defined(DEBUG) && 0
 static void
-log_trace(sp, msg, rno, p)
-	SCR *sp;
-	char *msg;
-	db_recno_t rno;
-	u_char *p;
+log_trace(SCR *sp, char *msg, recno_t rno, u_char *p)
 {
 	LMARK lm;
 	MARK m;
-	db_recno_t lno;
+	recno_t lno;
 
 	switch (*p) {
 	case LOG_CURSOR_INIT:
 		memmove(&m, p + sizeof(u_char), sizeof(MARK));
-		vtrace(sp, "%lu: %s:  C_INIT: %u/%u\n", rno, msg, m.lno, m.cno);
+		TRACE(sp, "%lu: %s:  C_INIT: %u/%u\n", rno, msg, m.lno, m.cno);
 		break;
 	case LOG_CURSOR_END:
 		memmove(&m, p + sizeof(u_char), sizeof(MARK));
-		vtrace(sp, "%lu: %s:   C_END: %u/%u\n", rno, msg, m.lno, m.cno);
+		TRACE(sp, "%lu: %s:   C_END: %u/%u\n", rno, msg, m.lno, m.cno);
 		break;
-	case LOG_LINE_APPEND_F:
-		memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-		vtrace(sp, "%lu: %s:  APPEND_F: %lu\n", rno, msg, lno);
+	case LOG_LINE_APPEND:
+		memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
+		TRACE(sp, "%lu: %s:  APPEND: %lu\n", rno, msg, lno);
 		break;
-	case LOG_LINE_APPEND_B:
-		memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-		vtrace(sp, "%lu: %s:  APPEND_B: %lu\n", rno, msg, lno);
+	case LOG_LINE_INSERT:
+		memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
+		TRACE(sp, "%lu: %s:  INSERT: %lu\n", rno, msg, lno);
 		break;
-	case LOG_LINE_DELETE_F:
-		memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-		vtrace(sp, "%lu: %s:  DELETE_F: %lu\n", rno, msg, lno);
-		break;
-	case LOG_LINE_DELETE_B:
-		memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-		vtrace(sp, "%lu: %s:  DELETE_B: %lu\n", rno, msg, lno);
+	case LOG_LINE_DELETE:
+		memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
+		TRACE(sp, "%lu: %s:  DELETE: %lu\n", rno, msg, lno);
 		break;
 	case LOG_LINE_RESET_F:
-		memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-		vtrace(sp, "%lu: %s: RESET_F: %lu\n", rno, msg, lno);
+		memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
+		TRACE(sp, "%lu: %s: RESET_F: %lu\n", rno, msg, lno);
 		break;
 	case LOG_LINE_RESET_B:
-		memmove(&lno, p + sizeof(u_char), sizeof(db_recno_t));
-		vtrace(sp, "%lu: %s: RESET_B: %lu\n", rno, msg, lno);
+		memmove(&lno, p + sizeof(u_char), sizeof(recno_t));
+		TRACE(sp, "%lu: %s: RESET_B: %lu\n", rno, msg, lno);
 		break;
 	case LOG_MARK:
 		memmove(&lm, p + sizeof(u_char), sizeof(LMARK));
-		vtrace(sp,
+		TRACE(sp,
 		    "%lu: %s:    MARK: %u/%u\n", rno, msg, lm.lno, lm.cno);
 		break;
 	default:
@@ -805,3 +696,41 @@ log_trace(sp, msg, rno, p)
 	}
 }
 #endif
+
+/*
+ * apply_with --
+ *	Apply a realigned line from the log db to the file db.
+ */
+static int
+apply_with(int (*db_func)(SCR *, recno_t, CHAR_T *, size_t), SCR *sp,
+    recno_t lno, u_char *p, size_t len)
+{
+#ifdef USE_WIDECHAR
+	typedef unsigned long nword;
+
+	static size_t blen;
+	static nword *bp;
+	nword *lp = (nword *)((uintptr_t)p / sizeof(nword) * sizeof(nword));
+
+	if (lp != (nword *)p) {
+		int offl = ((uintptr_t)p - (uintptr_t)lp) << 3;
+		int offr = (sizeof(nword) << 3) - offl;
+		size_t i, cnt = (len + sizeof(nword) / 2) / sizeof(nword);
+
+		if (len > blen) {
+			blen = p2roundup(MAX(len, 512));
+			REALLOC(sp, bp, nword *, blen);
+			if (bp == NULL)
+				return (1);
+		}
+		for (i = 0; i < cnt; ++i)
+#if BYTE_ORDER == BIG_ENDIAN
+			bp[i] = (lp[i] << offl) ^ (lp[i+1] >> offr);
+#else
+			bp[i] = (lp[i] >> offl) ^ (lp[i+1] << offr);
+#endif
+		p = (u_char *)bp;
+	}
+#endif
+	return db_func(sp, lno, (CHAR_T *)p, len / sizeof(CHAR_T));
+}
