@@ -9,32 +9,22 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: msg.c,v 10.61 2003/07/18 23:17:30 skimo Exp $ (Berkeley) $Date: 2003/07/18 23:17:30 $";
-#endif /* not lint */
-
-#include <sys/param.h>
-#include <sys/types.h>		/* XXX: param.h may not have included types.h */
+#include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/stat.h>
-#include <sys/time.h>
 
 #include <bitstring.h>
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
+#include <locale.h>
+#include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
-#ifdef __STDC__
-#include <stdarg.h>
-#else
-#include <varargs.h>
-#endif
-
 #include "common.h"
 #include "../vi/vi.h"
 
@@ -42,18 +32,10 @@ static const char sccsid[] = "$Id: msg.c,v 10.61 2003/07/18 23:17:30 skimo Exp $
  * msgq --
  *	Display a message.
  *
- * PUBLIC: void msgq __P((SCR *, mtype_t, const char *, ...));
+ * PUBLIC: void msgq(SCR *, mtype_t, const char *, ...);
  */
 void
-#ifdef __STDC__
 msgq(SCR *sp, mtype_t mt, const char *fmt, ...)
-#else
-msgq(sp, mt, fmt, va_alist)
-	SCR *sp;
-	mtype_t mt;
-        const char *fmt;
-        va_dcl
-#endif
 {
 #ifndef NL_ARGMAX
 #define	__NL_ARGMAX	20		/* Set to 9 by System V. */
@@ -66,13 +48,17 @@ msgq(sp, mt, fmt, va_alist)
 	} str[__NL_ARGMAX];
 #endif
 	static int reenter;		/* STATIC: Re-entrancy check. */
-	CHAR_T ch;
 	GS *gp;
-	WIN *wp;
-	size_t blen, cnt1, cnt2, len, mlen, nlen, soff;
-	const char *p, *t, *u;
-	char *bp, *mp, *rbp, *s_rbp;
-        va_list ap;
+	size_t blen, len, mlen, nlen;
+	const char *p;
+	char *bp, *mp;
+	va_list ap;
+#ifndef NL_ARGMAX
+	int ch;
+	char *rbp, *s_rbp;
+	const char *t, *u;
+	size_t cnt1, cnt2, soff;
+#endif
 
 	/*
 	 * !!!
@@ -88,7 +74,6 @@ msgq(sp, mt, fmt, va_alist)
 			mt = M_INFO;
 	} else {
 		gp = sp->gp;
-		wp = sp->wp;
 		switch (mt) {
 		case M_BERR:
 			if (F_ISSET(sp, SC_VI) && !O_ISSET(sp, O_VERBOSE)) {
@@ -108,7 +93,6 @@ msgq(sp, mt, fmt, va_alist)
 			break;
 		case M_ERR:
 		case M_SYSERR:
-		case M_DBERR:
 			break;
 		default:
 			abort();
@@ -145,7 +129,7 @@ retry:		FREE_SPACE(sp, bp, blen);
 #define	REM	(blen - mlen)
 	mp = bp;
 	mlen = 0;
-	if (mt == M_SYSERR || mt == M_DBERR) {
+	if (mt == M_SYSERR) {
 		p = msg_cat(sp, "020|Error: ", &len);
 		if (REM < len)
 			goto retry;
@@ -159,14 +143,18 @@ retry:		FREE_SPACE(sp, bp, blen);
 	 * the file name and line number prefix.
 	 */
 	if ((mt == M_ERR || mt == M_SYSERR) &&
-	    sp != NULL && wp != NULL && wp->if_name != NULL) {
-		for (p = wp->if_name; *p != '\0'; ++p) {
-			len = snprintf(mp, REM, "%s", KEY_NAME(sp, *p));
+	    sp != NULL && gp != NULL && gp->if_name != NULL) {
+		CHAR_T *wp;
+		size_t wlen;
+
+		CHAR2INT(sp, gp->if_name, strlen(gp->if_name) + 1, wp, wlen);
+		for (; *wp != '\0'; ++wp) {
+			len = snprintf(mp, REM, "%s", KEY_NAME(sp, *wp));
 			mp += len;
 			if ((mlen += len) > blen)
 				goto retry;
 		}
-		len = snprintf(mp, REM, ", %d: ", wp->if_lno);
+		len = snprintf(mp, REM, ", %d: ", gp->if_lno);
 		mp += len;
 		if ((mlen += len) > blen)
 			goto retry;
@@ -206,12 +194,12 @@ retry:		FREE_SPACE(sp, bp, blen);
 		if (*p == '\0')
 			break;
 		++p;
-		if (!isdigit(*p)) {
+		if (!isdigit((u_char)*p)) {
 			if (*p == '%')
 				++p;
 			continue;
 		}
-		for (u = p; *++p != '\0' && isdigit(*p););
+		for (u = p; *++p != '\0' && isdigit((u_char)*p););
 		if (*p != '$')
 			continue;
 
@@ -275,12 +263,10 @@ retry:		FREE_SPACE(sp, bp, blen);
 	fmt = rbp;
 #endif
 
+#ifndef NL_ARGMAX
 format:	/* Format the arguments into the string. */
-#ifdef __STDC__
-        va_start(ap, fmt);
-#else
-        va_start(ap);
 #endif
+	va_start(ap, fmt);
 	len = vsnprintf(mp, REM, fmt, ap);
 	va_end(ap);
 	if (len >= nlen)
@@ -336,46 +322,42 @@ nofmt:	mp += len;
 			goto retry;
 		mt = M_ERR;
 	}
-	if (mt == M_DBERR) {
-		len = snprintf(mp, REM, ": %s", db_strerror(sp->db_error));
-		mp += len;
-		if ((mlen += len) > blen)
-			goto retry;
-		mt = M_ERR;
-	}
 
 	/* Add trailing newline. */
 	if ((mlen += 1) > blen)
 		goto retry;
 	*mp = '\n';
 
-	if (sp != NULL && sp->ep != NULL)
+	if (sp != NULL)
 		(void)ex_fflush(sp);
-	if (wp != NULL)
-		wp->scr_msg(sp, mt, bp, mlen);
+	if (gp != NULL)
+		gp->scr_msg(sp, mt, bp, mlen);
 	else
 		(void)fprintf(stderr, "%.*s", (int)mlen, bp);
 
 	/* Cleanup. */
-ret:	FREE_SPACE(sp, bp, blen);
+#ifndef NL_ARGMAX
+ret:
+#endif
+	FREE_SPACE(sp, bp, blen);
 alloc_err:
 	reenter = 0;
 }
 
 /*
- * msgq_str --
+ * msgq_wstr --
  *	Display a message with an embedded string.
  *
- * PUBLIC: void msgq_wstr __P((SCR *, mtype_t, CHAR_T *, char *));
+ * PUBLIC: void msgq_wstr(SCR *, mtype_t, const CHAR_T *, const char *);
  */
 void
-msgq_wstr(SCR *sp, mtype_t mtype, CHAR_T *str, char *fmt)
+msgq_wstr(SCR *sp, mtype_t mtype, const CHAR_T *str, const char *fmt)
 {
 	size_t nlen;
-	char *nstr;
+	CONST char *nstr;
 
 	if (str == NULL) {
-		msgq(sp, mtype, fmt);
+		msgq(sp, mtype, "%s", fmt);
 		return;
 	}
 	INT2CHAR(sp, str, STRLEN(str) + 1, nstr, nlen);
@@ -386,16 +368,16 @@ msgq_wstr(SCR *sp, mtype_t mtype, CHAR_T *str, char *fmt)
  * msgq_str --
  *	Display a message with an embedded string.
  *
- * PUBLIC: void msgq_str __P((SCR *, mtype_t, char *, char *));
+ * PUBLIC: void msgq_str(SCR *, mtype_t, const char *, const char *);
  */
 void
-msgq_str(SCR *sp, mtype_t mtype, char *str, char *fmt)
+msgq_str(SCR *sp, mtype_t mtype, const char *str, const char *fmt)
 {
 	int nf, sv_errno;
 	char *p;
 
 	if (str == NULL) {
-		msgq(sp, mtype, fmt);
+		msgq(sp, mtype, "%s", fmt);
 		return;
 	}
 
@@ -425,7 +407,7 @@ msgq_str(SCR *sp, mtype_t mtype, char *str, char *fmt)
  * the command 2d}, from the 'b' would report that two lines were deleted,
  * not one.
  *
- * PUBLIC: void mod_rpt __P((SCR *));
+ * PUBLIC: void mod_rpt(SCR *);
  */
 void
 mod_rpt(SCR *sp)
@@ -443,7 +425,7 @@ mod_rpt(SCR *sp)
 		"300|line",
 		"301|lines",
 	};
-	db_recno_t total;
+	recno_t total;
 	u_long rptval;
 	int first, cnt;
 	size_t blen, len, tlen;
@@ -498,8 +480,8 @@ mod_rpt(SCR *sp)
 				*p++ = ' ';
 				tlen += 2;
 			}
-			len = snprintf(p, MAXNUM, "%lu ", 
-				(unsigned long) sp->rptlines[cnt]);
+			len = snprintf(p, MAXNUM, "%lu ",
+			    (u_long)sp->rptlines[cnt]);
 			p += len;
 			tlen += len;
 			t = msg_cat(sp,
@@ -521,7 +503,7 @@ mod_rpt(SCR *sp)
 	++tlen;
 
 	(void)ex_fflush(sp);
-	sp->wp->scr_msg(sp, M_INFO, bp, tlen);
+	sp->gp->scr_msg(sp, M_INFO, bp, tlen);
 
 	FREE_SPACE(sp, bp, blen);
 alloc_err:
@@ -535,26 +517,32 @@ alloc_err:
  * msgq_status --
  *	Report on the file's status.
  *
- * PUBLIC: void msgq_status __P((SCR *, db_recno_t, u_int));
+ * PUBLIC: void msgq_status(SCR *, recno_t, u_int);
  */
 void
-msgq_status(SCR *sp, db_recno_t lno, u_int flags)
+msgq_status(SCR *sp, recno_t lno, u_int flags)
 {
-	db_recno_t last;
+	recno_t last;
 	size_t blen, len;
 	int cnt, needsep;
 	const char *t;
-	char **ap, *bp, *np, *p, *s;
+	char **ap, *bp, *np, *p, *s, *ep;
+	CHAR_T *wp;
+	size_t wlen;
 
 	/* Get sufficient memory. */
 	len = strlen(sp->frp->name);
 	GET_SPACE_GOTOC(sp, bp, blen, len * MAX_CHARACTER_COLUMNS + 128);
 	p = bp;
+	ep = bp + blen;
+
+	/* Convert the filename. */
+	CHAR2INT(sp, sp->frp->name, len + 1, wp, wlen);
 
 	/* Copy in the filename. */
-	for (p = bp, t = sp->frp->name; *t != '\0'; ++t) {
-		len = KEY_LEN(sp, *t);
-		memcpy(p, KEY_NAME(sp, *t), len);
+	for (; *wp != '\0'; ++wp) {
+		len = KEY_LEN(sp, *wp);
+		memcpy(p, KEY_NAME(sp, *wp), len);
 		p += len;
 	}
 	np = p;
@@ -565,7 +553,7 @@ msgq_status(SCR *sp, db_recno_t lno, u_int flags)
 	if (F_ISSET(sp, SC_STATUS_CNT) && sp->argv != NULL) {
 		for (cnt = 0, ap = sp->argv; *ap != NULL; ++ap, ++cnt);
 		if (cnt > 1) {
-			(void)sprintf(p,
+			(void)snprintf(p, ep - p,
 			    msg_cat(sp, "317|%d files to edit", NULL), cnt);
 			p += strlen(p);
 			*p++ = ':';
@@ -640,16 +628,17 @@ msgq_status(SCR *sp, db_recno_t lno, u_int flags)
 			p += len;
 		} else {
 			t = msg_cat(sp, "027|line %lu of %lu [%ld%%]", &len);
-			(void)sprintf(p, t, lno, last, (lno * 100) / last);
+			(void)snprintf(p, ep - p, t, (u_long)lno, (u_long)last,
+			    ((u_long)lno * 100) / last);
 			p += strlen(p);
 		}
 	} else {
 		t = msg_cat(sp, "029|line %lu", &len);
-		(void)sprintf(p, t, lno);
+		(void)snprintf(p, ep - p, t, (u_long)lno);
 		p += strlen(p);
 	}
 #ifdef DEBUG
-	(void)sprintf(p, " (pid %lu)", (u_long)getpid());
+	(void)snprintf(p, ep - p, " (pid %lu)", (u_long)getpid());
 	p += strlen(p);
 #endif
 	*p++ = '\n';
@@ -686,7 +675,7 @@ msgq_status(SCR *sp, db_recno_t lno, u_int flags)
 	/* Flush any waiting ex messages. */
 	(void)ex_fflush(sp);
 
-	sp->wp->scr_msg(sp, M_INFO, s, len);
+	sp->gp->scr_msg(sp, M_INFO, s, len);
 
 	FREE_SPACE(sp, bp, blen);
 alloc_err:
@@ -697,7 +686,7 @@ alloc_err:
  * msg_open --
  *	Open the message catalogs.
  *
- * PUBLIC: int msg_open __P((SCR *, char *));
+ * PUBLIC: int msg_open(SCR *, char *);
  */
 int
 msg_open(SCR *sp, char *file)
@@ -712,79 +701,70 @@ msg_open(SCR *sp, char *file)
 	 * message will be repeated every time nvi is started up.
 	 */
 	static int first = 1;
-	DB *db;
-	DBT data, key;
-	db_recno_t msgno;
-	char *p, *t, buf[MAXPATHLEN];
+	nl_catd catd;
+	char *p;
+	int rval = 0;
 
-	if ((p = strrchr(file, '/')) != NULL && p[1] == '\0' &&
-	    (((t = getenv("LC_MESSAGES")) != NULL && t[0] != '\0') ||
-	    ((t = getenv("LANG")) != NULL && t[0] != '\0'))) {
-		(void)snprintf(buf, sizeof(buf), "%s%s", file, t);
-		p = buf;
-	} else
-		p = file;
-	if ((sp->db_error = db_create(&db, 0, 0)) != 0 ||
-	    (sp->db_error = db->set_re_source(db, p)) != 0 ||
-	    (sp->db_error = db_open(db, NULL, DB_RECNO, 0, 0)) != 0) {
-		if (first) {
-			first = 0;
+	if ((p = strrchr(file, '/')) != NULL && p[1] == '\0') {
+		/* Confirms to XPG4. */
+		if ((p = join(file, setlocale(LC_MESSAGES, NULL))) == NULL) {
+			msgq(sp, M_SYSERR, NULL);
+			return (1);
+		}
+	} else {
+		/* Make sure it's recognized as a path by catopen(3). */
+		if ((p = join(".", file)) == NULL) {
+			msgq(sp, M_SYSERR, NULL);
 			return (1);
 		}
-		msgq_str(sp, M_DBERR, p, "%s");
-		return (1);
 	}
-
-	/*
-	 * Test record 1 for the magic string.  The msgq call is here so
-	 * the message catalog build finds it.
-	 */
-#define	VMC	"VI_MESSAGE_CATALOG"
-	memset(&key, 0, sizeof(key));
-	key.data = &msgno;
-	key.size = sizeof(db_recno_t);
-	memset(&data, 0, sizeof(data));
-	msgno = 1;
-	if ((sp->db_error = db->get(db, NULL, &key, &data, 0)) != 0 ||
-	    data.size != sizeof(VMC) - 1 ||
-	    memcmp(data.data, VMC, sizeof(VMC) - 1)) {
-		(void)db->close(db, DB_NOSYNC);
+	errno = 0;
+	if ((catd = catopen(p, NL_CAT_LOCALE)) == (nl_catd)-1) {
 		if (first) {
 			first = 0;
-			return (1);
+			rval = 1;
+			goto ret;
 		}
-		msgq_str(sp, M_DBERR, p,
-		    "030|The file %s is not a message catalog");
-		return (1);
+
+		/*
+		 * POSIX.1-2008 gives no instruction on how to report a
+		 * corrupt catalog file.  Errno == 0 is not rare; add
+		 * EFTYPE, which is seen on FreeBSD, for a good measure.
+		 */
+		if (errno == 0 || errno == EFTYPE)
+			msgq_str(sp, M_ERR, p,
+			    "030|The file %s is not a message catalog");
+		else
+			msgq_str(sp, M_SYSERR, p, "%s");
+		rval = 1;
+		goto ret;
 	}
 	first = 0;
 
-	if (sp->gp->msg != NULL)
-		(void)sp->gp->msg->close(sp->gp->msg, DB_NOSYNC);
-	sp->gp->msg = db;
-	return (0);
+	msg_close(sp->gp);
+	sp->gp->catd = catd;
+ret:	free(p);
+	return (rval);
 }
 
 /*
  * msg_close --
  *	Close the message catalogs.
  *
- * PUBLIC: void msg_close __P((GS *));
+ * PUBLIC: void msg_close(GS *);
  */
 void
 msg_close(GS *gp)
 {
-	if (gp->msg != NULL) {
-		(void)gp->msg->close(gp->msg, 0);
-		gp->msg = NULL;
-	}
+	if (gp->catd != (nl_catd)-1)
+		(void)catclose(gp->catd);
 }
 
 /*
  * msg_cont --
  *	Return common continuation messages.
  *
- * PUBLIC: const char *msg_cmsg __P((SCR *, cmsg_t, size_t *));
+ * PUBLIC: const char *msg_cmsg(SCR *, cmsg_t, size_t *);
  */
 const char *
 msg_cmsg(SCR *sp, cmsg_t which, size_t *lenp)
@@ -819,45 +799,32 @@ msg_cmsg(SCR *sp, cmsg_t which, size_t *lenp)
  * Only a single catalog message can be accessed at a time, if multiple
  * ones are needed, they must be copied into local memory.
  *
- * PUBLIC: const char *msg_cat __P((SCR *, const char *, size_t *));
+ * PUBLIC: const char *msg_cat(SCR *, const char *, size_t *);
  */
 const char *
 msg_cat(SCR *sp, const char *str, size_t *lenp)
 {
 	GS *gp;
-	DBT data, key;
-	db_recno_t msgno;
+	char *p;
+	int msgno;
 
 	/*
 	 * If it's not a catalog message, i.e. has doesn't have a leading
 	 * number and '|' symbol, we're done.
 	 */
-	if (isdigit(str[0]) &&
-	    isdigit(str[1]) && isdigit(str[2]) && str[3] == '|') {
-		memset(&key, 0, sizeof(key));
-		key.data = &msgno;
-		key.size = sizeof(db_recno_t);
-		memset(&data, 0, sizeof(data));
+	if (isdigit((u_char)str[0]) &&
+	    isdigit((u_char)str[1]) &&
+	    isdigit((u_char)str[2]) && str[3] == '|') {
 		msgno = atoi(str);
-
-		/*
-		 * XXX
-		 * Really sleazy hack -- we put an extra character on the
-		 * end of the format string, and then we change it to be
-		 * the nul termination of the string.  There ought to be
-		 * a better way.  Once we can allocate multiple temporary
-		 * memory buffers, maybe we can use one of them instead.
-		 */
-		gp = sp == NULL ? NULL : sp->gp;
-		if (gp != NULL && gp->msg != NULL &&
-		    gp->msg->get(gp->msg, NULL, &key, &data, 0) == 0 &&
-		    data.size != 0) {
-			if (lenp != NULL)
-				*lenp = data.size - 1;
-			((char *)data.data)[data.size - 1] = '\0';
-			return (data.data);
-		}
 		str = &str[4];
+
+		gp = sp == NULL ? NULL : sp->gp;
+		if (gp != NULL && gp->catd != (nl_catd)-1 &&
+		    (p = catgets(gp->catd, 1, msgno, str)) != NULL) {
+			if (lenp != NULL)
+				*lenp = strlen(p);
+			return (p);
+		}
 	}
 	if (lenp != NULL)
 		*lenp = strlen(str);
@@ -868,19 +835,22 @@ msg_cat(SCR *sp, const char *str, size_t *lenp)
  * msg_print --
  *	Return a printable version of a string, in allocated memory.
  *
- * PUBLIC: char *msg_print __P((SCR *, const char *, int *));
+ * PUBLIC: char *msg_print(SCR *, const char *, int *);
  */
 char *
 msg_print(SCR *sp, const char *s, int *needfree)
 {
 	size_t blen, nlen;
-	const char *cp;
 	char *bp, *ep, *p, *t;
+	CHAR_T *wp, *cp;
+	size_t wlen;
 
 	*needfree = 0;
 
-	for (cp = s; *cp != '\0'; ++cp)
-		if (!isprint(*cp))
+	/* XXX Not good for debugging ex_read & ex_filter.*/
+	CHAR2INT5(sp, EXP(sp)->ibcw, (char *)s, strlen(s) + 1, wp, wlen);
+	for (cp = wp; *cp != '\0'; ++cp)
+		if (!ISPRINT(*cp))
 			break;
 	if (*cp == '\0')
 		return ((char *)s);	/* SAFE: needfree set to 0. */
@@ -904,8 +874,8 @@ alloc_err:	return ("");
 	}
 	*needfree = 1;
 
-	for (p = bp, ep = (bp + blen) - 1, cp = s; *cp != '\0' && p < ep; ++cp)
-		for (t = KEY_NAME(sp, *cp); *t != '\0' && p < ep; *p++ = *t++);
+	for (p = bp, ep = (bp + blen) - 1; *wp != '\0' && p < ep; ++wp)
+		for (t = KEY_NAME(sp, *wp); *t != '\0' && p < ep; *p++ = *t++);
 	if (p == ep)
 		goto retry;
 	*p = '\0';
