@@ -26,20 +26,20 @@
  * put --
  *	Put text buffer contents into the file.
  *
- * PUBLIC: int put(SCR *, CB *, CHAR_T *, MARK *, MARK *, int, int);
+ * PUBLIC: int put(SCR *, CB *, CHAR_T *, MARK *, MARK *, int);
  */
 int
-put(SCR *sp, CB *cbp, CHAR_T *namep, MARK *cp, MARK *rp, int append, int cnt)
+put(SCR *sp, CB *cbp, CHAR_T *namep, MARK *cp, MARK *rp, int append)
 {
 	CHAR_T name;
 	TEXT *ltp, *tp;
 	recno_t lno;
 	size_t blen, clen, len;
-	int rval, i, isempty;
+	int rval;
 	CHAR_T *bp, *t;
 	CHAR_T *p;
 
-	if (cbp == NULL) {
+	if (cbp == NULL)
 		if (namep == NULL) {
 			cbp = sp->gp->dcbp;
 			if (cbp == NULL) {
@@ -56,7 +56,6 @@ put(SCR *sp, CB *cbp, CHAR_T *namep, MARK *cp, MARK *rp, int append, int cnt)
 				return (1);
 			}
 		}
-	}
 	tp = TAILQ_FIRST(cbp->textq);
 
 	/*
@@ -77,16 +76,11 @@ put(SCR *sp, CB *cbp, CHAR_T *namep, MARK *cp, MARK *rp, int append, int cnt)
 	if (cp->lno == 1) {
 		if (db_last(sp, &lno))
 			return (1);
-		if (lno == 0 && F_ISSET(cbp, CB_LMODE)) {
-			for (i = cnt; i > 0; i--) {
-				for (; tp != NULL;
-				    ++lno, ++sp->rptlines[L_ADDED],
-				    tp = TAILQ_NEXT(tp, q))
-					if (db_append(sp, 1, lno, tp->lb,
-					    tp->len))
-						return (1);
-				tp = TAILQ_FIRST(cbp->textq);
-			}
+		if (lno == 0) {
+			for (; tp != NULL;
+			    ++lno, ++sp->rptlines[L_ADDED], tp = TAILQ_NEXT(tp, q))
+				if (db_append(sp, 1, lno, tp->lb, tp->len))
+					return (1);
 			rp->lno = 1;
 			rp->cno = 0;
 			return (0);
@@ -97,14 +91,10 @@ put(SCR *sp, CB *cbp, CHAR_T *namep, MARK *cp, MARK *rp, int append, int cnt)
 	if (F_ISSET(cbp, CB_LMODE)) {
 		lno = append ? cp->lno : cp->lno - 1;
 		rp->lno = lno + 1;
-		for (i = cnt; i > 0; i--) {
-			for (; tp != NULL;
-			    ++lno, ++sp->rptlines[L_ADDED],
-			    tp = TAILQ_NEXT(tp, q))
-				if (db_append(sp, 1, lno, tp->lb, tp->len))
-					return (1);
-			tp = TAILQ_FIRST(cbp->textq);
-		}
+		for (; tp != NULL;
+		    ++lno, ++sp->rptlines[L_ADDED], tp = TAILQ_NEXT(tp, q))
+			if (db_append(sp, 1, lno, tp->lb, tp->len))
+				return (1);
 		rp->cno = 0;
 		(void)nonblank(sp, rp->lno, &rp->cno);
 		return (0);
@@ -120,11 +110,8 @@ put(SCR *sp, CB *cbp, CHAR_T *namep, MARK *cp, MARK *rp, int append, int cnt)
 	 * Get the first line.
 	 */
 	lno = cp->lno;
-	if (db_eget(sp, lno, &p, &len, &isempty)) {
-		if (!isempty)
-			return (1);
-		len = 0;
-	}
+	if (db_get(sp, lno, DBG_FATAL, &p, &len))
+		return (1);
 
 	GET_SPACE_RETW(sp, bp, blen, tp->len + len + 1);
 	t = bp;
@@ -138,10 +125,8 @@ put(SCR *sp, CB *cbp, CHAR_T *namep, MARK *cp, MARK *rp, int append, int cnt)
 
 	/* First line from the CB. */
 	if (tp->len != 0) {
-		for (i = cnt; i > 0; i--) {
-			MEMCPY(t, tp->lb, tp->len);
-			t += tp->len;
-		}
+		MEMCPY(t, tp->lb, tp->len);
+		t += tp->len;
 	}
 
 	/* Calculate length left in the original line. */
