@@ -9,12 +9,7 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: recover.c,v 10.31 2001/11/01 15:24:44 skimo Exp $ (Berkeley) $Date: 2001/11/01 15:24:44 $";
-#endif /* not lint */
-
-#include <sys/param.h>
-#include <sys/types.h>		/* XXX: param.h may not have included types.h */
+#include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/stat.h>
 
@@ -31,12 +26,15 @@ static const char sccsid[] = "$Id: recover.c,v 10.31 2001/11/01 15:24:44 skimo E
 #include <fcntl.h>
 #include <limits.h>
 #include <pwd.h>
+#include <netinet/in.h>		/* Required by resolv.h. */
+#include <resolv.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
 
+#include "../ex/version.h"
 #include "common.h"
 #include "pathnames.h"
 
@@ -57,12 +55,7 @@ static const char sccsid[] = "$Id: recover.c,v 10.31 2001/11/01 15:24:44 skimo E
  *		file exists, and is exclusively locked.
  *
  * In the EXF structure we maintain a file descriptor that is the locked
- * file descriptor for the mail recovery file.  NOTE: we sometimes have to
- * do locking with fcntl(2).  This is a problem because if you close(2) any
- * file descriptor associated with the file, ALL of the locks go away.  Be
- * sure to remember that if you have to modify the recovery code.  (It has
- * been rhetorically asked of what the designers could have been thinking
- * when they did that interface.  The answer is simple: they weren't.)
+ * file descriptor for the mail recovery file.
  *
  * To find out if a recovery file/backing file pair are in use, try to get
  * a lock on the recovery file.
@@ -93,40 +86,37 @@ static const char sccsid[] = "$Id: recover.c,v 10.31 2001/11/01 15:24:44 skimo E
  * means that the data structures (SCR, EXF, the underlying tree structures)
  * must be consistent when the signal arrives.
  *
- * The recovery mail file contains normal mail headers, with two additions,
- * which occur in THIS order, as the FIRST TWO headers:
+ * The recovery mail file contains normal mail headers, with two additional
  *
- *	X-vi-recover-file: file_name
- *	X-vi-recover-path: recover_path
+ *	X-vi-data: <file|path>;<base64 encoded path>
  *
- * Since newlines delimit the headers, this means that file names cannot have
- * newlines in them, but that's probably okay.  As these files aren't intended
- * to be long-lived, changing their format won't be too painful.
+ * MIME headers; the folding character is limited to ' '.
  *
- * Btree files are named "vi.XXXX" and recovery files are named "recover.XXXX".
+ * Btree files are named "vi.XXXXXX" and recovery files are named
+ * "recover.XXXXXX".
  */
 
-#define	VI_FHEADER	"X-vi-recover-file: "
-#define	VI_PHEADER	"X-vi-recover-path: "
+#define	VI_DHEADER	"X-vi-data:"
 
-static int	 rcv_copy __P((SCR *, int, char *));
-static void	 rcv_email __P((SCR *, char *));
-static char	*rcv_gets __P((char *, size_t, int));
-static int	 rcv_mailfile __P((SCR *, int, char *));
-static int	 rcv_mktemp __P((SCR *, char *, char *, int));
+static int	 rcv_copy(SCR *, int, char *);
+static void	 rcv_email(SCR *, char *);
+static int	 rcv_mailfile(SCR *, int, char *);
+static int	 rcv_mktemp(SCR *, char *, char *);
+static int	 rcv_dlnwrite(SCR *, const char *, const char *, FILE *);
+static int	 rcv_dlnread(SCR *, char **, char **, FILE *);
 
 /*
  * rcv_tmp --
  *	Build a file name that will be used as the recovery file.
  *
- * PUBLIC: int rcv_tmp __P((SCR *, EXF *, char *));
+ * PUBLIC: int rcv_tmp(SCR *, EXF *, char *);
  */
 int
 rcv_tmp(SCR *sp, EXF *ep, char *name)
 {
 	struct stat sb;
 	int fd;
-	char *dp, path[MAXPATHLEN];
+	char *dp, *path;
 
 	/*
 	 * !!!
@@ -150,21 +140,17 @@ rcv_tmp(SCR *sp, EXF *ep, char *name)
 		(void)chmod(dp, S_IRWXU | S_IRWXG | S_IRWXO | S_ISVTX);
 	}
 
-	/* Newlines delimit the mail messages. */
-	if (strchr(name, '\n')) {
-		msgq(sp, M_ERR,
-		    "055|Files with newlines in the name are unrecoverable");
+	if ((path = join(dp, "vi.XXXXXX")) == NULL)
+		goto err;
+	if ((fd = rcv_mktemp(sp, path, dp)) == -1) {
+		free(path);
 		goto err;
 	}
-
-	(void)snprintf(path, sizeof(path), "%s/vi.XXXXXX", dp);
-	if ((fd = rcv_mktemp(sp, path, dp, S_IRWXU)) == -1)
-		goto err;
+	(void)fchmod(fd, S_IRWXU);
 	(void)close(fd);
 
-	if ((ep->rcv_path = strdup(path)) == NULL) {
-		msgq(sp, M_SYSERR, NULL);
-		(void)unlink(path);
+	ep->rcv_path = path;
+	if (0) {
 err:		msgq(sp, M_ERR,
 		    "056|Modifications not recoverable if the session fails");
 		return (1);
@@ -179,13 +165,13 @@ err:		msgq(sp, M_ERR,
  * rcv_init --
  *	Force the file to be snapshotted for recovery.
  *
- * PUBLIC: int rcv_init __P((SCR *));
+ * PUBLIC: int rcv_init(SCR *);
  */
 int
 rcv_init(SCR *sp)
 {
 	EXF *ep;
-	db_recno_t lno;
+	recno_t lno;
 
 	ep = sp->ep;
 
@@ -212,7 +198,7 @@ rcv_init(SCR *sp)
 		/* Turn on a busy message, and sync it to backing store. */
 		sp->gp->scr_busy(sp,
 		    "057|Copying file for recovery...", BUSY_ON);
-		if (ep->db->sync(ep->db, 0)) {
+		if (ep->db->sync(ep->db, R_RECNOSYNC)) {
 			msgq_str(sp, M_SYSERR, ep->rcv_path,
 			    "058|Preservation failed: %s");
 			sp->gp->scr_busy(sp, NULL, BUSY_OFF);
@@ -241,14 +227,14 @@ err:	msgq(sp, M_ERR,
  *		sending email to the user if the file was modified
  *		ending the file session
  *
- * PUBLIC: int rcv_sync __P((SCR *, u_int));
+ * PUBLIC: int rcv_sync(SCR *, u_int);
  */
 int
 rcv_sync(SCR *sp, u_int flags)
 {
 	EXF *ep;
 	int fd, rval;
-	char *dp, buf[1024];
+	char *dp, *buf;
 
 	/* Make sure that there's something to recover/sync. */
 	ep = sp->ep;
@@ -257,7 +243,7 @@ rcv_sync(SCR *sp, u_int flags)
 
 	/* Sync the file if it's been modified. */
 	if (F_ISSET(ep, F_MODIFIED)) {
-		if (ep->db->sync(ep->db, 0)) {
+		if (ep->db->sync(ep->db, R_RECNOSYNC)) {
 			F_CLR(ep, F_RCV_ON | F_RCV_NORM);
 			msgq_str(sp, M_SYSERR,
 			    ep->rcv_path, "060|File backup failed: %s");
@@ -288,9 +274,14 @@ rcv_sync(SCR *sp, u_int flags)
 		if (opts_empty(sp, O_RECDIR, 0))
 			goto err;
 		dp = O_STR(sp, O_RECDIR);
-		(void)snprintf(buf, sizeof(buf), "%s/vi.XXXXXX", dp);
-		if ((fd = rcv_mktemp(sp, buf, dp, S_IRUSR | S_IWUSR)) == -1)
+		if ((buf = join(dp, "vi.XXXXXX")) == NULL) {
+			msgq(sp, M_SYSERR, NULL);
 			goto err;
+		}
+		if ((fd = rcv_mktemp(sp, buf, dp)) == -1) {
+			free(buf);
+			goto err;
+		}
 		sp->gp->scr_busy(sp,
 		    "061|Copying file for recovery...", BUSY_ON);
 		if (rcv_copy(sp, fd, ep->rcv_path) ||
@@ -299,6 +290,7 @@ rcv_sync(SCR *sp, u_int flags)
 			(void)close(fd);
 			rval = 1;
 		}
+		free(buf);
 		sp->gp->scr_busy(sp, NULL, BUSY_OFF);
 	}
 	if (0) {
@@ -322,22 +314,24 @@ rcv_mailfile(SCR *sp, int issync, char *cp_path)
 	EXF *ep;
 	GS *gp;
 	struct passwd *pw;
-	size_t len;
+	int len;
 	time_t now;
 	uid_t uid;
 	int fd;
-	char *dp, *p, *t, buf[4096], mpath[MAXPATHLEN];
+	FILE *fp;
+	char *dp, *p, *t, *qt, *buf, *mpath;
 	char *t1, *t2, *t3;
+	int st;
 
 	/*
 	 * XXX
-	 * MAXHOSTNAMELEN is in various places on various systems, including
-	 * <netdb.h> and <sys/socket.h>.  If not found, use a large default.
+	 * MAXHOSTNAMELEN/HOST_NAME_MAX are deprecated. We try sysconf(3)
+	 * first, then fallback to _POSIX_HOST_NAME_MAX.
 	 */
-#ifndef MAXHOSTNAMELEN
-#define	MAXHOSTNAMELEN	1024
-#endif
-	char host[MAXHOSTNAMELEN];
+	char *host;
+	long hostmax = sysconf(_SC_HOST_NAME_MAX);
+	if (hostmax < 0)
+		hostmax = _POSIX_HOST_NAME_MAX;
 
 	gp = sp->gp;
 	if ((pw = getpwuid(uid = getuid())) == NULL) {
@@ -349,9 +343,19 @@ rcv_mailfile(SCR *sp, int issync, char *cp_path)
 	if (opts_empty(sp, O_RECDIR, 0))
 		return (1);
 	dp = O_STR(sp, O_RECDIR);
-	(void)snprintf(mpath, sizeof(mpath), "%s/recover.XXXXXX", dp);
-	if ((fd = rcv_mktemp(sp, mpath, dp, S_IRUSR | S_IWUSR)) == -1)
+	if ((mpath = join(dp, "recover.XXXXXX")) == NULL) {
+		msgq(sp, M_SYSERR, NULL);
 		return (1);
+	}
+	if ((fd = rcv_mktemp(sp, mpath, dp)) == -1) {
+		free(mpath);
+		return (1);
+	}
+	if ((fp = fdopen(fd, "w")) == NULL) {
+		free(mpath);
+		close(fd);
+		return (1);
+	}
 
 	/*
 	 * XXX
@@ -361,56 +365,64 @@ rcv_mailfile(SCR *sp, int issync, char *cp_path)
 	 * and the lock, but it's pretty small.
 	 */
 	ep = sp->ep;
-	if (file_lock(sp, NULL, NULL, fd, 1) != LOCK_SUCCESS)
+	if (file_lock(sp, NULL, fd, 1) != LOCK_SUCCESS)
 		msgq(sp, M_SYSERR, "063|Unable to lock recovery file");
 	if (!issync) {
 		/* Save the recover file descriptor, and mail path. */
-		ep->rcv_fd = fd;
-		if ((ep->rcv_mpath = strdup(mpath)) == NULL) {
-			msgq(sp, M_SYSERR, NULL);
-			goto err;
-		}
+		ep->rcv_fd = dup(fd);
+		ep->rcv_mpath = mpath;
 		cp_path = ep->rcv_path;
 	}
 
-	/*
-	 * XXX
-	 * We can't use stdio(3) here.  The problem is that we may be using
-	 * fcntl(2), so if ANY file descriptor into the file is closed, the
-	 * lock is lost.  So, we could never close the FILE *, even if we
-	 * dup'd the fd first.
-	 */
 	t = sp->frp->name;
 	if ((p = strrchr(t, '/')) == NULL)
 		p = t;
 	else
 		++p;
 	(void)time(&now);
-	(void)gethostname(host, sizeof(host));
-	len = snprintf(buf, sizeof(buf),
-	    "%s%s\n%s%s\n%s\n%s\n%s%s\n%s%s\n%s\n\n",
-	    VI_FHEADER, t,			/* Non-standard. */
-	    VI_PHEADER, cp_path,		/* Non-standard. */
-	    "Reply-To: root",
-	    "From: root (Nvi recovery program)",
-	    "To: ", pw->pw_name,
+
+	if ((st = rcv_dlnwrite(sp, "file", t, fp))) {
+		if (st == 1)
+			goto werr;
+		goto err;
+	}
+	if ((st = rcv_dlnwrite(sp, "path", cp_path, fp))) {
+		if (st == 1)
+			goto werr;
+		goto err;
+	}
+
+	MALLOC(sp, host, hostmax + 1);
+	if (host == NULL)
+		goto err;
+	(void)gethostname(host, hostmax + 1);
+
+	len = fprintf(fp, "%s%s%s\n%s%s%s%s\n%s%.40s\n%s\n\n",
+	    "From: root@", host, " (Nvi recovery program)",
+	    "To: ", pw->pw_name, "@", host,
 	    "Subject: Nvi saved the file ", p,
 	    "Precedence: bulk");		/* For vacation(1). */
-	if (len > sizeof(buf) - 1)
-		goto lerr;
-	if (write(fd, buf, len) != len)
+	if (len < 0) {
+		free(host);
 		goto werr;
+	}
 
-	len = snprintf(buf, sizeof(buf),
-	    "%s%.24s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n\n",
+	if ((qt = quote(t)) == NULL) {
+		free(host);
+		msgq(sp, M_SYSERR, NULL);
+		goto err;
+	}
+	len = asprintf(&buf, "%s%.24s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n\n",
 	    "On ", ctime(&now), ", the user ", pw->pw_name,
 	    " was editing a file named ", t, " on the machine ",
 	    host, ", when it was saved for recovery. ",
 	    "You can recover most, if not all, of the changes ",
-	    "to this file using the -r option to ", gp->progname, ":\n\n\t",
-	    gp->progname, " -r ", t);
-	if (len > sizeof(buf) - 1) {
-lerr:		msgq(sp, M_ERR, "064|Recovery file buffer overrun");
+	    "to this file using the -r option to ", getprogname(), ":\n\n\t",
+	    getprogname(), " -r ", qt);
+	free(qt);
+	free(host);
+	if (buf == NULL) {
+		msgq(sp, M_SYSERR, NULL);
 		goto err;
 	}
 
@@ -444,23 +456,29 @@ lerr:		msgq(sp, M_ERR, "064|Recovery file buffer overrun");
 wout:		*t2++ = '\n';
 
 		/* t2 points one after the last character to display. */
-		if (write(fd, t1, t2 - t1) != t2 - t1)
+		if (fwrite(t1, 1, t2 - t1, fp) != t2 - t1) {
+			free(buf);
 			goto werr;
+		}
 	}
 
 	if (issync) {
+		fflush(fp);
 		rcv_email(sp, mpath);
-		if (close(fd)) {
-werr:			msgq(sp, M_SYSERR, "065|Recovery file");
-			goto err;
-		}
+		free(mpath);
 	}
+	if (fclose(fp)) {
+		free(buf);
+werr:		msgq(sp, M_SYSERR, "065|Recovery file");
+		goto err;
+	}
+	free(buf);
 	return (0);
 
 err:	if (!issync)
 		ep->rcv_fd = -1;
-	if (fd != -1)
-		(void)close(fd);
+	if (fp != NULL)
+		(void)fclose(fp);
 	return (1);
 }
 
@@ -472,7 +490,7 @@ err:	if (!issync)
  * rcv_list --
  *	List the files that can be recovered by this user.
  *
- * PUBLIC: int rcv_list __P((SCR *));
+ * PUBLIC: int rcv_list(SCR *);
  */
 int
 rcv_list(SCR *sp)
@@ -482,7 +500,9 @@ rcv_list(SCR *sp)
 	DIR *dirp;
 	FILE *fp;
 	int found;
-	char *p, *t, file[MAXPATHLEN], path[MAXPATHLEN];
+	char *p, *file, *path;
+	char *dtype, *data;
+	int st;
 
 	/* Open the recovery directory for reading. */
 	if (opts_empty(sp, O_RECDIR, 0))
@@ -498,18 +518,11 @@ rcv_list(SCR *sp)
 		if (strncmp(dp->d_name, "recover.", 8))
 			continue;
 
-		/*
-		 * If it's readable, it's recoverable.
-		 *
-		 * XXX
-		 * Should be "r", we don't want to write the file.  However,
-		 * if we're using fcntl(2), there's no way to lock a file
-		 * descriptor that's not open for writing.
-		 */
-		if ((fp = fopen(dp->d_name, "r+")) == NULL)
+		/* If it's readable, it's recoverable. */
+		if ((fp = fopen(dp->d_name, "r")) == NULL)
 			continue;
 
-		switch (file_lock(sp, NULL, NULL, fileno(fp), 1)) {
+		switch (file_lock(sp, NULL, fileno(fp), 1)) {
 		case LOCK_FAILED:
 			/*
 			 * XXX
@@ -528,17 +541,23 @@ rcv_list(SCR *sp)
 		}
 
 		/* Check the headers. */
-		if (fgets(file, sizeof(file), fp) == NULL ||
-		    strncmp(file, VI_FHEADER, sizeof(VI_FHEADER) - 1) ||
-		    (p = strchr(file, '\n')) == NULL ||
-		    fgets(path, sizeof(path), fp) == NULL ||
-		    strncmp(path, VI_PHEADER, sizeof(VI_PHEADER) - 1) ||
-		    (t = strchr(path, '\n')) == NULL) {
-			msgq_str(sp, M_ERR, dp->d_name,
-			    "066|%s: malformed recovery file");
-			goto next;
+		for (file = NULL, path = NULL;
+		    file == NULL || path == NULL;) {
+			if ((st = rcv_dlnread(sp, &dtype, &data, fp))) {
+				if (st == 1)
+					msgq_str(sp, M_ERR, dp->d_name,
+					    "066|%s: malformed recovery file");
+				goto next;
+			}
+			if (dtype == NULL)
+				continue;
+			if (!strcmp(dtype, "file"))
+				file = data;
+			else if (!strcmp(dtype, "path"))
+				path = data;
+			else
+				free(data);
 		}
-		*p = *t = '\0';
 
 		/*
 		 * If the file doesn't exist, it's an orphaned recovery file,
@@ -549,7 +568,7 @@ rcv_list(SCR *sp)
 		 * before deleting the email file.
 		 */
 		errno = 0;
-		if (stat(path + sizeof(VI_PHEADER) - 1, &sb) &&
+		if (stat(path, &sb) &&
 		    errno == ENOENT) {
 			(void)unlink(dp->d_name);
 			goto next;
@@ -558,14 +577,16 @@ rcv_list(SCR *sp)
 		/* Get the last modification time and display. */
 		(void)fstat(fileno(fp), &sb);
 		(void)printf("%.24s: %s\n",
-		    ctime(&sb.st_mtime), file + sizeof(VI_FHEADER) - 1);
+		    ctime(&sb.st_mtime), file);
 		found = 1;
 
 		/* Close, discarding lock. */
 next:		(void)fclose(fp);
+		free(file);
+		free(path);
 	}
 	if (found == 0)
-		(void)printf("vi: no files to recover.\n");
+		(void)printf("%s: No files to recover\n", getprogname());
 	(void)closedir(dirp);
 	return (0);
 }
@@ -574,7 +595,7 @@ next:		(void)fclose(fp);
  * rcv_read --
  *	Start a recovered file as the file to edit.
  *
- * PUBLIC: int rcv_read __P((SCR *, FREF *));
+ * PUBLIC: int rcv_read(SCR *, FREF *);
  */
 int
 rcv_read(SCR *sp, FREF *frp)
@@ -582,46 +603,41 @@ rcv_read(SCR *sp, FREF *frp)
 	struct dirent *dp;
 	struct stat sb;
 	DIR *dirp;
+	FILE *fp;
 	EXF *ep;
-	time_t rec_mtime;
-	int fd, found, locked, requested, sv_fd;
+	struct timespec rec_mtim = { 0, 0 };
+	int found, locked = 0, requested, sv_fd;
 	char *name, *p, *t, *rp, *recp, *pathp;
-	char file[MAXPATHLEN], path[MAXPATHLEN], recpath[MAXPATHLEN];
+	char *file, *path, *recpath;
+	char *dtype, *data;
+	int st;
 
 	if (opts_empty(sp, O_RECDIR, 0))
 		return (1);
 	rp = O_STR(sp, O_RECDIR);
 	if ((dirp = opendir(rp)) == NULL) {
-		msgq_str(sp, M_ERR, rp, "%s");
+		msgq_str(sp, M_SYSERR, rp, "%s");
 		return (1);
 	}
 
 	name = frp->name;
 	sv_fd = -1;
-	rec_mtime = 0;
 	recp = pathp = NULL;
 	for (found = requested = 0; (dp = readdir(dirp)) != NULL;) {
 		if (strncmp(dp->d_name, "recover.", 8))
 			continue;
-		(void)snprintf(recpath,
-		    sizeof(recpath), "%s/%s", rp, dp->d_name);
-
-		/*
-		 * If it's readable, it's recoverable.  It would be very
-		 * nice to use stdio(3), but, we can't because that would
-		 * require closing and then reopening the file so that we
-		 * could have a lock and still close the FP.  Another tip
-		 * of the hat to fcntl(2).
-		 *
-		 * XXX
-		 * Should be O_RDONLY, we don't want to write it.  However,
-		 * if we're using fcntl(2), there's no way to lock a file
-		 * descriptor that's not open for writing.
-		 */
-		if ((fd = open(recpath, O_RDWR, 0)) == -1)
+		if ((recpath = join(rp, dp->d_name)) == NULL) {
+			msgq(sp, M_SYSERR, NULL);
 			continue;
+		}
 
-		switch (file_lock(sp, NULL, NULL, fd, 1)) {
+		/* If it's readable, it's recoverable. */
+		if ((fp = fopen(recpath, "r")) == NULL) {
+			free(recpath);
+			continue;
+		}
+
+		switch (file_lock(sp, NULL, fileno(fp), 1)) {
 		case LOCK_FAILED:
 			/*
 			 * XXX
@@ -637,22 +653,28 @@ rcv_read(SCR *sp, FREF *frp)
 			break;
 		case LOCK_UNAVAIL:
 			/* If it's locked, it's live. */
-			(void)close(fd);
+			(void)fclose(fp);
 			continue;
 		}
 
 		/* Check the headers. */
-		if (rcv_gets(file, sizeof(file), fd) == NULL ||
-		    strncmp(file, VI_FHEADER, sizeof(VI_FHEADER) - 1) ||
-		    (p = strchr(file, '\n')) == NULL ||
-		    rcv_gets(path, sizeof(path), fd) == NULL ||
-		    strncmp(path, VI_PHEADER, sizeof(VI_PHEADER) - 1) ||
-		    (t = strchr(path, '\n')) == NULL) {
-			msgq_str(sp, M_ERR, recpath,
-			    "067|%s: malformed recovery file");
-			goto next;
+		for (file = NULL, path = NULL;
+		    file == NULL || path == NULL;) {
+			if ((st = rcv_dlnread(sp, &dtype, &data, fp))) {
+				if (st == 1)
+					msgq_str(sp, M_ERR, dp->d_name,
+					    "067|%s: malformed recovery file");
+				goto next;
+			}
+			if (dtype == NULL)
+				continue;
+			if (!strcmp(dtype, "file"))
+				file = data;
+			else if (!strcmp(dtype, "path"))
+				path = data;
+			else
+				free(data);
 		}
-		*p = *t = '\0';
 		++found;
 
 		/*
@@ -664,52 +686,40 @@ rcv_read(SCR *sp, FREF *frp)
 		 * before deleting the email file.
 		 */
 		errno = 0;
-		if (stat(path + sizeof(VI_PHEADER) - 1, &sb) &&
+		if (stat(path, &sb) &&
 		    errno == ENOENT) {
 			(void)unlink(dp->d_name);
 			goto next;
 		}
 
 		/* Check the file name. */
-		if (strcmp(file + sizeof(VI_FHEADER) - 1, name))
+		if (strcmp(file, name))
 			goto next;
 
 		++requested;
 
-		/*
-		 * If we've found more than one, take the most recent.
-		 *
-		 * XXX
-		 * Since we're using st_mtime, for portability reasons,
-		 * we only get a single second granularity, instead of
-		 * getting it right.
-		 */
-		(void)fstat(fd, &sb);
-		if (recp == NULL || rec_mtime < sb.st_mtime) {
+		/* If we've found more than one, take the most recent. */
+		(void)fstat(fileno(fp), &sb);
+		if (recp == NULL ||
+		    timespeccmp(&rec_mtim, &sb.st_mtimespec, <)) {
 			p = recp;
 			t = pathp;
-			if ((recp = strdup(recpath)) == NULL) {
-				msgq(sp, M_SYSERR, NULL);
-				recp = p;
-				goto next;
-			}
-			if ((pathp = strdup(path)) == NULL) {
-				msgq(sp, M_SYSERR, NULL);
-				free(recp);
-				recp = p;
-				pathp = t;
-				goto next;
-			}
+			recp = recpath;
+			pathp = path;
 			if (p != NULL) {
 				free(p);
 				free(t);
 			}
-			rec_mtime = sb.st_mtime;
+			rec_mtim = sb.st_mtimespec;
 			if (sv_fd != -1)
 				(void)close(sv_fd);
-			sv_fd = fd;
-		} else
-next:			(void)close(fd);
+			sv_fd = dup(fileno(fp));
+		} else {
+next:			free(recpath);
+			free(path);
+		}
+		(void)fclose(fp);
+		free(file);
 	}
 	(void)closedir(dirp);
 
@@ -733,12 +743,13 @@ next:			(void)close(fd);
 	 * XXX
 	 * file_init() is going to set ep->rcv_path.
 	 */
-	if (file_init(sp, frp, pathp + sizeof(VI_PHEADER) - 1, 0)) {
+	if (file_init(sp, frp, pathp, 0)) {
 		free(recp);
 		free(pathp);
 		(void)close(sv_fd);
 		return (1);
 	}
+	free(pathp);
 
 	/*
 	 * We keep an open lock on the file so that the recover option can
@@ -779,47 +790,17 @@ err:	msgq_str(sp, M_SYSERR, fname, "%s");
 	return (1);
 }
 
-/*
- * rcv_gets --
- *	Fgets(3) for a file descriptor.
- */
-static char *
-rcv_gets(char *buf, size_t len, int fd)
-{
-	int nr;
-	char *p;
-
-	if ((nr = read(fd, buf, len - 1)) == -1)
-		return (NULL);
-	if ((p = strchr(buf, '\n')) == NULL)
-		return (NULL);
-	(void)lseek(fd, (off_t)((p - buf) + 1), SEEK_SET);
-	return (buf);
-}
-
 /*
  * rcv_mktemp --
  *	Paranoid make temporary file routine.
  */
 static int
-rcv_mktemp(SCR *sp, char *path, char *dname, int perms)
+rcv_mktemp(SCR *sp, char *path, char *dname)
 {
 	int fd;
 
-	/*
-	 * !!!
-	 * We expect mkstemp(3) to set the permissions correctly.  On
-	 * historic System V systems, mkstemp didn't.  Do it here, on
-	 * GP's.
-	 *
-	 * XXX
-	 * The variable perms should really be a mode_t, and it would
-	 * be nice to use fchmod(2) instead of chmod(2), here.
-	 */
 	if ((fd = mkstemp(path)) == -1)
 		msgq_str(sp, M_SYSERR, dname, "%s");
-	else
-		(void)chmod(path, perms);
 	return (fd);
 }
 
@@ -830,22 +811,131 @@ rcv_mktemp(SCR *sp, char *path, char *dname, int perms)
 static void
 rcv_email(SCR *sp, char *fname)
 {
-	struct stat sb;
-	char buf[MAXPATHLEN * 2 + 20];
+	char *buf;
 
-	if (_PATH_SENDMAIL[0] != '/' || stat(_PATH_SENDMAIL, &sb))
-		msgq_str(sp, M_SYSERR,
-		    _PATH_SENDMAIL, "071|not sending email: %s");
-	else {
-		/*
-		 * !!!
-		 * If you need to port this to a system that doesn't have
-		 * sendmail, the -t flag causes sendmail to read the message
-		 * for the recipients instead of specifying them some other
-		 * way.
-		 */
-		(void)snprintf(buf, sizeof(buf),
-		    "%s -t < %s", _PATH_SENDMAIL, fname);
-		(void)system(buf);
+	(void)asprintf(&buf, _PATH_SENDMAIL " -odb -t < %s", fname);
+	if (buf == NULL) {
+		msgq_str(sp, M_ERR, strerror(errno),
+		    "071|not sending email: %s");
+		return;
 	}
+	(void)system(buf);
+	free(buf);
+}
+
+/*
+ * rcv_dlnwrite --
+ *	Encode a string into an X-vi-data line and write it.
+ */
+static int
+rcv_dlnwrite(SCR *sp, const char *dtype, const char *src, FILE *fp)
+{
+	char *bp = NULL, *p;
+	size_t blen = 0;
+	size_t dlen, len;
+	int plen, xlen;
+
+	len = strlen(src);
+	dlen = strlen(dtype);
+	GET_SPACE_GOTOC(sp, bp, blen, (len + 2) / 3 * 4 + dlen + 2);
+	(void)memcpy(bp, dtype, dlen);
+	bp[dlen] = ';';
+	if ((xlen = b64_ntop((u_char *)src,
+	    len, bp + dlen + 1, blen)) == -1)
+		goto err;
+	xlen += dlen + 1;
+
+	/* Output as an MIME folding header. */
+	if ((plen = fprintf(fp, VI_DHEADER " %.*s\n",
+	    FMTCOLS - (int)sizeof(VI_DHEADER), bp)) < 0)
+		goto err;
+	plen -= (int)sizeof(VI_DHEADER) + 1;
+	for (p = bp, xlen -= plen; xlen > 0; xlen -= plen) {
+		p += plen;
+		if ((plen = fprintf(fp, " %.*s\n", FMTCOLS - 1, p)) < 0)
+			goto err;
+		plen -= 2;
+	}
+	FREE_SPACE(sp, bp, blen);
+	return (0);
+
+err:	FREE_SPACE(sp, bp, blen);
+	return (1);
+alloc_err:
+	msgq(sp, M_SYSERR, NULL);
+	return (-1);
+}
+
+/*
+ * rcv_dlnread --
+ *	Read an X-vi-data line and decode it.
+ */
+static int
+rcv_dlnread(SCR *sp, char **dtypep,
+	char **datap,		/* free *datap if != NULL after use. */
+	FILE *fp)
+{
+	int ch;
+	char buf[1024];
+	char *bp = NULL, *p, *src;
+	size_t blen = 0;
+	size_t len, off, dlen;
+	char *dtype, *data;
+	int xlen;
+
+	if (fgets(buf, sizeof(buf), fp) == NULL)
+		return (1);
+	if (strncmp(buf, VI_DHEADER, sizeof(VI_DHEADER) - 1)) {
+		*dtypep = NULL;
+		*datap = NULL;
+		return (0);
+	}
+
+	/* Fetch an MIME folding header. */
+	len = strlen(buf) - sizeof(VI_DHEADER) + 1;
+	GET_SPACE_GOTOC(sp, bp, blen, len);
+	(void)memcpy(bp, buf + sizeof(VI_DHEADER) - 1, len);
+	p = bp + len;
+	while ((ch = fgetc(fp)) == ' ') {
+		if (fgets(buf, sizeof(buf), fp) == NULL)
+			goto err;
+		off = strlen(buf);
+		len += off;
+		ADD_SPACE_GOTOC(sp, bp, blen, len);
+		p = bp + len - off;
+		(void)memcpy(p, buf, off);
+	}
+	bp[len] = '\0';
+	(void)ungetc(ch, fp);
+
+	for (p = bp; *p == ' ' || *p == '\n'; p++);
+	if ((src = strchr(p, ';')) == NULL)
+		goto err;
+	dlen = src - p;
+	src += 1;
+	len -= src - bp;
+
+	/* Memory looks like: "<data>\0<dtype>\0". */
+	MALLOC(sp, data, dlen + len / 4 * 3 + 2);
+	if (data == NULL)
+		goto err;
+	if ((xlen = (b64_pton(p + dlen + 1,
+	    (u_char *)data, len / 4 * 3 + 1))) == -1) {
+		free(data);
+		goto err;
+	}
+	data[xlen] = '\0';
+	dtype = data + xlen + 1;
+	(void)memcpy(dtype, p, dlen);
+	dtype[dlen] = '\0';
+	FREE_SPACE(sp, bp, blen);
+	*dtypep = dtype;
+	*datap = data;
+	return (0);
+
+err: 	FREE_SPACE(sp, bp, blen);
+	return (1);
+alloc_err:
+	msgq(sp, M_SYSERR, NULL);
+	return (-1);
 }
