@@ -154,10 +154,6 @@ ex(SCR **spp)
 			if (file_end(sp, NULL, F_ISSET(sp, SC_EXIT_FORCE)))
 				return (1);
 			*spp = screen_next(sp);
-			if (*spp) {
-				F_CLR(*spp, SC_SCR_VI);
-				F_SET(*spp, SC_SCR_EX);
-			}
 			return (screen_end(sp));
 		}
 	}
@@ -773,7 +769,6 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 	 * no longer useful.
 	 */
 	vi_address = ecp->clen != 0 && ecp->cp[0] != '\n';
-	ecp->trailing = 0;
 	for (p = ecp->cp; ecp->clen > 0; --ecp->clen, ++ecp->cp) {
 		ch = ecp->cp[0];
 		if (IS_ESCAPE(sp, ecp, ch) && ecp->clen > 1) {
@@ -789,7 +784,6 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 				ch = tmp;
 			}
 		} else if (ch == '\n' || ch == '|') {
-			ecp->trailing = 1;
 			if (ch == '\n')
 				F_SET(ecp, E_NEWLINE);
 			--ecp->clen;
@@ -817,14 +811,13 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 	 * this isn't a particularly complex trap, and if backslashes were
 	 * legal in set commands, this would have to be much more complicated.
 	 */
-	if (ecp->cmd == &cmds[C_SET]) {
+	if (ecp->cmd == &cmds[C_SET])
 		for (p = ecp->cp, len = ecp->clen; len > 0; --len, ++p)
 			if (IS_ESCAPE(sp, ecp, *p) && len > 1) {
 				--len;
 				++p;
 			} else if (*p == '\\')
 				*p = CH_LITERAL;
-	}
 
 	/*
 	 * Set the default addresses.  It's an error to specify an address for
@@ -1261,7 +1254,7 @@ addr_verify:
 				ex_badaddr(sp, ecp->cmd, A_ZERO, NUM_OK);
 				goto err;
 			}
-		} else if (!db_exist(sp, ecp->addr2.lno)) {
+		} else if (!db_exist(sp, ecp->addr2.lno))
 			if (FL_ISSET(ecp->iflags, E_C_COUNT)) {
 				if (db_last(sp, &lno))
 					goto err;
@@ -1270,7 +1263,6 @@ addr_verify:
 				ex_badaddr(sp, NULL, A_EOF, NUM_OK);
 				goto err;
 			}
-		}
 		/* FALLTHROUGH */
 	case 1:
 		if (ecp->addr1.lno == 0) {
@@ -1469,13 +1461,8 @@ addr_verify:
 		LF_INIT(FL_ISSET(ecp->iflags, E_C_HASH | E_C_LIST | E_C_PRINT));
 		if (!LF_ISSET(E_C_HASH | E_C_LIST | E_C_PRINT | E_NOAUTO) &&
 		    !F_ISSET(sp, SC_EX_GLOBAL) &&
-		    O_ISSET(sp, O_AUTOPRINT) && F_ISSET(ecp, E_AUTOPRINT)) {
-			/* Honor the number option if autoprint is set. */
-			if (F_ISSET(ecp, E_OPTNUM))
-				LF_INIT(E_C_HASH);
-			else
-				LF_INIT(E_C_PRINT);
-		}
+		    O_ISSET(sp, O_AUTOPRINT) && F_ISSET(ecp, E_AUTOPRINT))
+			LF_INIT(E_C_PRINT);
 
 		if (LF_ISSET(E_C_HASH | E_C_LIST | E_C_PRINT)) {
 			cur.lno = sp->lno;
@@ -2130,7 +2117,7 @@ ex_load(SCR *sp)
 
 			/* If it's a global/v command, fix up the last line. */
 			if (FL_ISSET(ecp->agv_flags,
-			    AGV_GLOBAL | AGV_V) && ecp->range_lno != OOBLNO) {
+			    AGV_GLOBAL | AGV_V) && ecp->range_lno != OOBLNO)
 				if (db_exist(sp, ecp->range_lno))
 					sp->lno = ecp->range_lno;
 				else {
@@ -2139,7 +2126,6 @@ ex_load(SCR *sp)
 					if (sp->lno == 0)
 						sp->lno = 1;
 				}
-			}
 			free(ecp->o_cp);
 		}
 
@@ -2277,8 +2263,7 @@ ex_comm_search(CHAR_T *name, size_t len)
 			return (NULL);
 		if (cp->name[0] != name[0])
 			continue;
-		if (STRLEN(cp->name) >= len &&
-			!MEMCMP(name, cp->name, len))
+		if (!MEMCMP(name, cp->name, len))
 			return (cp);
 	}
 	return (NULL);
