@@ -5,15 +5,13 @@
  *	Keith Bostic.  All rights reserved.
  *
  * See the LICENSE file for redistribution information.
- *
- *	$Id: ex.h,v 10.30 2004/03/16 14:11:33 skimo Exp $ (Berkeley) $Date: 2004/03/16 14:11:33 $
  */
 
 #define	PROMPTCHAR	':'		/* Prompt using a colon. */
 
 typedef struct _excmdlist {		/* Ex command table structure. */
 	CHAR_T *name;			/* Command name, underlying function. */
-	int (*fn) __P((SCR *, EXCMD *));
+	int (*fn)(SCR *, EXCMD *);
 
 #define	E_ADDR1		0x00000001	/* One address. */
 #define	E_ADDR2		0x00000002	/* Two addresses. */
@@ -49,6 +47,9 @@ extern EXCMDLIST const cmds[];		/* Table of ex commands. */
 	(F_ISSET(cmdp, E_VLITONLY) ?					\
 	    (ch) == CH_LITERAL : KEY_VAL(sp, ch) == K_VLNEXT)
 
+#define	IS_SHELLMETA(sp, ch)						\
+	((ch) <= CHAR_MAX && strchr(O_STR(sp, O_SHELLMETA), ch) != NULL)
+
 /*
  * File state must be checked for each command -- any ex command may be entered
  * at any time, and most of them won't work well if a file hasn't yet been read
@@ -64,16 +65,16 @@ extern EXCMDLIST const cmds[];		/* Table of ex commands. */
 /* Range structures for global and @ commands. */
 typedef struct _range RANGE;
 struct _range {				/* Global command range. */
-	CIRCLEQ_ENTRY(_range) q;	/* Linked list of ranges. */
-	db_recno_t start, stop;		/* Start/stop of the range. */
+	TAILQ_ENTRY(_range) q;		/* Linked list of ranges. */
+	recno_t start, stop;		/* Start/stop of the range. */
 };
 
 /* Ex command structure. */
 struct _excmd {
-	LIST_ENTRY(_excmd) q;		/* Linked list of commands. */
+	SLIST_ENTRY(_excmd) q;		/* Linked list of commands. */
 
 	char	 *if_name;		/* Associated file. */
-	db_recno_t	  if_lno;		/* Associated line number. */
+	recno_t	  if_lno;		/* Associated line number. */
 
 	/* Clear the structure for the ex parser. */
 #define	CLEAR_EX_PARSER(cmdp)						\
@@ -89,8 +90,8 @@ struct _excmd {
 	EXCMDLIST const *cmd;		/* Command: entry in command table. */
 	EXCMDLIST rcmd;			/* Command: table entry/replacement. */
 
-	CIRCLEQ_HEAD(_rh, _range) rq;	/* @/global range: linked list. */
-	db_recno_t   range_lno;		/* @/global range: set line number. */
+	TAILQ_HEAD(_rh, _range) rq[1];	/* @/global range: linked list. */
+	recno_t   range_lno;		/* @/global range: set line number. */
 	CHAR_T	 *o_cp;			/* Original @/global command. */
 	size_t	  o_clen;		/* Original @/global command length. */
 #define	AGV_AT		0x01		/* @ buffer execution. */
@@ -109,7 +110,7 @@ struct _excmd {
 }
 
 	CHAR_T	  buffer;		/* Command: named buffer. */
-	db_recno_t	  lineno;		/* Command: line number. */
+	recno_t	  lineno;		/* Command: line number. */
 	long	  count;		/* Command: signed count. */
 	long	  flagoff;		/* Command: signed flag offset. */
 	int	  addrcnt;		/* Command: addresses (0, 1 or 2). */
@@ -157,9 +158,10 @@ struct _excmd {
 
 /* Ex private, per-screen memory. */
 typedef struct _ex_private {
-	CIRCLEQ_HEAD(_tqh, _tagq) tq;	/* Tag queue. */
-	TAILQ_HEAD(_tagfh, _tagf) tagfq;/* Tag file list. */
-	LIST_HEAD(_csch, _csc) cscq;    /* Cscope connection list. */
+					/* Tag file list. */
+	TAILQ_HEAD(_tagfh, _tagf) tagfq[1];
+	TAILQ_HEAD(_tqh, _tagq) tq[1];	/* Tag queue. */
+	SLIST_HEAD(_csch, _csc) cscq[1];/* Cscope connection list. */
 	CHAR_T	*tag_last;		/* Saved last tag string. */
 
 	CHAR_T	*lastbcomm;		/* Last bang command. */
@@ -204,7 +206,6 @@ enum filtertype { FILTER_BANG, FILTER_RBANG, FILTER_READ, FILTER_WRITE };
 typedef enum {
 	EXM_EMPTYBUF,			/* Empty buffer. */
 	EXM_FILECOUNT,			/* Too many file names. */
-	EXM_LOCKED,			/* Another thread is active. */
 	EXM_NOCANON,			/* No terminal interface. */
 	EXM_NOCANON_F,			/* EXM_NOCANO: filter version. */
 	EXM_NOFILEYET,			/* Illegal until a file read in. */
@@ -219,7 +220,7 @@ typedef enum {
 /* Ex address error types. */
 enum badaddr { A_COMBO, A_EMPTY, A_EOF, A_NOTSET, A_ZERO };
 
-/* Ex common tag error messages. */                                         
+/* Ex common tag error messages. */
 typedef enum {
 	TAG_BADLNO,		/* Tag line doesn't exist. */
 	TAG_EMPTY,		/* Tags stack is empty. */
