@@ -9,10 +9,6 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: ex_bang.c,v 10.36 2001/06/25 15:19:14 skimo Exp $ (Berkeley) $Date: 2001/06/25 15:19:14 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/time.h>
@@ -44,7 +40,7 @@ static const char sccsid[] = "$Id: ex_bang.c,v 10.36 2001/06/25 15:19:14 skimo E
  * ways of getting here display the right things.  It took a long time to
  * get it right (wrong?), so be careful.
  *
- * PUBLIC: int ex_bang __P((SCR *, EXCMD *));
+ * PUBLIC: int ex_bang(SCR *, EXCMD *);
  */
 int
 ex_bang(SCR *sp, EXCMD *cmdp)
@@ -53,7 +49,7 @@ ex_bang(SCR *sp, EXCMD *cmdp)
 	ARGS *ap;
 	EX_PRIVATE *exp;
 	MARK rm;
-	db_recno_t lno;
+	recno_t lno;
 	int rval;
 	const char *msg;
 	char *np;
@@ -67,8 +63,7 @@ ex_bang(SCR *sp, EXCMD *cmdp)
 
 	/* Set the "last bang command" remembered value. */
 	exp = EXP(sp);
-	if (exp->lastbcomm != NULL)
-		free(exp->lastbcomm);
+	free(exp->lastbcomm);
 	if ((exp->lastbcomm = v_wstrdup(sp, ap->bp, ap->len)) == NULL) {
 		msgq(sp, M_SYSERR, NULL);
 		return (1);
@@ -88,7 +83,7 @@ ex_bang(SCR *sp, EXCMD *cmdp)
 		if (F_ISSET(sp, SC_VI))
 			vs_update(sp, "!", ap->bp);
 		else {
-			(void)ex_printf(sp, "!%s\n", ap->bp);
+			(void)ex_printf(sp, "!"WS"\n", ap->bp);
 			(void)ex_fflush(sp);
 		}
 	}
@@ -178,6 +173,10 @@ ex_bang(SCR *sp, EXCMD *cmdp)
 	if (!F_ISSET(sp, SC_VI) && !F_ISSET(sp, SC_EX_SILENT))
 		(void)ex_puts(sp, "!\n");
 
+	/* Apply expandtab to the new text */
+	if (O_ISSET(sp, O_EXPANDTAB))
+		ex_retab(sp, cmdp);
+
 	/*
 	 * XXX
 	 * The ! commands never return an error, so that autoprint always
