@@ -9,15 +9,9 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: ex_cscope.c,v 10.21 2003/11/05 17:11:54 skimo Exp $ (Berkeley) $Date: 2003/11/05 17:11:54 $";
-#endif /* not lint */
-
-#include <sys/param.h>
-#include <sys/types.h>		/* XXX: param.h may not have included types.h */
+#include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/stat.h>
-#include <sys/time.h>
 #include <sys/wait.h>
 
 #include <bitstring.h>
@@ -25,6 +19,7 @@ static const char sccsid[] = "$Id: ex_cscope.c,v 10.21 2003/11/05 17:11:54 skimo
 #include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
+#include <signal.h>
 #include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -62,15 +57,15 @@ find c|d|e|f|g|i|s|t buffer|pattern\n\
       s: find all uses of name\n\
       t: find assignments to name"
 
-static int cscope_add __P((SCR *, EXCMD *, CHAR_T *));
-static int cscope_find __P((SCR *, EXCMD*, CHAR_T *));
-static int cscope_help __P((SCR *, EXCMD *, CHAR_T *));
-static int cscope_kill __P((SCR *, EXCMD *, CHAR_T *));
-static int cscope_reset __P((SCR *, EXCMD *, CHAR_T *));
+static int cscope_add(SCR *, EXCMD *, CHAR_T *);
+static int cscope_find(SCR *, EXCMD*, CHAR_T *);
+static int cscope_help(SCR *, EXCMD *, CHAR_T *);
+static int cscope_kill(SCR *, EXCMD *, CHAR_T *);
+static int cscope_reset(SCR *, EXCMD *, CHAR_T *);
 
 typedef struct _cc {
 	char	 *name;
-	int	(*function) __P((SCR *, EXCMD *, CHAR_T *));
+	int	(*function)(SCR *, EXCMD *, CHAR_T *);
 	char	 *help_msg;
 	char	 *usage_msg;
 } CC;
@@ -89,23 +84,23 @@ static CC const cscope_cmds[] = {
 	{ NULL }
 };
 
-static TAGQ	*create_cs_cmd __P((SCR *, char *, size_t *));
-static int	 csc_help __P((SCR *, char *));
-static void	 csc_file __P((SCR *,
-		    CSC *, char *, char **, size_t *, int *));
-static int	 get_paths __P((SCR *, CSC *));
-static CC const	*lookup_ccmd __P((char *));
-static int	 parse __P((SCR *, CSC *, TAGQ *, int *));
-static int	 read_prompt __P((SCR *, CSC *));
-static int	 run_cscope __P((SCR *, CSC *, char *));
-static int	 start_cscopes __P((SCR *, EXCMD *));
-static int	 terminate __P((SCR *, CSC *, int));
+static TAGQ	*create_cs_cmd(SCR *, char *, size_t *);
+static int	 csc_help(SCR *, char *);
+static void	 csc_file(SCR *,
+		    CSC *, char *, char **, size_t *, int *);
+static int	 get_paths(SCR *, CSC *);
+static CC const	*lookup_ccmd(char *);
+static int	 parse(SCR *, CSC *, TAGQ *, int *);
+static int	 read_prompt(SCR *, CSC *);
+static int	 run_cscope(SCR *, CSC *, char *);
+static int	 start_cscopes(SCR *, EXCMD *);
+static int	 terminate(SCR *, CSC *, int);
 
 /*
  * ex_cscope --
  *	Perform an ex cscope.
  *
- * PUBLIC: int ex_cscope __P((SCR *, EXCMD *));
+ * PUBLIC: int ex_cscope(SCR *, EXCMD *);
  */
 int
 ex_cscope(SCR *sp, EXCMD *cmdp)
@@ -202,8 +197,8 @@ cscope_add(SCR *sp, EXCMD *cmdp, CHAR_T *dname)
 	CSC *csc;
 	size_t len;
 	int cur_argc;
-	char *dbname, path[MAXPATHLEN];
-	char *np;
+	char *dbname, *path;
+	char *np = NULL;
 	size_t nlen;
 
 	exp = EXP(sp);
@@ -238,16 +233,20 @@ cscope_add(SCR *sp, EXCMD *cmdp, CHAR_T *dname)
 	 * name regardless so that we can use it as a base for searches.
 	 */
 	if (stat(np, &sb)) {
-		msgq(sp, M_SYSERR, np);
+		msgq(sp, M_SYSERR, "%s", np);
 		return (1);
 	}
 	if (S_ISDIR(sb.st_mode)) {
-		(void)snprintf(path, sizeof(path),
-		    "%s/%s", np, CSCOPE_DBFILE);
-		if (stat(path, &sb)) {
-			msgq(sp, M_SYSERR, path);
+		if ((path = join(np, CSCOPE_DBFILE)) == NULL) {
+			msgq(sp, M_SYSERR, NULL);
 			return (1);
 		}
+		if (stat(path, &sb)) {
+			msgq(sp, M_SYSERR, "%s", path);
+			free(path);
+			return (1);
+		}
+		free(path);
 		dbname = CSCOPE_DBFILE;
 	} else if ((dbname = strrchr(np, '/')) != NULL)
 		*dbname++ = '\0';
@@ -258,11 +257,11 @@ cscope_add(SCR *sp, EXCMD *cmdp, CHAR_T *dname)
 
 	/* Allocate a cscope connection structure and initialize its fields. */
 	len = strlen(np);
-	CALLOC_RET(sp, csc, CSC *, 1, sizeof(CSC) + len);
+	CALLOC_RET(sp, csc, 1, sizeof(CSC) + len);
 	csc->dname = csc->buf;
 	csc->dlen = len;
 	memcpy(csc->dname, np, len);
-	csc->mtime = sb.st_mtime;
+	csc->mtim = sb.st_mtimespec;
 
 	/* Get the search paths for the cscope. */
 	if (get_paths(sp, csc))
@@ -277,7 +276,7 @@ cscope_add(SCR *sp, EXCMD *cmdp, CHAR_T *dname)
 	 * on error, we have to call terminate, which expects the csc to
 	 * be on the chain.
 	 */
-	LIST_INSERT_HEAD(&exp->cscq, csc, q);
+	SLIST_INSERT_HEAD(exp->cscq, csc, q);
 
 	/* Read the initial prompt from the cscope to make sure it's okay. */
 	return read_prompt(sp, csc);
@@ -297,7 +296,7 @@ get_paths(SCR *sp, CSC *csc)
 	struct stat sb;
 	int fd, nentries;
 	size_t len;
-	char *p, **pathp, buf[MAXPATHLEN * 2];
+	char *p, **pathp, *buf;
 
 	/*
 	 * EXTENSION #2:
@@ -311,19 +310,24 @@ get_paths(SCR *sp, CSC *csc)
 	 * directory.  To fix this, rewrite the each path using the cscope
 	 * directory as a prefix.
 	 */
-	(void)snprintf(buf, sizeof(buf), "%s/%s", csc->dname, CSCOPE_PATHS);
+	if ((buf = join(csc->dname, CSCOPE_PATHS)) == NULL) {
+		msgq(sp, M_SYSERR, NULL);
+		return (1);
+	}
 	if (stat(buf, &sb) == 0) {
 		/* Read in the CSCOPE_PATHS file. */
 		len = sb.st_size;
-		MALLOC_RET(sp, csc->pbuf, char *, len + 1);
+		MALLOC_RET(sp, csc->pbuf, len + 1);
 		if ((fd = open(buf, O_RDONLY, 0)) < 0 ||
 		    read(fd, csc->pbuf, len) != len) {
 			 msgq_str(sp, M_SYSERR, buf, "%s");
 			 if (fd >= 0)
 				(void)close(fd);
+			 free(buf);
 			 return (1);
 		}
 		(void)close(fd);
+		free(buf);
 		csc->pbuf[len] = '\0';
 
 		/* Count up the entries. */
@@ -332,13 +336,13 @@ get_paths(SCR *sp, CSC *csc)
 				++nentries;
 
 		/* Build an array of pointers to the paths. */
-		CALLOC_GOTO(sp,
-		    csc->paths, char **, nentries + 1, sizeof(char **));
+		CALLOC_GOTO(sp, csc->paths, nentries + 1, sizeof(char **));
 		for (pathp = csc->paths, p = strtok(csc->pbuf, ":");
 		    p != NULL; p = strtok(NULL, ":"))
 			*pathp++ = p;
 		return (0);
 	}
+	free(buf);
 
 	/*
 	 * If the CSCOPE_PATHS file doesn't exist, we look for files
@@ -348,15 +352,13 @@ get_paths(SCR *sp, CSC *csc)
 		msgq(sp, M_SYSERR, NULL);
 		return (1);
 	}
-	CALLOC_GOTO(sp, csc->paths, char **, 2, sizeof(char *));
+	CALLOC_GOTO(sp, csc->paths, 2, sizeof(char *));
 	csc->paths[0] = csc->pbuf;
 	return (0);
 
 alloc_err:
-	if (csc->pbuf != NULL) {
-		free(csc->pbuf);
-		csc->pbuf = NULL;
-	}
+	free(csc->pbuf);
+	csc->pbuf = NULL;
 	return (1);
 }
 
@@ -368,18 +370,19 @@ static int
 run_cscope(SCR *sp, CSC *csc, char *dbname)
 {
 	int to_cs[2], from_cs[2];
-	char cmd[MAXPATHLEN * 2];
+	char *cmd;
 
 	/*
 	 * Cscope reads from to_cs[0] and writes to from_cs[1]; vi reads from
 	 * from_cs[0] and writes to to_cs[1].
 	 */
-	to_cs[0] = to_cs[1] = from_cs[0] = from_cs[0] = -1;
+	to_cs[0] = to_cs[1] = from_cs[0] = from_cs[1] = -1;
 	if (pipe(to_cs) < 0 || pipe(from_cs) < 0) {
 		msgq(sp, M_SYSERR, "pipe");
 		goto err;
 	}
 	switch (csc->pid = vfork()) {
+		char *dn, *dbn;
 	case -1:
 		msgq(sp, M_SYSERR, "vfork");
 err:		if (to_cs[0] != -1)
@@ -401,11 +404,23 @@ err:		if (to_cs[0] != -1)
 		(void)close(from_cs[0]);
 
 		/* Run the cscope command. */
-#define	CSCOPE_CMD_FMT		"cd '%s' && exec cscope -dl -f %s"
-		(void)snprintf(cmd, sizeof(cmd),
-		    CSCOPE_CMD_FMT, csc->dname, dbname);
+#define	CSCOPE_CMD_FMT		"cd %s && exec cscope -dl -f %s"
+		if ((dn = quote(csc->dname)) == NULL)
+			goto nomem;
+		if ((dbn = quote(dbname)) == NULL) {
+			free(dn);
+			goto nomem;
+		}
+		(void)asprintf(&cmd, CSCOPE_CMD_FMT, dn, dbn);
+		free(dbn);
+		free(dn);
+		if (cmd == NULL) {
+nomem:			msgq(sp, M_SYSERR, NULL);
+			_exit (1);
+		}
 		(void)execl(_PATH_BSHELL, "sh", "-c", cmd, (char *)NULL);
 		msgq_str(sp, M_SYSERR, cmd, "execl: %s");
+		free(cmd);
 		_exit (127);
 		/* NOTREACHED */
 	default:			/* parent. */
@@ -438,7 +453,7 @@ cscope_find(SCR *sp, EXCMD *cmdp, CHAR_T *pattern)
 	FREF *frp;
 	TAGQ *rtqp, *tqp;
 	TAG *rtp;
-	db_recno_t lno;
+	recno_t lno;
 	size_t cno, search;
 	int force, istmp, matches;
 	char *np = NULL;
@@ -447,7 +462,7 @@ cscope_find(SCR *sp, EXCMD *cmdp, CHAR_T *pattern)
 	exp = EXP(sp);
 
 	/* Check for connections. */
-	if (exp->cscq.lh_first == NULL) {
+	if (SLIST_EMPTY(exp->cscq)) {
 		msgq(sp, M_ERR, "310|No cscope connections running");
 		return (1);
 	}
@@ -459,15 +474,15 @@ cscope_find(SCR *sp, EXCMD *cmdp, CHAR_T *pattern)
 	 */
 	rtp = NULL;
 	rtqp = NULL;
-	if (exp->tq.cqh_first == (void *)&exp->tq) {
+	if (TAILQ_EMPTY(exp->tq)) {
 		/* Initialize the `local context' tag queue structure. */
-		CALLOC_GOTO(sp, rtqp, TAGQ *, 1, sizeof(TAGQ));
-		CIRCLEQ_INIT(&rtqp->tagq);
+		CALLOC_GOTO(sp, rtqp, 1, sizeof(TAGQ));
+		TAILQ_INIT(rtqp->tagq);
 
 		/* Initialize and link in its tag structure. */
-		CALLOC_GOTO(sp, rtp, TAG *, 1, sizeof(TAG));
-		CIRCLEQ_INSERT_HEAD(&rtqp->tagq, rtp, q);
-		rtqp->current = rtp; 
+		CALLOC_GOTO(sp, rtp, 1, sizeof(TAG));
+		TAILQ_INSERT_HEAD(rtqp->tagq, rtp, q);
+		rtqp->current = rtp;
 	}
 
 	/* Create the cscope command. */
@@ -475,6 +490,8 @@ cscope_find(SCR *sp, EXCMD *cmdp, CHAR_T *pattern)
 	np = strdup(np);
 	if ((tqp = create_cs_cmd(sp, np, &search)) == NULL)
 		goto err;
+	free(np);
+	np = NULL;
 
 	/*
 	 * Stick the current context in a convenient place, we'll lose it
@@ -487,35 +504,30 @@ cscope_find(SCR *sp, EXCMD *cmdp, CHAR_T *pattern)
 
 	/* Search all open connections for a match. */
 	matches = 0;
-	for (csc = exp->cscq.lh_first; csc != NULL; csc = csc_next) {
-		/* Copy csc->q.lh_next here in case csc is killed. */
-		csc_next = csc->q.le_next;
-
+	/* Copy next connect here in case csc is killed. */
+	SLIST_FOREACH_SAFE(csc, exp->cscq, q, csc_next) {
 		/*
 		 * Send the command to the cscope program.  (We skip the
 		 * first two bytes of the command, because we stored the
 		 * search cscope command character and a leading space
 		 * there.)
 		 */
-		(void)fprintf(csc->to_fp, "%d%s\n", search, tqp->tag + 2);
+		(void)fprintf(csc->to_fp, "%lu%s\n", search, tqp->tag + 2);
 		(void)fflush(csc->to_fp);
 
 		/* Read the output. */
-		if (parse(sp, csc, tqp, &matches)) {
-			if (rtqp != NULL)
-				free(rtqp);
-			tagq_free(sp, tqp);
-			return (1);
-		}
+		if (parse(sp, csc, tqp, &matches))
+			goto nomatch;
 	}
 
 	if (matches == 0) {
 		msgq(sp, M_INFO, "278|No matches for query");
-		return (0);
+nomatch:	free(rtp);
+		free(rtqp);
+		tagq_free(sp, tqp);
+		return (1);
 	}
 
-	tqp->current = tqp->tagq.cqh_first;
-
 	/* Try to switch to the first tag. */
 	force = FL_ISSET(cmdp->iflags, E_C_FORCE);
 	if (F_ISSET(cmdp, E_NEWSCREEN)) {
@@ -534,13 +546,13 @@ cscope_find(SCR *sp, EXCMD *cmdp, CHAR_T *pattern)
 	 * in place, so we can pop all the way back to the current mark.
 	 * Note, it doesn't point to much of anything, it's a placeholder.
 	 */
-	if (exp->tq.cqh_first == (void *)&exp->tq) {
-		CIRCLEQ_INSERT_HEAD(&exp->tq, rtqp, q);
+	if (TAILQ_EMPTY(exp->tq)) {
+		TAILQ_INSERT_HEAD(exp->tq, rtqp, q);
 	} else
-		rtqp = exp->tq.cqh_first;
+		rtqp = TAILQ_FIRST(exp->tq);
 
 	/* Link the current TAGQ structure into place. */
-	CIRCLEQ_INSERT_HEAD(&exp->tq, tqp, q);
+	TAILQ_INSERT_HEAD(exp->tq, tqp, q);
 
 	(void)cscope_search(sp, tqp, tqp->current);
 
@@ -567,12 +579,9 @@ cscope_find(SCR *sp, EXCMD *cmdp, CHAR_T *pattern)
 
 err:
 alloc_err:
-	if (rtqp != NULL)
-		free(rtqp);
-	if (rtp != NULL)
-		free(rtp);
-	if (np != NULL)
-		free(np);
+	free(rtqp);
+	free(rtp);
+	free(np);
 	return (1);
 }
 
@@ -601,8 +610,8 @@ create_cs_cmd(SCR *sp, char *pattern, size_t *searchp)
 		goto usage;
 
 	/* Skip leading blanks, check for command character. */
-	for (; isblank(pattern[0]); ++pattern);
-	if (pattern[0] == '\0' || !isblank(pattern[1]))
+	for (; cmdskip(pattern[0]); ++pattern);
+	if (pattern[0] == '\0' || !cmdskip(pattern[1]))
 		goto usage;
 	for (*searchp = 0, p = CSCOPE_QUERIES;
 	    *p != '\0' && *p != pattern[0]; ++*searchp, ++p);
@@ -614,7 +623,7 @@ create_cs_cmd(SCR *sp, char *pattern, size_t *searchp)
 	}
 
 	/* Skip <blank> characters to the pattern. */
-	for (p = pattern + 1; *p != '\0' && isblank(*p); ++p);
+	for (p = pattern + 1; *p != '\0' && cmdskip(*p); ++p);
 	if (*p == '\0') {
 usage:		(void)csc_help(sp, "find");
 		return (NULL);
@@ -625,16 +634,16 @@ usage:		(void)csc_help(sp, "find");
 	if (p[0] == '"' && p[1] != '\0' && p[2] == '\0')
 		CBNAME(sp, cbp, p[1]);
 	if (cbp != NULL) {
-		INT2CHAR(sp, cbp->textq.cqh_first->lb,
-			cbp->textq.cqh_first->len, p, tlen);
+		INT2CHAR(sp, TAILQ_FIRST(cbp->textq)->lb,
+			TAILQ_FIRST(cbp->textq)->len, p, tlen);
 	} else
 		tlen = strlen(p);
 
 	/* Allocate and initialize the TAGQ structure. */
-	CALLOC(sp, tqp, TAGQ *, 1, sizeof(TAGQ) + tlen + 3);
+	CALLOC(sp, tqp, 1, sizeof(TAGQ) + tlen + 3);
 	if (tqp == NULL)
 		return (NULL);
-	CIRCLEQ_INIT(&tqp->tagq);
+	TAILQ_INIT(tqp->tagq);
 	tqp->tag = tqp->buf;
 	tqp->tag[0] = pattern[0];
 	tqp->tag[1] = ' ';
@@ -654,10 +663,12 @@ static int
 parse(SCR *sp, CSC *csc, TAGQ *tqp, int *matchesp)
 {
 	TAG *tp;
-	db_recno_t slno;
-	size_t dlen, nlen, slen;
-	int ch, i, isolder, nlines;
-	char *dname, *name, *search, *p, *t, dummy[2], buf[2048];
+	recno_t slno = 0;
+	size_t dlen, nlen = 0, slen = 0;
+	int ch, i, isolder = 0, nlines;
+	char *dname = NULL, *name = NULL, *search, *p, *t, dummy[2], buf[2048];
+	CHAR_T *wp;
+	size_t wlen;
 
 	for (;;) {
 		if (!fgets(buf, sizeof(buf), csc->from_fp))
@@ -706,7 +717,7 @@ parse(SCR *sp, CSC *csc, TAGQ *tqp, int *matchesp)
 			case 1:			/* Context. */
 				break;
 			case 2:			/* Line number. */
-				slno = (db_recno_t)atol(t);
+				slno = (recno_t)atol(t);
 				break;
 			}
 		if (i != 3 || p == NULL || t == NULL)
@@ -733,10 +744,12 @@ parse(SCR *sp, CSC *csc, TAGQ *tqp, int *matchesp)
 		 * Allocate and initialize a tag structure plus the variable
 		 * length cscope information that follows it.
 		 */
-		CALLOC_RET(sp, tp,
-		    TAG *, 1, sizeof(TAG) + dlen + 2 + nlen + 1 + slen + 1);
+		CALLOC_RET(sp, tp, 1,
+			   sizeof(TAG) + dlen + 2 + nlen + 1 + (slen + 1) * sizeof(CHAR_T));
 		tp->fname = (char *)tp->buf;
-		if (dlen != 0) {
+		if (dlen == 1 && *dname == '.')
+			--dlen;
+		else if (dlen != 0) {
 			memcpy(tp->fname, dname, dlen);
 			tp->fname[dlen] = '/';
 			++dlen;
@@ -746,13 +759,22 @@ parse(SCR *sp, CSC *csc, TAGQ *tqp, int *matchesp)
 		tp->slno = slno;
 		if (slen != 0) {
 			tp->search = (CHAR_T*)(tp->fname + tp->fnlen + 1);
-			MEMCPYW(tp->search, search, (tp->slen = slen) + 1);
+			CHAR2INT(sp, search, slen + 1, wp, wlen);
+			MEMCPY(tp->search, wp, (tp->slen = slen) + 1);
 		}
-		CIRCLEQ_INSERT_TAIL(&tqp->tagq, tp, q);
+		TAILQ_INSERT_TAIL(tqp->tagq, tp, q);
+
+		/* Try to preset the tag within the current file. */
+		if (sp->frp != NULL && sp->frp->name != NULL &&
+		    tqp->current == NULL && !strcmp(tp->fname, sp->frp->name))
+			tqp->current = tp;
 
 		++*matchesp;
 	}
 
+	if (tqp->current == NULL)
+		tqp->current = TAILQ_FIRST(tqp->tagq);
+
 	return read_prompt(sp, csc);
 
 io_err:	if (feof(csc->from_fp))
@@ -770,7 +792,7 @@ static void
 csc_file(SCR *sp, CSC *csc, char *name, char **dirp, size_t *dlenp, int *isolderp)
 {
 	struct stat sb;
-	char **pp, buf[MAXPATHLEN];
+	char **pp, *buf;
 
 	/*
 	 * Check for the file in all of the listed paths.  If we don't
@@ -780,13 +802,20 @@ csc_file(SCR *sp, CSC *csc, char *name, char **dirp, size_t *dlenp, int *isolder
 	 * lives.
 	 */
 	for (pp = csc->paths; *pp != NULL; ++pp) {
-		(void)snprintf(buf, sizeof(buf), "%s/%s", *pp, name);
-		if (stat(buf, &sb) == 0) {
-			*dirp = *pp;
-			*dlenp = strlen(*pp);
-			*isolderp = sb.st_mtime < csc->mtime;
+		if ((buf = join(*pp, name)) == NULL) {
+			msgq(sp, M_SYSERR, NULL);
+			*dlenp = 0;
 			return;
 		}
+		if (stat(buf, &sb) == 0) {
+			free(buf);
+			*dirp = *pp;
+			*dlenp = strlen(*pp);
+			*isolderp = timespeccmp(
+			    &sb.st_mtimespec, &csc->mtim, <);
+			return;
+		}
+		free(buf);
 	}
 	*dlenp = 0;
 }
@@ -821,7 +850,7 @@ csc_help(SCR *sp, char *cmd)
 			return (1);
 		} else {
 			ex_printf(sp,
-		          "Command: %s (%s)\n", ccp->name, ccp->help_msg);
+			  "Command: %s (%s)\n", ccp->name, ccp->help_msg);
 			ex_printf(sp, "  Usage: %s\n", ccp->usage_msg);
 			return (0);
 		}
@@ -841,9 +870,13 @@ cscope_kill(SCR *sp, EXCMD *cmdp, CHAR_T *cn)
 {
 	char *np;
 	size_t nlen;
+	int n = 1;
 
-	INT2CHAR(sp, cn, STRLEN(cn) + 1, np, nlen);
-	return (terminate(sp, NULL, atoi(np)));
+	if (*cn) {
+		INT2CHAR(sp, cn, STRLEN(cn) + 1, np, nlen);
+		n = atoi(np);
+	}
+	return (terminate(sp, NULL, n));
 }
 
 /*
@@ -854,47 +887,53 @@ static int
 terminate(SCR *sp, CSC *csc, int n)
 {
 	EX_PRIVATE *exp;
-	int i, pstat;
+	int i = 0, pstat;
+	CSC *cp, *pre_cp = NULL;
 
 	exp = EXP(sp);
 
 	/*
-	 * We either get a csc structure or a number.  If not provided a
-	 * csc structure, find the right one.
+	 * We either get a csc structure or a number.  Locate and remove
+	 * the candidate which matches the structure or the number.
 	 */
-	if (csc == NULL) {
-		if (n < 1)
-			goto badno;
-		for (i = 1, csc = exp->cscq.lh_first;
-		    csc != NULL; csc = csc->q.le_next, i++)
-			if (i == n)
-				break;
-		if (csc == NULL) {
-badno:			msgq(sp, M_ERR, "312|%d: no such cscope session", n);
-			return (1);
+	if (csc == NULL && n < 1)
+		goto badno;
+	SLIST_FOREACH(cp, exp->cscq, q) {
+		++i;
+		if (csc == NULL ? i != n : cp != csc) {
+			pre_cp = cp;
+			continue;
 		}
+		if (cp == SLIST_FIRST(exp->cscq))
+			SLIST_REMOVE_HEAD(exp->cscq, q);
+		else
+			SLIST_REMOVE_AFTER(pre_cp, q);
+		csc = cp;
+		break;
+	}
+	if (csc == NULL) {
+badno:		msgq(sp, M_ERR, "312|%d: no such cscope session", n);
+		return (1);
 	}
 
 	/*
 	 * XXX
 	 * Theoretically, we have the only file descriptors to the process,
 	 * so closing them should let it exit gracefully, deleting temporary
-	 * files, etc.  The original vi cscope integration sent the cscope
-	 * connection a SIGTERM signal, so I'm not sure if closing the file
-	 * descriptors is sufficient.
+	 * files, etc.  However, the earlier created cscope processes seems
+	 * to refuse to quit unless we send a SIGTERM signal.
 	 */
 	if (csc->from_fp != NULL)
 		(void)fclose(csc->from_fp);
 	if (csc->to_fp != NULL)
 		(void)fclose(csc->to_fp);
+	if (i > 1)
+		(void)kill(csc->pid, SIGTERM);
 	(void)waitpid(csc->pid, &pstat, 0);
 
 	/* Discard cscope connection information. */
-	LIST_REMOVE(csc, q);
-	if (csc->pbuf != NULL)
-		free(csc->pbuf);
-	if (csc->paths != NULL)
-		free(csc->paths);
+	free(csc->pbuf);
+	free(csc->paths);
 	free(csc);
 	return (0);
 }
@@ -905,14 +944,24 @@ badno:			msgq(sp, M_ERR, "312|%d: no such cscope session", n);
  */
 static int
 cscope_reset(SCR *sp, EXCMD *cmdp, CHAR_T *notusedp)
+{
+	return cscope_end(sp);
+}
+
+/*
+ * cscope_end --
+ *	End all cscope connections.
+ *
+ * PUBLIC: int cscope_end(SCR *);
+ */
+int
+cscope_end(SCR *sp)
 {
 	EX_PRIVATE *exp;
 
-	for (exp = EXP(sp); exp->cscq.lh_first != NULL;) {
-		static CHAR_T one[] = {'1', 0};
-		if (cscope_kill(sp, cmdp, one))
+	for (exp = EXP(sp); !SLIST_EMPTY(exp->cscq);)
+		if (terminate(sp, NULL, 1))
 			return (1);
-	}
 	return (0);
 }
 
@@ -920,24 +969,23 @@ cscope_reset(SCR *sp, EXCMD *cmdp, CHAR_T *notusedp)
  * cscope_display --
  *	Display current connections.
  *
- * PUBLIC: int cscope_display __P((SCR *));
+ * PUBLIC: int cscope_display(SCR *);
  */
 int
 cscope_display(SCR *sp)
 {
 	EX_PRIVATE *exp;
 	CSC *csc;
-	int i;
+	int i = 0;
 
 	exp = EXP(sp);
-	if (exp->cscq.lh_first == NULL) {
+	if (SLIST_EMPTY(exp->cscq)) {
 		ex_printf(sp, "No cscope connections.\n");
 		return (0);
 	}
-	for (i = 1,
-	    csc = exp->cscq.lh_first; csc != NULL; ++i, csc = csc->q.le_next)
-		ex_printf(sp,
-		    "%2d %s (process %lu)\n", i, csc->dname, (u_long)csc->pid);
+	SLIST_FOREACH(csc, exp->cscq, q)
+		ex_printf(sp, "%2d %s (process %lu)\n",
+		    ++i, csc->dname, (u_long)csc->pid);
 	return (0);
 }
 
@@ -945,7 +993,7 @@ cscope_display(SCR *sp)
  * cscope_search --
  *	Search a file for a cscope entry.
  *
- * PUBLIC: int cscope_search __P((SCR *, TAGQ *, TAG *));
+ * PUBLIC: int cscope_search(SCR *, TAGQ *, TAG *);
  */
 int
 cscope_search(SCR *sp, TAGQ *tqp, TAG *tp)
@@ -967,7 +1015,7 @@ cscope_search(SCR *sp, TAGQ *tqp, TAG *tp)
 		m.lno = 1;
 		m.cno = 0;
 		if (f_search(sp, &m, &m,
-		    tp->search, tp->slen, NULL, SEARCH_CSCOPE | SEARCH_FIRST)) {
+		    tp->search, tp->slen, NULL, SEARCH_CSCOPE | SEARCH_FILE)) {
 			tag_msg(sp, TAG_SEARCH, tqp->tag);
 			return (1);
 		}
