@@ -9,12 +9,9 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "@(#)ex_global.c	10.21 (Berkeley) 6/30/96";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
+#include <sys/time.h>
 
 #include <bitstring.h>
 #include <ctype.h>
@@ -29,18 +26,16 @@ static const char sccsid[] = "@(#)ex_global.c	10.21 (Berkeley) 6/30/96";
 
 enum which {GLOBAL, V};
 
-static int ex_g_setup __P((SCR *, EXCMD *, enum which));
+static int ex_g_setup(SCR *, EXCMD *, enum which);
 
 /*
  * ex_global -- [line [,line]] g[lobal][!] /pattern/ [commands]
  *	Exec on lines matching a pattern.
  *
- * PUBLIC: int ex_global __P((SCR *, EXCMD *));
+ * PUBLIC: int ex_global(SCR *, EXCMD *);
  */
 int
-ex_global(sp, cmdp)
-	SCR *sp;
-	EXCMD *cmdp;
+ex_global(SCR *sp, EXCMD *cmdp)
 {
 	return (ex_g_setup(sp,
 	    cmdp, FL_ISSET(cmdp->iflags, E_C_FORCE) ? V : GLOBAL));
@@ -50,12 +45,10 @@ ex_global(sp, cmdp)
  * ex_v -- [line [,line]] v /pattern/ [commands]
  *	Exec on lines not matching a pattern.
  *
- * PUBLIC: int ex_v __P((SCR *, EXCMD *));
+ * PUBLIC: int ex_v(SCR *, EXCMD *);
  */
 int
-ex_v(sp, cmdp)
-	SCR *sp;
-	EXCMD *cmdp;
+ex_v(SCR *sp, EXCMD *cmdp)
 {
 	return (ex_g_setup(sp, cmdp, V));
 }
@@ -65,11 +58,9 @@ ex_v(sp, cmdp)
  *	Ex global and v commands.
  */
 static int
-ex_g_setup(sp, cmdp, cmd)
-	SCR *sp;
-	EXCMD *cmdp;
-	enum which cmd;
+ex_g_setup(SCR *sp, EXCMD *cmdp, enum which cmd)
 {
+	CHAR_T *ptrn, *p, *t;
 	EXCMD *ecp;
 	MARK abs;
 	RANGE *rp;
@@ -79,14 +70,13 @@ ex_g_setup(sp, cmdp, cmd)
 	regmatch_t match[1];
 	size_t len;
 	int cnt, delim, eval;
-	char *ptrn, *p, *t;
+	CHAR_T *dbp;
 
 	NEEDFILE(sp, cmdp);
 
 	if (F_ISSET(sp, SC_EX_GLOBAL)) {
-		msgq(sp, M_ERR,
-	"124|The %s command can't be used as part of a global or v command",
-		    cmdp->cmd->name);
+		msgq_wstr(sp, M_ERR, cmdp->cmd->name,
+	"124|The %s command can't be used as part of a global or v command");
 		return (1);
 	}
 
@@ -96,8 +86,8 @@ ex_g_setup(sp, cmdp, cmd)
 	 */
 	if (cmdp->argc == 0)
 		goto usage;
-	for (p = cmdp->argv[0]->bp; isblank(*p); ++p);
-	if (*p == '\0' || isalnum(*p) ||
+	for (p = cmdp->argv[0]->bp; cmdskip(*p); ++p);
+	if (*p == '\0' || is09azAZ(*p) ||
 	    *p == '\\' || *p == '|' || *p == '\n') {
 usage:		ex_emsg(sp, cmdp->cmd->usage, EXM_USAGE);
 		return (1);
@@ -138,13 +128,14 @@ usage:		ex_emsg(sp, cmdp->cmd->usage, EXM_USAGE);
 		}
 
 		/* Re-compile the RE if necessary. */
-		if (!F_ISSET(sp, SC_RE_SEARCH) && re_compile(sp,
-		    sp->re, sp->re_len, NULL, NULL, &sp->re_c, RE_C_SEARCH))
+		if (!F_ISSET(sp, SC_RE_SEARCH) &&
+		    re_compile(sp, sp->re, sp->re_len,
+		    NULL, NULL, &sp->re_c, RE_C_SEARCH))
 			return (1);
 	} else {
 		/* Compile the RE. */
-		if (re_compile(sp, ptrn, t - ptrn,
-		    &sp->re, &sp->re_len, &sp->re_c, RE_C_SEARCH))
+		if (re_compile(sp, ptrn, t - ptrn, &sp->re,
+		    &sp->re_len, &sp->re_c, RE_C_SEARCH))
 			return (1);
 
 		/*
@@ -162,8 +153,8 @@ usage:		ex_emsg(sp, cmdp->cmd->usage, EXM_USAGE);
 		return (1);
 
 	/* Get an EXCMD structure. */
-	CALLOC_RET(sp, ecp, EXCMD *, 1, sizeof(EXCMD));
-	CIRCLEQ_INIT(&ecp->rq);
+	CALLOC_RET(sp, ecp, 1, sizeof(EXCMD));
+	TAILQ_INIT(ecp->rq);
 
 	/*
 	 * Get a copy of the command string; the default command is print.
@@ -172,18 +163,18 @@ usage:		ex_emsg(sp, cmdp->cmd->usage, EXM_USAGE);
 	 * because the ex parser may step on the command string when it's
 	 * parsing it.
 	 */
-	if ((len = strlen(p)) == 0) {
-		p = "pp";
+	if ((len = cmdp->argv[0]->len - (p - cmdp->argv[0]->bp)) == 0) {
+		p = L("p");
 		len = 1;
 	}
 
-	MALLOC_RET(sp, ecp->cp, char *, len * 2);
+	MALLOC_RET(sp, ecp->cp, (len * 2) * sizeof(CHAR_T));
 	ecp->o_cp = ecp->cp;
 	ecp->o_clen = len;
-	memcpy(ecp->cp + len, p, len);
+	MEMCPY(ecp->cp + len, p, len);
 	ecp->range_lno = OOBLNO;
 	FL_SET(ecp->agv_flags, cmd == GLOBAL ? AGV_GLOBAL : AGV_V);
-	LIST_INSERT_HEAD(&sp->gp->ecq, ecp, q);
+	SLIST_INSERT_HEAD(sp->gp->ecq, ecp, q);
 
 	/*
 	 * For each line...  The semantics of global matching are that we first
@@ -203,7 +194,7 @@ usage:		ex_emsg(sp, cmdp->cmd->usage, EXM_USAGE);
 	    end = cmdp->addr2.lno; start <= end; ++start) {
 		if (cnt-- == 0) {
 			if (INTERRUPTED(sp)) {
-				LIST_REMOVE(ecp, q);
+				SLIST_REMOVE_HEAD(sp->gp->ecq, q);
 				free(ecp->cp);
 				free(ecp);
 				break;
@@ -212,11 +203,12 @@ usage:		ex_emsg(sp, cmdp->cmd->usage, EXM_USAGE);
 			btype = BUSY_UPDATE;
 			cnt = INTERRUPT_CHECK;
 		}
-		if (db_get(sp, start, DBG_FATAL, &p, &len))
+		if (db_get(sp, start, DBG_FATAL, &dbp, &len))
 			return (1);
 		match[0].rm_so = 0;
 		match[0].rm_eo = len;
-		switch (eval = regexec(&sp->re_c, p, 0, match, REG_STARTEND)) {
+		switch (eval =
+		    regexec(&sp->re_c, dbp, 0, match, REG_STARTEND)) {
 		case 0:
 			if (cmd == V)
 				continue;
@@ -231,18 +223,18 @@ usage:		ex_emsg(sp, cmdp->cmd->usage, EXM_USAGE);
 		}
 
 		/* If follows the last entry, extend the last entry's range. */
-		if ((rp = ecp->rq.cqh_last) != (void *)&ecp->rq &&
+		if ((rp = TAILQ_LAST(ecp->rq, _rh)) != NULL &&
 		    rp->stop == start - 1) {
 			++rp->stop;
 			continue;
 		}
 
 		/* Allocate a new range, and append it to the list. */
-		CALLOC(sp, rp, RANGE *, 1, sizeof(RANGE));
+		CALLOC(sp, rp, 1, sizeof(RANGE));
 		if (rp == NULL)
 			return (1);
 		rp->start = rp->stop = start;
-		CIRCLEQ_INSERT_TAIL(&ecp->rq, rp, q);
+		TAILQ_INSERT_TAIL(ecp->rq, rp, q);
 	}
 	search_busy(sp, BUSY_OFF);
 	return (0);
@@ -252,13 +244,10 @@ usage:		ex_emsg(sp, cmdp->cmd->usage, EXM_USAGE);
  * ex_g_insdel --
  *	Update the ranges based on an insertion or deletion.
  *
- * PUBLIC: int ex_g_insdel __P((SCR *, lnop_t, recno_t));
+ * PUBLIC: int ex_g_insdel(SCR *, lnop_t, recno_t);
  */
 int
-ex_g_insdel(sp, op, lno)
-	SCR *sp;
-	lnop_t op;
-	recno_t lno;
+ex_g_insdel(SCR *sp, lnop_t op, recno_t lno)
 {
 	EXCMD *ecp;
 	RANGE *nrp, *rp;
@@ -270,12 +259,10 @@ ex_g_insdel(sp, op, lno)
 	if (op == LINE_RESET)
 		return (0);
 
-	for (ecp = sp->gp->ecq.lh_first; ecp != NULL; ecp = ecp->q.le_next) {
+	SLIST_FOREACH(ecp, sp->gp->ecq, q) {
 		if (!FL_ISSET(ecp->agv_flags, AGV_AT | AGV_GLOBAL | AGV_V))
 			continue;
-		for (rp = ecp->rq.cqh_first; rp != (void *)&ecp->rq; rp = nrp) {
-			nrp = rp->q.cqe_next;
-
+		TAILQ_FOREACH_SAFE(rp, ecp->rq, q, nrp) {
 			/* If range less than the line, ignore it. */
 			if (rp->stop < lno)
 				continue;
@@ -303,16 +290,15 @@ ex_g_insdel(sp, op, lno)
 			 */
 			if (op == LINE_DELETE) {
 				if (rp->start > --rp->stop) {
-					CIRCLEQ_REMOVE(&ecp->rq, rp, q);
+					TAILQ_REMOVE(ecp->rq, rp, q);
 					free(rp);
 				}
 			} else {
-				CALLOC_RET(sp, nrp, RANGE *, 1, sizeof(RANGE));
+				CALLOC_RET(sp, nrp, 1, sizeof(RANGE));
 				nrp->start = lno + 1;
 				nrp->stop = rp->stop + 1;
 				rp->stop = lno - 1;
-				CIRCLEQ_INSERT_AFTER(&ecp->rq, rp, nrp, q);
-				rp = nrp;
+				TAILQ_INSERT_AFTER(ecp->rq, rp, nrp, q);
 			}
 		}
 
