@@ -9,10 +9,6 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: ex_subst.c,v 10.50 2002/02/09 21:18:23 skimo Exp $ (Berkeley) $Date: 2002/02/09 21:18:23 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/time.h>
@@ -32,12 +28,12 @@ static const char sccsid[] = "$Id: ex_subst.c,v 10.50 2002/02/09 21:18:23 skimo
 #define	SUB_FIRST	0x01		/* The 'r' flag isn't reasonable. */
 #define	SUB_MUSTSETR	0x02		/* The 'r' flag is required. */
 
-static int re_conv __P((SCR *, CHAR_T **, size_t *, int *));
-static int re_cscope_conv __P((SCR *, CHAR_T **, size_t *, int *));
-static int re_sub __P((SCR *,
-		CHAR_T *, CHAR_T **, size_t *, size_t *, regmatch_t [10]));
-static int re_tag_conv __P((SCR *, CHAR_T **, size_t *, int *));
-static int s __P((SCR *, EXCMD *, CHAR_T *, regex_t *, u_int));
+static int re_conv(SCR *, CHAR_T **, size_t *, int *);
+static int re_cscope_conv(SCR *, CHAR_T **, size_t *, int *);
+static int re_sub(SCR *,
+		CHAR_T *, CHAR_T **, size_t *, size_t *, regmatch_t [10]);
+static int re_tag_conv(SCR *, CHAR_T **, size_t *, int *);
+static int s(SCR *, EXCMD *, CHAR_T *, regex_t *, u_int);
 
 /*
  * ex_s --
@@ -45,7 +41,7 @@ static int s __P((SCR *, EXCMD *, CHAR_T *, regex_t *, u_int));
  *
  *	Substitute on lines matching a pattern.
  *
- * PUBLIC: int ex_s __P((SCR *, EXCMD *));
+ * PUBLIC: int ex_s(SCR *, EXCMD *);
  */
 int
 ex_s(SCR *sp, EXCMD *cmdp)
@@ -71,14 +67,14 @@ ex_s(SCR *sp, EXCMD *cmdp)
 		goto subagain;
 	for (p = cmdp->argv[0]->bp,
 	    len = cmdp->argv[0]->len; len > 0; --len, ++p) {
-		if (!isblank(*p))
+		if (!cmdskip(*p))
 			break;
 	}
 	if (len == 0)
 subagain:	return (ex_subagain(sp, cmdp));
 
 	delim = *p++;
-	if (isalnum(delim) || delim == '\\')
+	if (is09azAZ(delim) || delim == '\\')
 		return (s(sp, cmdp, p, &sp->subre_c, SUB_MUSTSETR));
 
 	/*
@@ -140,7 +136,7 @@ subagain:	return (ex_subagain(sp, cmdp));
 		/* Re-compile the RE if necessary. */
 		if (!F_ISSET(sp, SC_RE_SEARCH) &&
 		    re_compile(sp, sp->re, sp->re_len,
-		    NULL, NULL, &sp->re_c, SEARCH_CSEARCH | SEARCH_MSG))
+		    NULL, NULL, &sp->re_c, RE_C_SEARCH))
 			return (1);
 		flags = 0;
 	} else {
@@ -152,10 +148,10 @@ subagain:	return (ex_subagain(sp, cmdp));
 		 * ref counting the pattern string and (opaque) structure.
 		 */
 		if (re_compile(sp, ptrn, t - ptrn, &sp->re,
-		    &sp->re_len, &sp->re_c, SEARCH_CSEARCH | SEARCH_MSG))
+		    &sp->re_len, &sp->re_c, RE_C_SEARCH))
 			return (1);
 		if (re_compile(sp, ptrn, t - ptrn, &sp->subre,
-		    &sp->subre_len, &sp->subre_c, SEARCH_CSUBST | SEARCH_MSG))
+		    &sp->subre_len, &sp->subre_c, RE_C_SUBST))
 			return (1);
 		
 		flags = SUB_FIRST;
@@ -187,41 +183,40 @@ subagain:	return (ex_subagain(sp, cmdp));
 	 * V and then percolated elsewhere, presumably around the time
 	 * that it was added to their version of ed(1).
 	 */
-	if (p[0] == L('\0') || p[0] == delim) {
+	if (p[0] == '\0' || p[0] == delim) {
 		if (p[0] == delim)
 			++p;
-		if (sp->repl != NULL)
-			free(sp->repl);
+		free(sp->repl);
 		sp->repl = NULL;
 		sp->repl_len = 0;
-	} else if (p[0] == L('%') && (p[1] == L('\0') || p[1] == delim))
+	} else if (p[0] == '%' && (p[1] == '\0' || p[1] == delim))
 		p += p[1] == delim ? 2 : 1;
 	else {
 		for (rep = p, len = 0;
-		    p[0] != L('\0') && p[0] != delim; ++p, ++len)
-			if (p[0] == L('~'))
+		    p[0] != '\0' && p[0] != delim; ++p, ++len)
+			if (p[0] == '~')
 				len += sp->repl_len;
 		GET_SPACE_RETW(sp, bp, blen, len);
 		for (t = bp, len = 0, p = rep;;) {
-			if (p[0] == L('\0') || p[0] == delim) {
+			if (p[0] == '\0' || p[0] == delim) {
 				if (p[0] == delim)
 					++p;
 				break;
 			}
-			if (p[0] == L('\\')) {
+			if (p[0] == '\\') {
 				if (p[1] == delim)
 					++p;
-				else if (p[1] == L('\\')) {
+				else if (p[1] == '\\') {
 					*t++ = *p++;
 					++len;
-				} else if (p[1] == L('~')) {
+				} else if (p[1] == '~') {
 					++p;
 					if (!O_ISSET(sp, O_MAGIC))
 						goto tilde;
 				}
-			} else if (p[0] == L('~') && O_ISSET(sp, O_MAGIC)) {
+			} else if (p[0] == '~' && O_ISSET(sp, O_MAGIC)) {
 tilde:				++p;
-				MEMCPYW(t, sp->repl, sp->repl_len);
+				MEMCPY(t, sp->repl, sp->repl_len);
 				t += sp->repl_len;
 				len += sp->repl_len;
 				continue;
@@ -230,14 +225,13 @@ tilde:				++p;
 			++len;
 		}
 		if ((sp->repl_len = len) != 0) {
-			if (sp->repl != NULL)
-				free(sp->repl);
-			if ((sp->repl = malloc(len * sizeof(CHAR_T))) == NULL) {
-				msgq(sp, M_SYSERR, NULL);
+			free(sp->repl);
+			MALLOC(sp, sp->repl, len * sizeof(CHAR_T));
+			if (sp->repl == NULL) {
 				FREE_SPACEW(sp, bp, blen);
 				return (1);
 			}
-			MEMCPYW(sp->repl, bp, len);
+			MEMCPY(sp->repl, bp, len);
 		}
 		FREE_SPACEW(sp, bp, blen);
 	}
@@ -250,7 +244,7 @@ tilde:				++p;
  *
  *	Substitute using the last substitute RE and replacement pattern.
  *
- * PUBLIC: int ex_subagain __P((SCR *, EXCMD *));
+ * PUBLIC: int ex_subagain(SCR *, EXCMD *);
  */
 int
 ex_subagain(SCR *sp, EXCMD *cmdp)
@@ -261,7 +255,7 @@ ex_subagain(SCR *sp, EXCMD *cmdp)
 	}
 	if (!F_ISSET(sp, SC_RE_SUBST) &&
 	    re_compile(sp, sp->subre, sp->subre_len,
-	    NULL, NULL, &sp->subre_c, SEARCH_CSUBST | SEARCH_MSG))
+	    NULL, NULL, &sp->subre_c, RE_C_SUBST))
 		return (1);
 	return (s(sp,
 	    cmdp, cmdp->argc ? cmdp->argv[0]->bp : NULL, &sp->subre_c, 0));
@@ -273,7 +267,7 @@ ex_subagain(SCR *sp, EXCMD *cmdp)
  *
  *	Substitute using the last RE and last substitute replacement pattern.
  *
- * PUBLIC: int ex_subtilde __P((SCR *, EXCMD *));
+ * PUBLIC: int ex_subtilde(SCR *, EXCMD *);
  */
 int
 ex_subtilde(SCR *sp, EXCMD *cmdp)
@@ -283,7 +277,7 @@ ex_subtilde(SCR *sp, EXCMD *cmdp)
 		return (1);
 	}
 	if (!F_ISSET(sp, SC_RE_SEARCH) && re_compile(sp, sp->re,
-	    sp->re_len, NULL, NULL, &sp->re_c, SEARCH_CSEARCH | SEARCH_MSG))
+	    sp->re_len, NULL, NULL, &sp->re_c, RE_C_SEARCH))
 		return (1);
 	return (s(sp,
 	    cmdp, cmdp->argc ? cmdp->argv[0]->bp : NULL, &sp->re_c, 0));
@@ -316,20 +310,20 @@ ex_subtilde(SCR *sp, EXCMD *cmdp)
 
 #define	BUILD(sp, l, len) {						\
 	if (lbclen + (len) > lblen) {					\
-		lblen += MAX(lbclen + (len), 256);			\
+		lblen = p2roundup(MAX(lbclen + (len), 256));		\
 		REALLOC(sp, lb, CHAR_T *, lblen * sizeof(CHAR_T));	\
 		if (lb == NULL) {					\
 			lbclen = 0;					\
 			return (1);					\
 		}							\
 	}								\
-	MEMCPYW(lb + lbclen, l, len);					\
+	MEMCPY(lb + lbclen, l, len);					\
 	lbclen += len;							\
 }
 
 #define	NEEDSP(sp, len, pnt) {						\
 	if (lbclen + (len) > lblen) {					\
-		lblen += MAX(lbclen + (len), 256);			\
+		lblen = p2roundup(MAX(lbclen + (len), 256));		\
 		REALLOC(sp, lb, CHAR_T *, lblen * sizeof(CHAR_T));	\
 		if (lb == NULL) {					\
 			lbclen = 0;					\
@@ -344,8 +338,8 @@ s(SCR *sp, EXCMD *cmdp, CHAR_T *s, regex_t *re, u_int flags)
 {
 	EVENT ev;
 	MARK from, to;
-	TEXTH tiq;
-	db_recno_t elno, lno, slno;
+	TEXTH tiq[] = {{ 0 }};
+	recno_t elno, lno, slno;
 	u_long ul;
 	regmatch_t match[10];
 	size_t blen, cnt, last, lbclen, lblen, len, llen;
@@ -353,7 +347,7 @@ s(SCR *sp, EXCMD *cmdp, CHAR_T *s, regex_t *re, u_int flags)
 	int cflag, lflag, nflag, pflag, rflag;
 	int didsub, do_eol_match, eflags, empty_ok, eval;
 	int linechanged, matched, quit, rval;
-	CHAR_T *p, *lb, *bp;
+	CHAR_T *bp, *lb;
 	enum nresult nret;
 
 	NEEDFILE(sp, cmdp);
@@ -407,7 +401,7 @@ s(SCR *sp, EXCMD *cmdp, CHAR_T *s, regex_t *re, u_int flags)
 			if (lno != OOBLNO)
 				goto usage;
 			errno = 0;
-			nret = nget_uslong(sp, &ul, s, &s, 10);
+			nret = nget_uslong(&ul, s, &s, 10);
 			lno = ul;
 			if (*s == '\0')		/* Loop increment correction. */
 				--s;
@@ -437,10 +431,8 @@ s(SCR *sp, EXCMD *cmdp, CHAR_T *s, regex_t *re, u_int flags)
 			sp->c_suffix = !sp->c_suffix;
 
 			/* Ex text structure initialization. */
-			if (F_ISSET(sp, SC_EX)) {
-				memset(&tiq, 0, sizeof(TEXTH));
-				CIRCLEQ_INIT(&tiq);
-			}
+			if (F_ISSET(sp, SC_EX))
+				TAILQ_INIT(tiq);
 			break;
 		case 'g':
 			sp->g_suffix = !sp->g_suffix;
@@ -468,7 +460,7 @@ s(SCR *sp, EXCMD *cmdp, CHAR_T *s, regex_t *re, u_int flags)
 			goto usage;
 		}
 
-	if (*s != '\0' || !rflag && LF_ISSET(SUB_MUSTSETR)) {
+	if (*s != '\0' || (!rflag && LF_ISSET(SUB_MUSTSETR))) {
 usage:		ex_emsg(sp, cmdp->cmd->usage, EXM_USAGE);
 		return (1);
 	}
@@ -511,7 +503,7 @@ noargs:	if (F_ISSET(sp, SC_VI) && sp->c_suffix && (lflag || nflag || pflag)) {
 				GET_SPACE_RETW(sp, bp, blen, llen);
 			} else
 				ADD_SPACE_RETW(sp, bp, blen, llen);
-			MEMCPYW(bp, s, llen);
+			MEMCPY(bp, s, llen);
 			s = bp;
 		}
 
@@ -653,9 +645,9 @@ nextmatch:	match[0].rm_so = 0;
 				if (ex_print(sp, cmdp, &from, &to, 0) ||
 				    ex_scprint(sp, &from, &to))
 					goto lquit;
-				if (ex_txt(sp, &tiq, 0, TXT_CR))
+				if (ex_txt(sp, tiq, 0, TXT_CR))
 					goto err;
-				ev.e_c = tiq.cqh_first->lb[0];
+				ev.e_c = TAILQ_FIRST(tiq)->lb[0];
 			}
 
 			switch (ev.e_c) {
@@ -745,7 +737,7 @@ skip:		offset += match[0].rm_eo;
 			if (db_get(sp, lno, DBG_FATAL, &s, &llen))
 				goto err;
 			ADD_SPACE_RETW(sp, bp, blen, llen)
-			MEMCPYW(bp, s, llen);
+			MEMCPY(bp, s, llen);
 			s = bp;
 			len = llen - offset;
 
@@ -872,8 +864,7 @@ err:		rval = 1;
 
 	if (bp != NULL)
 		FREE_SPACEW(sp, bp, blen);
-	if (lb != NULL)
-		free(lb);
+	free(lb);
 	return (rval);
 }
 
@@ -881,8 +872,8 @@ err:		rval = 1;
  * re_compile --
  *	Compile the RE.
  *
- * PUBLIC: int re_compile __P((SCR *,
- * PUBLIC:     CHAR_T *, size_t, CHAR_T **, size_t *, regex_t *, u_int));
+ * PUBLIC: int re_compile(SCR *,
+ * PUBLIC:     CHAR_T *, size_t, CHAR_T **, size_t *, regex_t *, u_int);
  */
 int
 re_compile(SCR *sp, CHAR_T *ptrn, size_t plen, CHAR_T **ptrnp, size_t *lenp, regex_t *rep, u_int flags)
@@ -893,34 +884,26 @@ re_compile(SCR *sp, CHAR_T *ptrn, size_t plen, CHAR_T **ptrnp, size_t *lenp, reg
 
 	/* Set RE flags. */
 	reflags = 0;
-	if (LF_ISSET(SEARCH_EXTEND))
-		reflags |= REG_EXTENDED;
-	if (LF_ISSET(SEARCH_IC))
-		reflags |= REG_ICASE;
-	if (LF_ISSET(SEARCH_LITERAL))
-		reflags |= REG_NOSPEC;
-	if (!LF_ISSET(SEARCH_NOOPT | SEARCH_CSCOPE | SEARCH_TAG)) {
+	if (!LF_ISSET(RE_C_CSCOPE | RE_C_TAG)) {
 		if (O_ISSET(sp, O_EXTENDED))
 			reflags |= REG_EXTENDED;
 		if (O_ISSET(sp, O_IGNORECASE))
 			reflags |= REG_ICASE;
-		if (O_ISSET(sp, O_ICLOWER))
-			goto iclower;
-	}
-	if (LF_ISSET(SEARCH_ICL)) {
-iclower:	for (p = ptrn, len = plen; len > 0; ++p, --len)
-			if (isupper(*p))
-				break;
-		if (len == 0)
-			reflags |= REG_ICASE;
+		if (O_ISSET(sp, O_ICLOWER)) {
+			for (p = ptrn, len = plen; len > 0; ++p, --len)
+				if (ISUPPER(*p))
+					break;
+			if (len == 0)
+				reflags |= REG_ICASE;
+		}
 	}
 
 	/* If we're replacing a saved value, clear the old one. */
-	if (LF_ISSET(SEARCH_CSEARCH) && F_ISSET(sp, SC_RE_SEARCH)) {
+	if (LF_ISSET(RE_C_SEARCH) && F_ISSET(sp, SC_RE_SEARCH)) {
 		regfree(&sp->re_c);
 		F_CLR(sp, SC_RE_SEARCH);
 	}
-	if (LF_ISSET(SEARCH_CSUBST) && F_ISSET(sp, SC_RE_SUBST)) {
+	if (LF_ISSET(RE_C_SUBST) && F_ISSET(sp, SC_RE_SUBST)) {
 		regfree(&sp->subre_c);
 		F_CLR(sp, SC_RE_SUBST);
 	}
@@ -932,7 +915,7 @@ iclower:	for (p = ptrn, len = plen; len > 0; ++p, --len)
 	 */
 	if (ptrnp != NULL) {
 		replaced = 0;
-		if (LF_ISSET(SEARCH_CSCOPE)) {
+		if (LF_ISSET(RE_C_CSCOPE)) {
 			if (re_cscope_conv(sp, &ptrn, &plen, &replaced))
 				return (1);
 			/*
@@ -942,18 +925,17 @@ iclower:	for (p = ptrn, len = plen; len > 0; ++p, --len)
 			 * not be right or safe.
 			 */
 			reflags |= REG_EXTENDED;
-		} else if (LF_ISSET(SEARCH_TAG)) {
+		} else if (LF_ISSET(RE_C_TAG)) {
 			if (re_tag_conv(sp, &ptrn, &plen, &replaced))
 				return (1);
-		} else if (!LF_ISSET(SEARCH_LITERAL))
+		} else
 			if (re_conv(sp, &ptrn, &plen, &replaced))
 				return (1);
 
 		/* Discard previous pattern. */
-		if (*ptrnp != NULL) {
-			free(*ptrnp);
-			*ptrnp = NULL;
-		}
+		free(*ptrnp);
+		*ptrnp = NULL;
+
 		if (lenp != NULL)
 			*lenp = plen;
 
@@ -964,9 +946,9 @@ iclower:	for (p = ptrn, len = plen; len > 0; ++p, --len)
 		 * Regcomp isn't 8-bit clean, so the pattern is nul-terminated
 		 * for now.  There's just no other solution.  
 		 */
-		MALLOC(sp, *ptrnp, CHAR_T *, (plen + 1) * sizeof(CHAR_T));
+		MALLOC(sp, *ptrnp, (plen + 1) * sizeof(CHAR_T));
 		if (*ptrnp != NULL) {
-			MEMCPYW(*ptrnp, ptrn, plen);
+			MEMCPY(*ptrnp, ptrn, plen);
 			(*ptrnp)[plen] = '\0';
 		}
 
@@ -986,14 +968,14 @@ iclower:	for (p = ptrn, len = plen; len > 0; ++p, --len)
 	 * contained a nul.  Bummer!
 	 */
 	if ((rval = regcomp(rep, ptrn, /* plen, */ reflags)) != 0) {
-		if (LF_ISSET(SEARCH_MSG))
+		if (!LF_ISSET(RE_C_SILENT))
 			re_error(sp, rval, rep); 
 		return (1);
 	}
 
-	if (LF_ISSET(SEARCH_CSEARCH))
+	if (LF_ISSET(RE_C_SEARCH))
 		F_SET(sp, SC_RE_SEARCH);
-	if (LF_ISSET(SEARCH_CSUBST))
+	if (LF_ISSET(RE_C_SUBST))
 		F_SET(sp, SC_RE_SUBST);
 
 	return (0);
@@ -1117,7 +1099,7 @@ re_conv(SCR *sp, CHAR_T **ptrnp, size_t *plenp, int *replacedp)
 					if (O_ISSET(sp, O_MAGIC))
 						*t++ = '~';
 					else {
-						MEMCPYW(t,
+						MEMCPY(t,
 						    sp->repl, sp->repl_len);
 						t += sp->repl_len;
 					}
@@ -1138,7 +1120,7 @@ re_conv(SCR *sp, CHAR_T **ptrnp, size_t *plenp, int *replacedp)
 			break;
 		case '~':
 			if (O_ISSET(sp, O_MAGIC)) {
-				MEMCPYW(t, sp->repl, sp->repl_len);
+				MEMCPY(t, sp->repl, sp->repl_len);
 				t += sp->repl_len;
 			} else
 				*t++ = '~';
@@ -1213,7 +1195,7 @@ re_tag_conv(SCR *sp, CHAR_T **ptrnp, size_t *plenp, int *replacedp)
 		if (p[0] == '\\' && (p[1] == '/' || p[1] == '?')) {
 			++p;
 			--len;
-		} else if (strchr("^.[]$*", p[0]))
+		} else if (STRCHR(L("^.[]$*"), p[0]))
 			*t++ = '\\';
 		*t++ = *p++;
 	}
@@ -1264,20 +1246,20 @@ re_cscope_conv(SCR *sp, CHAR_T **ptrnp, size_t *plenp, int *replacedp)
 	t = bp;
 
 	*t++ = '^';
-	MEMCPYW(t, wp, wlen);
+	MEMCPY(t, wp, wlen);
 	t += wlen;
 
 	for (len = *plenp; len > 0; ++p, --len)
 		if (*p == ' ') {
-			MEMCPYW(t, wp, wlen);
+			MEMCPY(t, wp, wlen);
 			t += wlen;
 		} else {
-			if (strchr("\\^.[]$*+?()|{}", *p))
+			if (STRCHR(L("\\^.[]$*+?()|{}"), *p))
 				*t++ = '\\';
 			*t++ = *p;
 		}
 
-	MEMCPYW(t, wp, wlen);
+	MEMCPY(t, wp, wlen);
 	t += wlen;
 	*t++ = '$';
 
@@ -1290,7 +1272,7 @@ re_cscope_conv(SCR *sp, CHAR_T **ptrnp, size_t *plenp, int *replacedp)
  * re_error --
  *	Report a regular expression error.
  *
- * PUBLIC: void re_error __P((SCR *, int, regex_t *));
+ * PUBLIC: void re_error(SCR *, int, regex_t *);
  */
 void
 re_error(SCR *sp, int errcode, regex_t *preg)
@@ -1299,9 +1281,8 @@ re_error(SCR *sp, int errcode, regex_t *preg)
 	char *oe;
 
 	s = regerror(errcode, preg, "", 0);
-	if ((oe = malloc(s)) == NULL)
-		msgq(sp, M_SYSERR, NULL);
-	else {
+	MALLOC(sp, oe, s);
+	if (oe != NULL) {
 		(void)regerror(errcode, preg, oe, s);
 		msgq(sp, M_ERR, "RE error: %s", oe);
 		free(oe);
@@ -1313,12 +1294,13 @@ re_error(SCR *sp, int errcode, regex_t *preg)
  * 	Do the substitution for a regular expression.
  */
 static int
-re_sub(SCR *sp, CHAR_T *ip, CHAR_T **lbp, size_t *lbclenp, size_t *lblenp, regmatch_t *match)
-	        
-	           			/* Input line. */
-	             
-	                         
-	                     
+re_sub(
+	SCR *sp,
+	CHAR_T *ip,			/* Input line. */
+	CHAR_T **lbp,
+	size_t *lbclenp,
+	size_t *lblenp,
+	regmatch_t match[10])
 {
 	enum { C_NOTSET, C_LOWER, C_ONELOWER, C_ONEUPPER, C_UPPER } conv;
 	size_t lbclen, lblen;		/* Local copies. */
@@ -1352,8 +1334,8 @@ re_sub(SCR *sp, CHAR_T *ip, CHAR_T **lbp, size_t *lbclenp, size_t *lblenp, regma
 	 * all escaping characters.  This (hopefully) matches historic practice.
 	 */
 #define	OUTCH(ch, nltrans) {						\
-	CHAR_T __ch = (ch);						\
-	u_int __value = KEY_VAL(sp, __ch);				\
+	ARG_CHAR_T __ch = (ch);						\
+	e_key_t __value = KEY_VAL(sp, __ch);				\
 	if (nltrans && (__value == K_CR || __value == K_NL)) {		\
 		NEEDNEWLINE(sp);					\
 		sp->newl[sp->newl_cnt++] = lbclen;			\
@@ -1363,15 +1345,15 @@ re_sub(SCR *sp, CHAR_T *ip, CHAR_T **lbp, size_t *lbclenp, size_t *lblenp, regma
 			conv = C_NOTSET;				\
 			/* FALLTHROUGH */				\
 		case C_LOWER:						\
-			if (isupper(__ch))				\
-				__ch = tolower(__ch);			\
+			if (ISUPPER(__ch))				\
+				__ch = TOLOWER(__ch);			\
 			break;						\
 		case C_ONEUPPER:					\
 			conv = C_NOTSET;				\
 			/* FALLTHROUGH */				\
 		case C_UPPER:						\
-			if (islower(__ch))				\
-				__ch = toupper(__ch);			\
+			if (ISLOWER(__ch))				\
+				__ch = TOUPPER(__ch);			\
 			break;						\
 		default:						\
 			abort();					\
@@ -1406,7 +1388,7 @@ re_sub(SCR *sp, CHAR_T *ip, CHAR_T **lbp, size_t *lbclenp, size_t *lblenp, regma
 			case '5': case '6': case '7': case '8': case '9':
 				no = *rp++ - '0';
 subzero:			if (match[no].rm_so == -1 ||
-			    	    match[no].rm_eo == -1)
+				    match[no].rm_eo == -1)
 					break;
 				mlen = match[no].rm_eo - match[no].rm_so;
 				for (t = ip + match[no].rm_so; mlen--; ++t)
@@ -1433,6 +1415,9 @@ subzero:			if (match[no].rm_so == -1 ||
 				++rp;
 				conv = C_UPPER;
 				continue;
+			case '\r':
+				OUTCH(ch, 0);
+				continue;
 			default:
 				++rp;
 				break;
