@@ -9,10 +9,6 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: v_increment.c,v 10.16 2001/06/25 15:19:31 skimo Exp $ (Berkeley) $Date: 2001/06/25 15:19:31 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/time.h>
@@ -28,26 +24,26 @@ static const char sccsid[] = "$Id: v_increment.c,v 10.16 2001/06/25 15:19:31 ski
 #include "../common/common.h"
 #include "vi.h"
 
-static char * const fmt[] = {
+static CHAR_T * const fmt[] = {
 #define	DEC	0
-	"%ld",
+	L("%ld"),
 #define	SDEC	1
-	"%+ld",
+	L("%+ld"),
 #define	HEXC	2
-	"0X%0*lX",
+	L("0X%0*lX"),
 #define	HEXL	3
-	"0x%0*lx",
+	L("0x%0*lx"),
 #define	OCTAL	4
-	"%#0*lo",
+	L("%#0*lo"),
 };
 
-static void inc_err __P((SCR *, enum nresult));
+static void inc_err(SCR *, enum nresult);
 
 /*
  * v_increment -- [count]#[#+-]
  *	Increment/decrement a keyword number.
  *
- * PUBLIC: int v_increment __P((SCR *, VICMD *));
+ * PUBLIC: int v_increment(SCR *, VICMD *);
  */
 int
 v_increment(SCR *sp, VICMD *vp)
@@ -57,7 +53,7 @@ v_increment(SCR *sp, VICMD *vp)
 	long change, ltmp, lval;
 	size_t beg, blen, end, len, nlen, wlen;
 	int base, isempty, rval;
-	char *ntype, nbuf[100];
+	CHAR_T *ntype, nbuf[100];
 	CHAR_T *bp, *p, *t;
 
 	/* Validate the operator. */
@@ -90,7 +86,7 @@ v_increment(SCR *sp, VICMD *vp)
 	 * implies moving the cursor to its beginning, if we moved, refresh
 	 * now.
 	 */
-	for (beg = vp->m_start.cno; beg < len && isspace(p[beg]); ++beg);
+	for (beg = vp->m_start.cno; beg < len && ISSPACE(p[beg]); ++beg);
 	if (beg >= len)
 		goto nonum;
 	if (beg != vp->m_start.cno) {
@@ -99,9 +95,9 @@ v_increment(SCR *sp, VICMD *vp)
 	}
 
 #undef	ishex
-#define	ishex(c)	(isdigit(c) || strchr("abcdefABCDEF", c))
+#define	ishex(c)	(ISXDIGIT(c))
 #undef	isoctal
-#define	isoctal(c)	(isdigit(c) && (c) != '8' && (c) != '9')
+#define	isoctal(c)	((c) >= '0' && (c) <= '7')
 
 	/*
 	 * Look for 0[Xx], or leading + or - signs, guess at the base.
@@ -176,7 +172,7 @@ nonum:			msgq(sp, M_ERR, "181|Cursor not in a number");
 	 */
 	GET_SPACE_RETW(sp, bp, blen, len + 50);
 	if (end == len) {
-		MEMMOVEW(bp, &p[beg], wlen);
+		MEMMOVE(bp, &p[beg], wlen);
 		bp[wlen] = '\0';
 		t = bp;
 	} else
@@ -187,7 +183,7 @@ nonum:			msgq(sp, M_ERR, "181|Cursor not in a number");
 	 * in signed longs.
 	 */
 	if (base == 10) {
-		if ((nret = nget_slong(sp, &lval, t, NULL, 10)) != NUM_OK)
+		if ((nret = nget_slong(&lval, t, NULL, 10)) != NUM_OK)
 			goto err;
 		ltmp = vp->character == '-' ? -change : change;
 		if (lval > 0 && ltmp > 0 && !NPFITS(LONG_MAX, lval, ltmp)) {
@@ -202,9 +198,9 @@ nonum:			msgq(sp, M_ERR, "181|Cursor not in a number");
 		/* If we cross 0, signed numbers lose their sign. */
 		if (lval == 0 && ntype == fmt[SDEC])
 			ntype = fmt[DEC];
-		nlen = snprintf(nbuf, sizeof(nbuf), ntype, lval);
+		nlen = SPRINTF(nbuf, sizeof(nbuf), ntype, lval);
 	} else {
-		if ((nret = nget_uslong(sp, &ulval, t, NULL, base)) != NUM_OK)
+		if ((nret = nget_uslong(&ulval, t, NULL, base)) != NUM_OK)
 			goto err;
 		if (vp->character == '+') {
 			if (!NPFITS(ULONG_MAX, ulval, change)) {
@@ -224,13 +220,13 @@ nonum:			msgq(sp, M_ERR, "181|Cursor not in a number");
 		if (base == 16)
 			wlen -= 2;
 
-		nlen = snprintf(nbuf, sizeof(nbuf), ntype, wlen, ulval);
+		nlen = SPRINTF(nbuf, sizeof(nbuf), ntype, wlen, ulval);
 	}
 
 	/* Build the new line. */
-	MEMMOVEW(bp, p, beg);
-	MEMMOVEW(bp + beg, nbuf, nlen);
-	MEMMOVEW(bp + beg + nlen, p + end, len - beg - (end - beg));
+	MEMMOVE(bp, p, beg);
+	MEMMOVE(bp + beg, nbuf, nlen);
+	MEMMOVE(bp + beg + nlen, p + end, len - beg - (end - beg));
 	len = beg + nlen + (len - beg - (end - beg));
 
 	nret = NUM_OK;
