@@ -9,14 +9,9 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: v_txt.c,v 10.108 2003/07/18 21:27:42 skimo Exp $ (Berkeley) $Date: 2003/07/18 21:27:42 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/stat.h>
-#include <sys/time.h>
 
 #include <bitstring.h>
 #include <ctype.h>
@@ -30,25 +25,25 @@ static const char sccsid[] = "$Id: v_txt.c,v 10.108 2003/07/18 21:27:42 skimo Ex
 #include "../common/common.h"
 #include "vi.h"
 
-static int	 txt_abbrev __P((SCR *, TEXT *, CHAR_T *, int, int *, int *));
-static void	 txt_ai_resolve __P((SCR *, TEXT *, int *));
-static TEXT	*txt_backup __P((SCR *, TEXTH *, TEXT *, u_int32_t *));
-static int	 txt_dent __P((SCR *, TEXT *, int));
-static int	 txt_emark __P((SCR *, TEXT *, size_t));
-static void	 txt_err __P((SCR *, TEXTH *));
-static int	 txt_fc __P((SCR *, TEXT *, int *));
-static int	 txt_fc_col __P((SCR *, int, ARGS **));
-static int	 txt_hex __P((SCR *, TEXT *));
-static int	 txt_insch __P((SCR *, TEXT *, CHAR_T *, u_int));
-static int	 txt_isrch __P((SCR *, VICMD *, TEXT *, u_int8_t *));
-static int	 txt_map_end __P((SCR *));
-static int	 txt_map_init __P((SCR *));
-static int	 txt_margin __P((SCR *, TEXT *, TEXT *, int *, u_int32_t));
-static void	 txt_nomorech __P((SCR *));
-static void	 txt_Rresolve __P((SCR *, TEXTH *, TEXT *, const size_t));
-static int	 txt_resolve __P((SCR *, TEXTH *, u_int32_t));
-static int	 txt_showmatch __P((SCR *, TEXT *));
-static void	 txt_unmap __P((SCR *, TEXT *, u_int32_t *));
+static int	 txt_abbrev(SCR *, TEXT *, CHAR_T *, int, int *, int *);
+static void	 txt_ai_resolve(SCR *, TEXT *, int *);
+static TEXT	*txt_backup(SCR *, TEXTH *, TEXT *, u_int32_t *);
+static int	 txt_dent(SCR *, TEXT *, int, int);
+static int	 txt_emark(SCR *, TEXT *, size_t);
+static void	 txt_err(SCR *, TEXTH *);
+static int	 txt_fc(SCR *, TEXT *, int *);
+static int	 txt_fc_col(SCR *, int, ARGS **);
+static int	 txt_hex(SCR *, TEXT *);
+static int	 txt_insch(SCR *, TEXT *, CHAR_T *, u_int);
+static int	 txt_isrch(SCR *, VICMD *, TEXT *, u_int8_t *);
+static int	 txt_map_end(SCR *);
+static int	 txt_map_init(SCR *);
+static int	 txt_margin(SCR *, TEXT *, TEXT *, int *, u_int32_t);
+static void	 txt_nomorech(SCR *);
+static void	 txt_Rresolve(SCR *, TEXTH *, TEXT *, const size_t);
+static int	 txt_resolve(SCR *, TEXTH *, u_int32_t);
+static int	 txt_showmatch(SCR *, TEXT *);
+static void	 txt_unmap(SCR *, TEXT *, u_int32_t *);
 
 /* Cursor character (space is hard to track on the screen). */
 #if defined(DEBUG) && 0
@@ -60,7 +55,7 @@ static void	 txt_unmap __P((SCR *, TEXT *, u_int32_t *));
  * v_tcmd --
  *	Fill a buffer from the terminal for vi.
  *
- * PUBLIC: int v_tcmd __P((SCR *, VICMD *, ARG_CHAR_T, u_int));
+ * PUBLIC: int v_tcmd(SCR *, VICMD *, ARG_CHAR_T, u_int);
  */
 int
 v_tcmd(SCR *sp, VICMD *vp, ARG_CHAR_T prompt, u_int flags)
@@ -235,27 +230,27 @@ txt_map_end(SCR *sp)
  * v_txt --
  *	Vi text input.
  *
- * PUBLIC: int v_txt __P((SCR *, VICMD *, MARK *,
- * PUBLIC:    const CHAR_T *, size_t, ARG_CHAR_T, db_recno_t, u_long, u_int32_t));
+ * PUBLIC: int v_txt(SCR *, VICMD *, MARK *,
+ * PUBLIC:    const CHAR_T *, size_t, ARG_CHAR_T, recno_t, u_long, u_int32_t);
  */
 int
-v_txt(SCR *sp, VICMD *vp, MARK *tm, const CHAR_T *lp, size_t len, ARG_CHAR_T prompt, db_recno_t ai_line, u_long rcount, u_int32_t flags)
-	        
-	          
-	         		/* To MARK. */
-	                 	/* Input line. */
-	           		/* Input line length. */
-	                  	/* Prompt to display. */
-	                   	/* Line number to use for autoindent count. */
-	              		/* Replay count. */
-	                	/* TXT_* flags. */
+v_txt(
+	SCR *sp,
+	VICMD *vp,
+	MARK *tm,		/* To MARK. */
+	const CHAR_T *lp,	/* Input line. */
+	size_t len,		/* Input line length. */
+	ARG_CHAR_T prompt,	/* Prompt to display. */
+	recno_t ai_line,	/* Line number to use for autoindent count. */
+	u_long rcount,		/* Replay count. */
+	u_int32_t flags)	/* TXT_* flags. */
 {
-	EVENT ev, *evp;		/* Current event. */
+	EVENT ev, *evp = NULL;	/* Current event. */
 	EVENT fc;		/* File name completion event. */
 	GS *gp;
 	TEXT *ntp, *tp;		/* Input text structures. */
 	TEXT ait;		/* Autoindent text structure. */
-	TEXT wmt;		/* Wrapmargin text structure. */
+	TEXT wmt = {{ 0 }};	/* Wrapmargin text structure. */
 	TEXTH *tiqh;
 	VI_PRIVATE *vip;
 	abb_t abb;		/* State of abbreviation checks. */
@@ -275,6 +270,7 @@ v_txt(SCR *sp, VICMD *vp, MARK *tm, const CHAR_T *lp, size_t len, ARG_CHAR_T pro
 	int showmatch;		/* Showmatch set on this character. */
 	int wm_set, wm_skip;	/* Wrapmargin happened, blank skip flags. */
 	int max, tmp;
+	int nochange;
 	CHAR_T *p;
 
 	gp = sp->gp;
@@ -292,10 +288,11 @@ v_txt(SCR *sp, VICMD *vp, MARK *tm, const CHAR_T *lp, size_t len, ARG_CHAR_T pro
 	 * default to 0 -- text_init() handles this.)  If changing a line,
 	 * copy it into the TEXT buffer.
 	 */
-	tiqh = &sp->tiq;
-	if (tiqh->cqh_first != (void *)tiqh) {
-		tp = tiqh->cqh_first;
-		if (tp->q.cqe_next != (void *)tiqh || tp->lb_len < len + 32) {
+	tiqh = sp->tiq;
+	if (!TAILQ_EMPTY(tiqh)) {
+		tp = TAILQ_FIRST(tiqh);
+		if (TAILQ_NEXT(tp, q) != NULL ||
+		    tp->lb_len < (len + 32) * sizeof(CHAR_T)) {
 			text_lfree(tiqh);
 			goto newtp;
 		}
@@ -303,13 +300,13 @@ v_txt(SCR *sp, VICMD *vp, MARK *tm, const CHAR_T *lp, size_t len, ARG_CHAR_T pro
 		if (lp != NULL) {
 			tp->len = len;
 			BINC_RETW(sp, tp->lb, tp->lb_len, len);
-			MEMMOVEW(tp->lb, lp, len);
+			MEMMOVE(tp->lb, lp, len);
 		} else
 			tp->len = 0;
 	} else {
 newtp:		if ((tp = text_init(sp, lp, len, len + 32)) == NULL)
 			return (1);
-		CIRCLEQ_INSERT_HEAD(tiqh, tp, q);
+		TAILQ_INSERT_HEAD(tiqh, tp, q);
 	}
 
 	/* Set default termination condition. */
@@ -458,6 +455,7 @@ newtp:		if ((tp = text_init(sp, lp, len, len + 32)) == NULL)
 	/* Other text input mode setup. */
 	quote = Q_NOTSET;
 	carat = C_NOTSET;
+	nochange = 0;
 	FL_INIT(is_flags,
 	    LF_ISSET(TXT_SEARCHINCR) ? IS_RESTART | IS_RUNNING : 0);
 	filec_redraw = hexcnt = showmatch = 0;
@@ -493,7 +491,7 @@ next:	if (v_event_get(sp, evp, 0, ec_flags))
 		    sp->rows ? 1 : sp->rows - vip->totalcount;
 		fc.e_tlno = sp->rows;
 		vip->linecount = vip->lcontinue = vip->totalcount = 0;
-		(void)v_erepaint(sp, &fc);
+		(void)vs_repaint(sp, &fc);
 		(void)vs_refresh(sp, 1);
 	}
 
@@ -515,7 +513,7 @@ next:	if (v_event_get(sp, evp, 0, ec_flags))
 		 */
 		goto k_escape;
 	case E_REPAINT:
-		if (v_erepaint(sp, &ev))
+		if (vs_repaint(sp, &ev))
 			return (1);
 		goto next;
 	case E_WRESIZE:
@@ -534,7 +532,7 @@ next:	if (v_event_get(sp, evp, 0, ec_flags))
 	 * This was not documented as far as I know, and is a great test of vi
 	 * clones.
 	 */
-	if (rcol == 0 && !LF_ISSET(TXT_REPLAY) && evp->e_c == '\0') {
+	if (LF_ISSET(TXT_RECORD) && rcol == 0 && evp->e_c == '\0') {
 		if (vip->rep == NULL)
 			goto done;
 
@@ -574,7 +572,7 @@ next:	if (v_event_get(sp, evp, 0, ec_flags))
 
 	/* Abbreviation overflow check.  See comment in txt_abbrev(). */
 #define	MAX_ABBREVIATION_EXPANSION	256
-	if (FL_ISSET(evp->e_flags, CH_ABBREVIATED)) {
+	if (F_ISSET(&evp->e_ch, CH_ABBREVIATED)) {
 		if (++abcnt > MAX_ABBREVIATION_EXPANSION) {
 			if (v_event_flush(sp, CH_ABBREVIATED))
 				msgq(sp, M_ERR,
@@ -608,35 +606,26 @@ replay:	if (LF_ISSET(TXT_REPLAY)) {
 	}
 
 	/* If quoted by someone else, simply insert the character. */
-	if (FL_ISSET(evp->e_flags, CH_QUOTED))
+	if (F_ISSET(&evp->e_ch, CH_QUOTED))
 		goto insq_ch;
 
 	/*
 	 * !!!
-	 * If this character was quoted by a K_VLNEXT or a backslash, replace
-	 * the placeholder (a carat or a backslash) with the new character.
-	 * If it was quoted by a K_VLNEXT, we've already adjusted the cursor
-	 * because it has to appear on top of the placeholder character.  If
-	 * it was quoted by a backslash, adjust the cursor now, the cursor
-	 * doesn't appear on top of it.  Historic practice in both cases.
+	 * If this character was quoted by a K_VLNEXT, replace the placeholder
+	 * (a carat) with the new character.  We've already adjusted the cursor
+	 * because it has to appear on top of the placeholder character.
+	 * Historic practice.
 	 *
 	 * Skip tests for abbreviations; ":ab xa XA" followed by "ixa^V<space>"
 	 * doesn't perform an abbreviation.  Special case, ^V^J (not ^V^M) is
 	 * the same as ^J, historically.
 	 */
-	if (quote == Q_BTHIS || quote == Q_VTHIS) {
+	if (quote == Q_VTHIS) {
 		FL_CLR(ec_flags, EC_QUOTED);
 		if (LF_ISSET(TXT_MAPINPUT))
 			FL_SET(ec_flags, EC_MAPINPUT);
 
-		if (quote == Q_BTHIS &&
-		    (evp->e_value == K_VERASE || evp->e_value == K_VKILL)) {
-			quote = Q_NOTSET;
-			--tp->cno;
-			++tp->owrite;
-			goto insl_ch;
-		}
-		if (quote == Q_VTHIS && evp->e_value != K_NL) {
+		if (evp->e_value != K_NL) {
 			quote = Q_NOTSET;
 			goto insl_ch;
 		}
@@ -649,7 +638,7 @@ replay:	if (LF_ISSET(TXT_REPLAY)) {
 	 * this test delimits the value by any non-hex character.  Offset by
 	 * one, we use 0 to mean that we've found <CH_HEX>.
 	 */
-	if (hexcnt > 1 && !isxdigit(evp->e_c)) {
+	if (hexcnt > 1 && !ISXDIGIT(evp->e_c)) {
 		hexcnt = 0;
 		if (txt_hex(sp, tp))
 			goto err;
@@ -660,7 +649,6 @@ replay:	if (LF_ISSET(TXT_REPLAY)) {
 	case K_NL:				/* New line. */
 		/* Return in script windows and the command line. */
 k_cr:		if (LF_ISSET(TXT_CR)) {
-			static CHAR_T cr[] = { '\r', 0 };
 			/*
 			 * If this was a map, we may have not displayed
 			 * the line.  Display it, just in case.
@@ -672,7 +660,7 @@ k_cr:		if (LF_ISSET(TXT_CR)) {
 				if (vs_change(sp, tp->lno, LINE_RESET))
 					goto err;
 			} else if (F_ISSET(sp, SC_SCRIPT))
-				(void)v_event_push(sp, NULL, cr, 1, CH_NOMAP);
+				(void)v_event_push(sp, NULL, L("\r"), 1, CH_NOMAP);
 
 			/* Set term condition: if empty. */
 			if (tp->cno <= tp->offset)
@@ -770,7 +758,7 @@ k_cr:		if (LF_ISSET(TXT_CR)) {
 		if ((ntp = text_init(sp, p,
 		    insert + owrite, insert + owrite + 32)) == NULL)
 			goto err;
-		CIRCLEQ_INSERT_TAIL(&sp->tiq, ntp, q);
+		TAILQ_INSERT_TAIL(sp->tiq, ntp, q);
 
 		/* Set up bookkeeping for the new line. */
 		ntp->insert = insert;
@@ -785,7 +773,8 @@ k_cr:		if (LF_ISSET(TXT_CR)) {
 		 * characters may have been erased.
 		 */
 		if (LF_ISSET(TXT_AUTOINDENT)) {
-			if (carat == C_NOCHANGE) {
+			if (nochange) {
+				nochange = 0;
 				if (v_txt_auto(sp, OOBLNO, &ait, ait.ai, ntp))
 					goto err;
 				FREE_SPACEW(sp, ait.lb, ait.lb_len);
@@ -809,7 +798,7 @@ k_cr:		if (LF_ISSET(TXT_CR)) {
 #define	WMTSPACE	wmt.offset + wmt.owrite + wmt.insert
 				BINC_GOTOW(sp, ntp->lb,
 				    ntp->lb_len, ntp->len + WMTSPACE + 32);
-				MEMMOVEW(ntp->lb + ntp->cno, wmt.lb, WMTSPACE);
+				MEMMOVE(ntp->lb + ntp->cno, wmt.lb, WMTSPACE);
 				ntp->len += WMTSPACE;
 				ntp->cno += wmt.offset;
 				ntp->owrite = wmt.owrite;
@@ -884,7 +873,7 @@ k_escape:	LINE_RESOLVE;
 		 * characters, and making them into insert characters.
 		 */
 		if (LF_ISSET(TXT_REPLACE))
-			txt_Rresolve(sp, &sp->tiq, tp, len);
+			txt_Rresolve(sp, sp->tiq, tp, len);
 
 		/*
 		 * If there are any overwrite characters, copy down
@@ -892,7 +881,7 @@ k_escape:	LINE_RESOLVE;
 		 */
 		if (tp->owrite) {
 			if (tp->insert)
-				MEMMOVEW(tp->lb + tp->cno,
+				MEMMOVE(tp->lb + tp->cno,
 				    tp->lb + tp->cno + tp->owrite, tp->insert);
 			tp->len -= tp->owrite;
 		}
@@ -907,7 +896,7 @@ k_escape:	LINE_RESOLVE;
 		 * This is wrong, should pass back a length.
 		 */
 		if (LF_ISSET(TXT_RESOLVE)) {
-			if (txt_resolve(sp, &sp->tiq, flags))
+			if (txt_resolve(sp, sp->tiq, flags))
 				goto err;
 		} else {
 			BINC_GOTOW(sp, tp->lb, tp->lb_len, tp->len + 1);
@@ -955,10 +944,11 @@ k_escape:	LINE_RESOLVE;
 			ait.lb = NULL;
 			ait.lb_len = 0;
 			BINC_GOTOW(sp, ait.lb, ait.lb_len, tp->ai);
-			MEMMOVEW(ait.lb, tp->lb, tp->ai);
+			MEMMOVE(ait.lb, tp->lb, tp->ai);
 			ait.ai = ait.len = tp->ai;
 
-			carat = C_NOCHANGE;
+			carat = C_NOTSET;
+			nochange = 1;
 			goto leftmargin;
 		case C_ZEROSET:		/* 0^D */
 			if (tp->ai == 0 || tp->cno > tp->ai + tp->offset + 1)
@@ -974,7 +964,7 @@ leftmargin:		tp->lb[tp->cno - 1] = ' ';
 			if (tp->ai == 0 || tp->cno > tp->ai + tp->offset)
 				goto ins_ch;
 
-			(void)txt_dent(sp, tp, 0);
+			(void)txt_dent(sp, tp, O_SHIFTWIDTH, 0);
 			break;
 		default:
 			abort();
@@ -993,7 +983,7 @@ leftmargin:		tp->lb[tp->cno - 1] = ' ';
 		 */
 		if (tp->cno == 0) {
 			if ((ntp =
-			    txt_backup(sp, &sp->tiq, tp, &flags)) == NULL)
+			    txt_backup(sp, sp->tiq, tp, &flags)) == NULL)
 				goto err;
 			tp = ntp;
 			break;
@@ -1045,7 +1035,7 @@ leftmargin:		tp->lb[tp->cno - 1] = ' ';
 		 */
 		if (tp->cno == 0) {
 			if ((ntp =
-			    txt_backup(sp, &sp->tiq, tp, &flags)) == NULL)
+			    txt_backup(sp, sp->tiq, tp, &flags)) == NULL)
 				goto err;
 			tp = ntp;
 		}
@@ -1075,7 +1065,7 @@ leftmargin:		tp->lb[tp->cno - 1] = ' ';
 		}
 
 		/* Skip over trailing space characters. */
-		while (tp->cno > max && isblank(tp->lb[tp->cno - 1])) {
+		while (tp->cno > max && ISBLANK(tp->lb[tp->cno - 1])) {
 			--tp->cno;
 			++tp->owrite;
 		}
@@ -1104,12 +1094,12 @@ leftmargin:		tp->lb[tp->cno - 1] = ' ';
 		 */
 		if (LF_ISSET(TXT_TTYWERASE))
 			while (tp->cno > max) {
+				if (ISBLANK(tp->lb[tp->cno - 1]))
+					break;
 				--tp->cno;
 				++tp->owrite;
 				if (FL_ISSET(is_flags, IS_RUNNING))
 					tp->lb[tp->cno] = ' ';
-				if (isblank(tp->lb[tp->cno - 1]))
-					break;
 			}
 		else {
 			if (LF_ISSET(TXT_ALTWERASE)) {
@@ -1117,19 +1107,17 @@ leftmargin:		tp->lb[tp->cno - 1] = ' ';
 				++tp->owrite;
 				if (FL_ISSET(is_flags, IS_RUNNING))
 					tp->lb[tp->cno] = ' ';
-				if (isblank(tp->lb[tp->cno - 1]))
-					break;
 			}
 			if (tp->cno > max)
 				tmp = inword(tp->lb[tp->cno - 1]);
 			while (tp->cno > max) {
+				if (tmp != inword(tp->lb[tp->cno - 1])
+				    || ISBLANK(tp->lb[tp->cno - 1]))
+					break;
 				--tp->cno;
 				++tp->owrite;
 				if (FL_ISSET(is_flags, IS_RUNNING))
 					tp->lb[tp->cno] = ' ';
-				if (tmp != inword(tp->lb[tp->cno - 1])
-				    || isblank(tp->lb[tp->cno - 1]))
-					break;
 			}
 		}
 
@@ -1146,7 +1134,7 @@ leftmargin:		tp->lb[tp->cno - 1] = ' ';
 		 */
 		if (tp->cno == 0) {
 			if ((ntp =
-			    txt_backup(sp, &sp->tiq, tp, &flags)) == NULL)
+			    txt_backup(sp, sp->tiq, tp, &flags)) == NULL)
 				goto err;
 			tp = ntp;
 		}
@@ -1192,39 +1180,9 @@ leftmargin:		tp->lb[tp->cno - 1] = ' ';
 	case K_CNTRLT:			/* Add autoindent characters. */
 		if (!LF_ISSET(TXT_CNTRLT))
 			goto ins_ch;
-		if (txt_dent(sp, tp, 1))
+		if (txt_dent(sp, tp, O_SHIFTWIDTH, 1))
 			goto err;
 		goto ebuf_chk;
-	case K_RIGHTBRACE:
-	case K_RIGHTPAREN:
-		if (LF_ISSET(TXT_SHOWMATCH))
-			showmatch = 1;
-		goto ins_ch;
-	case K_BACKSLASH:		/* Quote next erase/kill. */
-		/*
-		 * !!!
-		 * Historic vi tried to make abbreviations after a backslash
-		 * escape work.  If you did ":ab x y", and inserted "x\^H",
-		 * (assuming the erase character was ^H) you got "x^H", and
-		 * no abbreviation was done.  If you inserted "x\z", however,
-		 * it tried to back up and do the abbreviation, i.e. replace
-		 * 'x' with 'y'.  The problem was it got it wrong, and you
-		 * ended up with "zy\".
-		 *
-		 * This is really hard to do (you have to remember the
-		 * word/non-word state, for example), and doesn't make any
-		 * sense to me.  Both backslash and the characters it
-		 * (usually) escapes will individually trigger the
-		 * abbreviation, so I don't see why the combination of them
-		 * wouldn't.  I don't expect to get caught on this one,
-		 * particularly since it never worked right, but I've been
-		 * wrong before.
-		 *
-		 * Do the tests for abbreviations, so ":ab xa XA",
-		 * "ixa\<K_VERASE>" performs the abbreviation.
-		 */
-		quote = Q_BNEXT;
-		goto insq_ch;
 	case K_VLNEXT:			/* Quote next character. */
 		evp->e_c = '^';
 		quote = Q_VNEXT;
@@ -1246,7 +1204,23 @@ leftmargin:		tp->lb[tp->cno - 1] = ' ';
 	case K_HEXCHAR:
 		hexcnt = 1;
 		goto insq_ch;
+	case K_TAB:
+		if (sp->showmode != SM_COMMAND && quote != Q_VTHIS &&
+		    O_ISSET(sp, O_EXPANDTAB)) {
+			if (txt_dent(sp, tp, O_TABSTOP, 1))
+				goto err;
+			goto ebuf_chk;
+		}
+		goto insq_ch;
 	default:			/* Insert the character. */
+		if (LF_ISSET(TXT_SHOWMATCH)) {
+			CHAR_T *match_chars, *cp;
+
+			match_chars = VIP(sp)->mcs;
+			cp = STRCHR(match_chars, evp->e_c);
+			if (cp != NULL && (cp - match_chars) & 1)
+				showmatch = 1;
+		}
 ins_ch:		/*
 		 * Historically, vi eliminated nul's out of hand.  If the
 		 * beautify option was set, it also deleted any unknown
@@ -1313,7 +1287,7 @@ insl_ch:	if (txt_insch(sp, tp, &evp->e_c, flags))
 		 * number of hex bytes.  Offset by one, we use 0 to mean
 		 * that we've found <CH_HEX>.
 		 */
-		if (hexcnt != 0 && hexcnt++ == sizeof(CHAR_T) * 2 + 1) {
+		if (hexcnt != 0 && hexcnt++ == 3) {
 			hexcnt = 0;
 			if (txt_hex(sp, tp))
 				goto err;
@@ -1361,19 +1335,15 @@ ebuf_chk:	if (tp->cno >= tp->len) {
 		}
 
 		/* Step the quote state forward. */
-		if (quote != Q_NOTSET) {
-			if (quote == Q_BNEXT)
-				quote = Q_BTHIS;
-			if (quote == Q_VNEXT)
-				quote = Q_VTHIS;
-		}
+		if (quote == Q_VNEXT)
+			quote = Q_VTHIS;
 		break;
 	}
 
 #ifdef DEBUG
 	if (tp->cno + tp->insert + tp->owrite != tp->len) {
 		msgq(sp, M_ERR,
-		    "len %u != cno: %u ai: %u insert %u overwrite %u",
+		    "len %zu != cno: %zu ai: %zu insert %zu overwrite %zu",
 		    tp->len, tp->cno, tp->ai, tp->insert, tp->owrite);
 		if (LF_ISSET(TXT_REPLAY))
 			goto done;
@@ -1457,7 +1427,7 @@ done:	/* Leave input mode. */
 err:
 alloc_err:
 	F_CLR(sp, SC_TINPUT);
-	txt_err(sp, &sp->tiq);
+	txt_err(sp, sp->tiq);
 	return (1);
 }
 
@@ -1559,7 +1529,7 @@ txt_abbrev(SCR *sp, TEXT *tp, CHAR_T *pushcp, int isinfoline, int *didsubp, int
 	 */
 search:	if (isinfoline)
 		if (off == tp->ai || off == tp->offset)
-			if (ex_is_abbrev(sp, p, len)) {
+			if (ex_is_abbrev(p, len)) {
 				*turnoffp = 1;
 				return (0);
 			} else
@@ -1610,7 +1580,7 @@ search:	if (isinfoline)
 		tp->owrite += len;
 	else {
 		if (tp->insert)
-			MEMMOVEW(tp->lb + tp->cno + qp->olen,
+			MEMMOVE(tp->lb + tp->cno + qp->olen,
 			    tp->lb + tp->cno + tp->owrite + len, tp->insert);
 		tp->owrite += qp->olen;
 		tp->len -= len - qp->olen;
@@ -1661,7 +1631,7 @@ txt_unmap(SCR *sp, TEXT *tp, u_int32_t *ec_flagsp)
 	 * if the user erases the line and starts another command, we go ahead
 	 * an turn mapping back on.
 	 */
-	if ((off == tp->ai || off == tp->offset) && ex_is_unmap(sp, p, len))
+	if ((off == tp->ai || off == tp->offset) && ex_is_unmap(p, len))
 		FL_CLR(*ec_flagsp, EC_MAPINPUT);
 	else
 		FL_SET(*ec_flagsp, EC_MAPINPUT);
@@ -1719,13 +1689,19 @@ txt_ai_resolve(SCR *sp, TEXT *tp, int *changedp)
 	/*
 	 * If there are no spaces, or no tabs after spaces and less than
 	 * ts spaces, it's already minimal.
+	 * Keep analysing if expandtab is set.
 	 */
-	if (!spaces || !tab_after_sp && spaces < ts)
+	if ((!spaces || (!tab_after_sp && spaces < ts)) &&
+	    !O_ISSET(sp, O_EXPANDTAB))
 		return;
 
 	/* Count up spaces/tabs needed to get to the target. */
-	for (cno = 0, tabs = 0; cno + COL_OFF(cno, ts) <= scno; ++tabs)
-		cno += COL_OFF(cno, ts);
+	cno = 0;
+	tabs = 0;
+	if (!O_ISSET(sp, O_EXPANDTAB)) {
+		for (; cno + COL_OFF(cno, ts) <= scno; ++tabs)
+			cno += COL_OFF(cno, ts);
+	}
 	spaces = scno - cno;
 
 	/*
@@ -1739,7 +1715,7 @@ txt_ai_resolve(SCR *sp, TEXT *tp, int *changedp)
 
 	/* Shift the rest of the characters down, adjust the counts. */
 	del = old - new;
-	MEMMOVEW(p - del, p, tp->len - old);
+	MEMMOVE(p - del, p, tp->len - old);
 	tp->len -= del;
 	tp->cno -= del;
 
@@ -1756,10 +1732,10 @@ txt_ai_resolve(SCR *sp, TEXT *tp, int *changedp)
  *	Handle autoindent.  If aitp isn't NULL, use it, otherwise,
  *	retrieve the line.
  *
- * PUBLIC: int v_txt_auto __P((SCR *, db_recno_t, TEXT *, size_t, TEXT *));
+ * PUBLIC: int v_txt_auto(SCR *, recno_t, TEXT *, size_t, TEXT *);
  */
 int
-v_txt_auto(SCR *sp, db_recno_t lno, TEXT *aitp, size_t len, TEXT *tp)
+v_txt_auto(SCR *sp, recno_t lno, TEXT *aitp, size_t len, TEXT *tp)
 {
 	size_t nlen;
 	CHAR_T *p, *t;
@@ -1793,11 +1769,11 @@ v_txt_auto(SCR *sp, db_recno_t lno, TEXT *aitp, size_t len, TEXT *tp)
 
 	/* Copy the buffer's current contents up. */
 	if (tp->len != 0)
-		MEMMOVEW(tp->lb + nlen, tp->lb, tp->len);
+		MEMMOVE(tp->lb + nlen, tp->lb, tp->len);
 	tp->len += nlen;
 
 	/* Copy the indentation into the new buffer. */
-	MEMMOVEW(tp->lb, t, nlen);
+	MEMMOVE(tp->lb, t, nlen);
 
 	/* Set the autoindent count. */
 	tp->ai = nlen;
@@ -1815,7 +1791,7 @@ txt_backup(SCR *sp, TEXTH *tiqh, TEXT *tp, u_int32_t *flagsp)
 	TEXT *ntp;
 
 	/* Get a handle on the previous TEXT structure. */
-	if ((ntp = tp->q.cqe_prev) == (void *)tiqh) {
+	if ((ntp = TAILQ_PREV(tp, _texth, q)) == NULL) {
 		if (!FL_ISSET(*flagsp, TXT_REPLAY))
 			msgq(sp, M_BERR,
 			    "193|Already at the beginning of the insert");
@@ -1836,7 +1812,7 @@ txt_backup(SCR *sp, TEXTH *tiqh, TEXT *tp, u_int32_t *flagsp)
 		FL_CLR(*flagsp, TXT_APPENDEOL);
 
 	/* Release the current TEXT. */
-	CIRCLEQ_REMOVE(tiqh, tp, q);
+	TAILQ_REMOVE(tiqh, tp, q);
 	text_free(tp);
 
 	/* Update the old line on the screen. */
@@ -1874,7 +1850,7 @@ txt_backup(SCR *sp, TEXTH *tiqh, TEXT *tp, u_int32_t *flagsp)
  * Technically, txt_dent should be part of the screen interface, as it requires
  * knowledge of character sizes, including <space>s, on the screen.  It's here
  * because it's a complicated little beast, and I didn't want to shove it down
- * into the screen.  It's probable that KEY_LEN will call into the screen once
+ * into the screen.  It's probable that KEY_COL will call into the screen once
  * there are screens with different character representations.
  *
  * txt_dent --
@@ -1884,15 +1860,15 @@ txt_backup(SCR *sp, TEXTH *tiqh, TEXT *tp, u_int32_t *flagsp)
  * changes.
  */
 static int
-txt_dent(SCR *sp, TEXT *tp, int isindent)
+txt_dent(SCR *sp, TEXT *tp, int swopt, int isindent)
 {
 	CHAR_T ch;
 	u_long sw, ts;
-	size_t cno, current, spaces, target, tabs, off;
+	size_t cno, current, spaces, target, tabs;
 	int ai_reset;
 
 	ts = O_VAL(sp, O_TABSTOP);
-	sw = O_VAL(sp, O_SHIFTWIDTH);
+	sw = O_VAL(sp, swopt);
 
 	/*
 	 * Since we don't know what precedes the character(s) being inserted
@@ -1917,8 +1893,10 @@ txt_dent(SCR *sp, TEXT *tp, int isindent)
 	target = current;
 	if (isindent)
 		target += COL_OFF(target, sw);
-	else
-		target -= --target % sw;
+	else {
+		--target;
+		target -= target % sw;
+	}
 
 	/*
 	 * The AI characters will be turned into overwrite characters if the
@@ -1957,9 +1935,12 @@ txt_dent(SCR *sp, TEXT *tp, int isindent)
 	if (current >= target)
 		spaces = tabs = 0;
 	else {
-		for (cno = current,
-		    tabs = 0; cno + COL_OFF(cno, ts) <= target; ++tabs)
-			cno += COL_OFF(cno, ts);
+		cno = current;
+		tabs = 0;
+		if (!O_ISSET(sp, O_EXPANDTAB)) {
+			for (; cno + COL_OFF(cno, ts) <= target; ++tabs)
+				cno += COL_OFF(cno, ts);
+		}
 		spaces = target - cno;
 	}
 
@@ -1980,23 +1961,23 @@ txt_dent(SCR *sp, TEXT *tp, int isindent)
 
 /*
  * txt_fc --
- *	File name completion.
+ *	File name and ex command completion.
  */
 static int
 txt_fc(SCR *sp, TEXT *tp, int *redrawp)
 {
 	struct stat sb;
 	ARGS **argv;
-	CHAR_T s_ch;
 	EXCMD cmd;
 	size_t indx, len, nlen, off;
-	int argc, trydir;
-	CHAR_T *p, *t;
-	char *np;
+	int argc;
+	CHAR_T *p, *t, *bp;
+	char *np, *epd = NULL;
 	size_t nplen;
+	int fstwd = 1;
 
-	trydir = 0;
 	*redrawp = 0;
+	ex_cinit(sp, &cmd, 0, 0, OOBLNO, OOBLNO, 0);
 
 	/*
 	 * Find the beginning of this "word" -- if we're at the beginning
@@ -2005,67 +1986,53 @@ txt_fc(SCR *sp, TEXT *tp, int *redrawp)
 	if (tp->cno == 1) {
 		len = 0;
 		p = tp->lb;
-	} else
-retry:		for (len = 0,
-		    off = tp->cno - 1, p = tp->lb + off;; --off, --p) {
-			if (isblank(*p)) {
-				++p;
-				break;
-			}
-			++len;
-			if (off == tp->ai || off == tp->offset)
-				break;
+	} else {
+		CHAR_T *ap;
+
+		for (len = 0,
+		    off = MAX(tp->ai, tp->offset), ap = tp->lb + off, p = ap;
+		    off < tp->cno; ++off, ++ap) {
+			if (IS_ESCAPE(sp, &cmd, *ap)) {
+				if (++off == tp->cno)
+					break;
+				++ap;
+				len += 2;
+			} else if (cmdskip(*ap)) {
+				p = ap + 1;
+				if (len > 0)
+					fstwd = 0;
+				len = 0;
+			} else
+				++len;
 		}
+	}
 
 	/*
-	 * Get enough space for a wildcard character.
-	 *
-	 * XXX
-	 * This won't work for "foo\", since the \ will escape the expansion
-	 * character.  I'm not sure if that's a bug or not...
+	 * If we are at the first word, do ex command completion instead of
+	 * file name completion.
 	 */
-	off = p - tp->lb;
-	BINC_RETW(sp, tp->lb, tp->lb_len, tp->len + 1);
-	p = tp->lb + off;
-
-	s_ch = p[len];
-	p[len] = '*';
-
-	/* Build an ex command, and call the ex expansion routines. */
-	ex_cinit(sp, &cmd, 0, 0, OOBLNO, OOBLNO, 0);
-	if (argv_exp2(sp, &cmd, p, len + 1)) {
-		p[len] = s_ch;
-		return (0);
+	if (fstwd)
+		(void)argv_flt_ex(sp, &cmd, p, len);
+	else {
+		if ((bp = argv_uesc(sp, &cmd, p, len)) == NULL)
+			return (1);
+		if (argv_flt_path(sp, &cmd, bp, STRLEN(bp))) {
+			FREE_SPACEW(sp, bp, 0);
+			return (0);
+		}
+		FREE_SPACEW(sp, bp, 0);
 	}
 	argc = cmd.argc;
 	argv = cmd.argv;
 
-	p[len] = s_ch;
-
 	switch (argc) {
 	case 0:				/* No matches. */
-		if (!trydir)
-			(void)sp->gp->scr_bell(sp);
+		(void)sp->gp->scr_bell(sp);
 		return (0);
 	case 1:				/* One match. */
-		/* If something changed, do the exchange. */
+		/* Always overwrite the old text. */
 		nlen = STRLEN(cmd.argv[0]->bp);
-		if (len != nlen || MEMCMP(cmd.argv[0]->bp, p, len))
-			break;
-
-		/* If haven't done a directory test, do it now. */
-		INT2CHAR(sp, cmd.argv[0]->bp, cmd.argv[0]->len + 1,
-			 np, nplen);
-		if (!trydir &&
-		    !stat(np, &sb) && S_ISDIR(sb.st_mode)) {
-			p += len;
-			goto isdir;
-		}
-
-		/* If nothing changed, period, ring the bell. */
-		if (!trydir)
-			(void)sp->gp->scr_bell(sp);
-		return (0);
+		break;
 	default:			/* Multiple matches. */
 		*redrawp = 1;
 		if (txt_fc_col(sp, argc, argv))
@@ -2083,8 +2050,17 @@ retry:		for (len = 0,
 		break;
 	}
 
+	/* Escape the matched part of the path. */
+	if (fstwd)
+		bp = cmd.argv[0]->bp;
+	else {
+		if ((bp = argv_esc(sp, &cmd, cmd.argv[0]->bp, nlen)) == NULL)
+			return (1);
+		nlen = STRLEN(bp);
+	}
+
 	/* Overwrite the expanded text first. */
-	for (t = cmd.argv[0]->bp; len > 0 && nlen > 0; --len, --nlen)
+	for (t = bp; len > 0 && nlen > 0; --len, --nlen)
 		*p++ = *t++;
 
 	/* If lost text, make the remaining old text overwrite characters. */
@@ -2107,30 +2083,37 @@ retry:		for (len = 0,
 		tp->len += nlen;
 
 		if (tp->insert != 0)
-			(void)MEMMOVEW(p + nlen, p, tp->insert);
+			(void)MEMMOVE(p + nlen, p, tp->insert);
 		while (nlen--)
 			*p++ = *t++;
 	}
 
-	/* If a single match and it's a directory, retry it. */
+	if (!fstwd)
+		FREE_SPACEW(sp, bp, 0);
+
+	/* If not a single match of path, we've done. */
+	if (argc != 1 || fstwd)
+		return (0);
+
+	/* If a single match and it's a directory, append a '/'. */
 	INT2CHAR(sp, cmd.argv[0]->bp, cmd.argv[0]->len + 1, np, nplen);
-	if (argc == 1 && !stat(np, &sb) && S_ISDIR(sb.st_mode)) {
-isdir:		if (tp->owrite == 0) {
+	if ((epd = expanduser(np)) != NULL)
+		np = epd;
+	if (!stat(np, &sb) && S_ISDIR(sb.st_mode)) {
+		if (tp->owrite == 0) {
 			off = p - tp->lb;
 			BINC_RETW(sp, tp->lb, tp->lb_len, tp->len + 1);
 			p = tp->lb + off;
 			if (tp->insert != 0)
-				(void)MEMMOVEW(p + 1, p, tp->insert);
+				(void)MEMMOVE(p + 1, p, tp->insert);
 			++tp->len;
 		} else
 			--tp->owrite;
 
 		++tp->cno;
 		*p++ = '/';
-
-		trydir = 1;
-		goto retry;
 	}
+	free(epd);
 	return (0);
 }
 
@@ -2175,7 +2158,7 @@ txt_fc_col(SCR *sp, int argc, ARGS **argv)
 	 */
 	for (ac = argc, av = argv, colwidth = 0; ac > 0; --ac, ++av) {
 		for (col = 0, p = av[0]->bp + prefix; *p != '\0'; ++p)
-			col += KEY_LEN(sp, *p);
+			col += KEY_COL(sp, *p);
 		if (col > colwidth)
 			colwidth = col;
 	}
@@ -2196,17 +2179,17 @@ txt_fc_col(SCR *sp, int argc, ARGS **argv)
 		goto intr;
 
 	/* If the largest file name is too large, just print them. */
-	if (colwidth > sp->cols) {
+	if (colwidth >= sp->cols) {
 		for (ac = argc, av = argv; ac > 0; --ac, ++av) {
 			INT2CHAR(sp, av[0]->bp+prefix, av[0]->len+1-prefix,
 				 np, nlen);
 			pp = msg_print(sp, np, &nf);
 			(void)ex_printf(sp, "%s\n", pp);
+			if (nf)
+				FREE_SPACE(sp, pp, 0);
 			if (F_ISSET(gp, G_INTERRUPTED))
 				break;
 		}
-		if (nf)
-			FREE_SPACE(sp, pp, 0);
 		CHK_INTR;
 	} else {
 		/* Figure out the number of columns. */
@@ -2259,7 +2242,7 @@ static int
 txt_emark(SCR *sp, TEXT *tp, size_t cno)
 {
 	CHAR_T ch;
-	char *kp;
+	u_char *kp;
 	size_t chlen, nlen, olen;
 	CHAR_T *p;
 
@@ -2269,11 +2252,11 @@ txt_emark(SCR *sp, TEXT *tp, size_t cno)
 	 * The end mark may not be the same size as the current character.
 	 * Don't let the line shift.
 	 */
-	nlen = KEY_LEN(sp, ch);
+	nlen = KEY_COL(sp, ch);
 	if (tp->lb[cno] == '\t')
 		(void)vs_columns(sp, tp->lb, tp->lno, &cno, &olen);
 	else
-		olen = KEY_LEN(sp, tp->lb[cno]);
+		olen = KEY_COL(sp, tp->lb[cno]);
 
 	/*
 	 * If the line got longer, well, it's weird, but it's easy.  If
@@ -2284,17 +2267,19 @@ txt_emark(SCR *sp, TEXT *tp, size_t cno)
 		BINC_RETW(sp, tp->lb, tp->lb_len, tp->len + olen);
 		chlen = olen - nlen;
 		if (tp->insert != 0)
-			MEMMOVEW(tp->lb + cno + 1 + chlen,
+			MEMMOVE(tp->lb + cno + 1 + chlen,
 			    tp->lb + cno + 1, tp->insert);
 
 		tp->len += chlen;
 		tp->owrite += chlen;
 		p = tp->lb + cno;
-		if (tp->lb[cno] == '\t')
+		if (tp->lb[cno] == '\t' ||
+		    KEY_NEEDSWIDE(sp, tp->lb[cno]))
 			for (cno += chlen; chlen--;)
 				*p++ = ' ';
 		else
-			for (kp = KEY_NAME(sp, tp->lb[cno]),
+			for (kp = (u_char *)
+			    KEY_NAME(sp, tp->lb[cno]),
 			    cno += chlen; chlen--;)
 				*p++ = *kp++;
 	}
@@ -2309,7 +2294,7 @@ txt_emark(SCR *sp, TEXT *tp, size_t cno)
 static void
 txt_err(SCR *sp, TEXTH *tiqh)
 {
-	db_recno_t lno;
+	recno_t lno;
 
 	/*
 	 * The problem with input processing is that the cursor is at an
@@ -2320,7 +2305,7 @@ txt_err(SCR *sp, TEXTH *tiqh)
 	 * We depend on at least one line number being set in the text
 	 * chain.
 	 */
-	for (lno = tiqh->cqh_first->lno;
+	for (lno = TAILQ_FIRST(tiqh)->lno;
 	    !db_exist(sp, lno) && lno > 0; --lno);
 
 	sp->lno = lno == 0 ? 1 : lno;
@@ -2373,7 +2358,7 @@ txt_hex(SCR *sp, TEXT *tp)
 	/* Get the value. */
 	errno = 0;
 	value = STRTOL(wp, NULL, 16);
-	if (errno || value > MAX_CHAR_T) {
+	if (errno || value > UCHAR_MAX) {
 nothex:		tp->lb[tp->cno] = savec;
 		return (0);
 	}
@@ -2388,12 +2373,12 @@ nothex:		tp->lb[tp->cno] = savec;
 
 	/* Copy down any overwrite characters. */
 	if (tp->owrite)
-		MEMMOVEW(tp->lb + tp->cno, tp->lb + tp->cno + len, 
+		MEMMOVE(tp->lb + tp->cno, tp->lb + tp->cno + len, 
 		    tp->owrite);
 
 	/* Copy down any insert characters. */
 	if (tp->insert)
-		MEMMOVEW(tp->lb + tp->cno + tp->owrite,
+		MEMMOVE(tp->lb + tp->cno + tp->owrite,
 		    tp->lb + tp->cno + tp->owrite + len, 
 		    tp->insert);
 
@@ -2422,7 +2407,7 @@ nothex:		tp->lb[tp->cno] = savec;
 static int
 txt_insch(SCR *sp, TEXT *tp, CHAR_T *chp, u_int flags)
 {
-	char *kp;
+	u_char *kp;
 	CHAR_T savech;
 	size_t chlen, cno, copydown, olen, nlen;
 	CHAR_T *p;
@@ -2451,7 +2436,7 @@ txt_insch(SCR *sp, TEXT *tp, CHAR_T *chp, u_int flags)
 			(void)vs_columns(sp, tp->lb, tp->lno, &cno, &nlen);
 			tp->lb[cno] = savech;
 		} else
-			nlen = KEY_LEN(sp, *chp);
+			nlen = KEY_COL(sp, *chp);
 
 		/*
 		 * Eat overwrite characters until we run out of them or we've
@@ -2466,7 +2451,7 @@ txt_insch(SCR *sp, TEXT *tp, CHAR_T *chp, u_int flags)
 				(void)vs_columns(sp,
 				    tp->lb, tp->lno, &cno, &olen);
 			else
-				olen = KEY_LEN(sp, tp->lb[cno]);
+				olen = KEY_COL(sp, tp->lb[cno]);
 
 			if (olen == nlen) {
 				nlen = 0;
@@ -2479,17 +2464,18 @@ txt_insch(SCR *sp, TEXT *tp, CHAR_T *chp, u_int flags)
 				BINC_RETW(sp,
 				    tp->lb, tp->lb_len, tp->len + olen);
 				chlen = olen - nlen;
-				MEMMOVEW(tp->lb + cno + 1 + chlen,
+				MEMMOVE(tp->lb + cno + 1 + chlen,
 				    tp->lb + cno + 1, 
 				    tp->owrite + tp->insert);
 
 				tp->len += chlen;
 				tp->owrite += chlen;
-				if (tp->lb[cno] == '\t')
+				if (tp->lb[cno] == '\t' ||
+				   KEY_NEEDSWIDE(sp, tp->lb[cno]))
 					for (p = tp->lb + cno + 1; chlen--;)
 						*p++ = ' ';
 				else
-					for (kp =
+					for (kp = (u_char *)
 					    KEY_NAME(sp, tp->lb[cno]) + nlen,
 					    p = tp->lb + cno + 1; chlen--;)
 						*p++ = *kp++;
@@ -2504,7 +2490,7 @@ txt_insch(SCR *sp, TEXT *tp, CHAR_T *chp, u_int flags)
 		 * into position.
 		 */
 		if (copydown != 0 && (tp->len -= copydown) != 0)
-			MEMMOVEW(tp->lb + cno, tp->lb + cno + copydown,
+			MEMMOVE(tp->lb + cno, tp->lb + cno + copydown,
 			    tp->owrite + tp->insert + copydown);
 
 		/* If we had enough overwrite characters, we're done. */
@@ -2522,7 +2508,7 @@ txt_insch(SCR *sp, TEXT *tp, CHAR_T *chp, u_int flags)
 		if (tp->insert == 1)
 			tp->lb[tp->cno + 1] = tp->lb[tp->cno];
 		else
-			MEMMOVEW(tp->lb + tp->cno + 1,
+			MEMMOVE(tp->lb + tp->cno + 1,
 			    tp->lb + tp->cno, tp->owrite + tp->insert);
 	}
 	tp->lb[tp->cno++] = *chp;
@@ -2537,7 +2523,7 @@ static int
 txt_isrch(SCR *sp, VICMD *vp, TEXT *tp, u_int8_t *is_flagsp)
 {
 	MARK start;
-	db_recno_t lno;
+	recno_t lno;
 	u_int sf;
 
 	/* If it's a one-line screen, we don't do incrementals. */
@@ -2567,7 +2553,7 @@ txt_isrch(SCR *sp, VICMD *vp, TEXT *tp, u_int8_t *is_flagsp)
 	 * If it's a magic shell character, and not quoted, reset the cursor
 	 * to the starting point.
 	 */
-	if (strchr(O_STR(sp, O_SHELLMETA), tp->lb[tp->cno - 1]) != NULL &&
+	if (IS_SHELLMETA(sp, tp->lb[tp->cno - 1]) &&
 	    (tp->cno == 2 || tp->lb[tp->cno - 2] != '\\'))
 		vp->m_final = vp->m_start;
 
@@ -2658,7 +2644,7 @@ txt_resolve(SCR *sp, TEXTH *tiqh, u_int32_t flags)
 {
 	VI_PRIVATE *vip;
 	TEXT *tp;
-	db_recno_t lno;
+	recno_t lno;
 	int changed;
 
 	/*
@@ -2669,23 +2655,23 @@ txt_resolve(SCR *sp, TEXTH *tiqh, u_int32_t flags)
 	 * about the line will be wrong.
 	 */
 	vip = VIP(sp);
-	tp = tiqh->cqh_first;
+	tp = TAILQ_FIRST(tiqh);
 
 	if (LF_ISSET(TXT_AUTOINDENT))
 		txt_ai_resolve(sp, tp, &changed);
 	else
 		changed = 0;
 	if (db_set(sp, tp->lno, tp->lb, tp->len) ||
-	    changed && vs_change(sp, tp->lno, LINE_RESET))
+	    (changed && vs_change(sp, tp->lno, LINE_RESET)))
 		return (1);
 
-	for (lno = tp->lno; (tp = tp->q.cqe_next) != (void *)&sp->tiq; ++lno) {
+	for (lno = tp->lno; (tp = TAILQ_NEXT(tp, q)) != NULL; ++lno) {
 		if (LF_ISSET(TXT_AUTOINDENT))
 			txt_ai_resolve(sp, tp, &changed);
 		else
 			changed = 0;
 		if (db_append(sp, 0, lno, tp->lb, tp->len) ||
-		    changed && vs_change(sp, tp->lno, LINE_RESET))
+		    (changed && vs_change(sp, tp->lno, LINE_RESET)))
 			return (1);
 	}
 
@@ -2737,7 +2723,7 @@ txt_showmatch(SCR *sp, TEXT *tp)
 	cs.cs_cno = tp->cno - 1;
 	if (cs_init(sp, &cs))
 		return (1);
-	startc = (endc = cs.cs_ch)  == ')' ? '(' : '{';
+	startc = STRCHR(VIP(sp)->mcs, endc = cs.cs_ch)[-1];
 
 	/* Search for the match. */
 	for (cnt = 1;;) {
@@ -2758,7 +2744,7 @@ txt_showmatch(SCR *sp, TEXT *tp)
 	}
 
 	/* If the match is on the screen, move to it. */
-	if (cs.cs_lno < m.lno || cs.cs_lno == m.lno && cs.cs_cno < m.cno)
+	if (cs.cs_lno < m.lno || (cs.cs_lno == m.lno && cs.cs_cno < m.cno))
 		return (0);
 	sp->lno = cs.cs_lno;
 	sp->cno = cs.cs_cno;
@@ -2868,9 +2854,9 @@ txt_Rresolve(SCR *sp, TEXTH *tiqh, TEXT *tp, const size_t orig_len)
 	 * Calculate how many characters the user has entered,
 	 * plus the blanks erased by <carriage-return>/<newline>s.
 	 */
-	for (ttp = tiqh->cqh_first, input_len = 0;;) {
+	for (ttp = TAILQ_FIRST(tiqh), input_len = 0;;) {
 		input_len += ttp == tp ? tp->cno : ttp->len + ttp->R_erase;
-		if ((ttp = ttp->q.cqe_next) == (void *)&sp->tiq)
+		if ((ttp = TAILQ_NEXT(ttp, q)) == NULL)
 			break;
 	}
 
@@ -2891,9 +2877,9 @@ txt_Rresolve(SCR *sp, TEXTH *tiqh, TEXT *tp, const size_t orig_len)
 	if (input_len < orig_len) {
 		retain = MIN(tp->owrite, orig_len - input_len);
 		if (db_get(sp,
-		    tiqh->cqh_first->lno, DBG_FATAL | DBG_NOCACHE, &p, NULL))
+		    TAILQ_FIRST(tiqh)->lno, DBG_FATAL | DBG_NOCACHE, &p, NULL))
 			return;
-		MEMCPYW(tp->lb + tp->cno, p + input_len, retain);
+		MEMCPY(tp->lb + tp->cno, p + input_len, retain);
 		tp->len -= tp->owrite - retain;
 		tp->owrite = 0;
 		tp->insert += retain;
