@@ -9,10 +9,6 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: vi.c,v 10.73 2002/04/11 19:49:30 skimo Exp $ (Berkeley) $Date: 2002/04/11 19:49:30 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/time.h>
@@ -34,16 +30,16 @@ typedef enum {
 } gcret_t;
 
 static VIKEYS const
-	       *v_alias __P((SCR *, VICMD *, VIKEYS const *));
-static gcret_t	v_cmd __P((SCR *, VICMD *, VICMD *, VICMD *, int *, int *));
-static int	v_count __P((SCR *, VICMD *, ARG_CHAR_T, u_long *));
-static void	v_dtoh __P((SCR *));
-static int	v_init __P((SCR *));
-static gcret_t	v_key __P((SCR *, VICMD *, int, u_int32_t));
-static int	v_motion __P((SCR *, VICMD *, VICMD *, int *));
+	       *v_alias(SCR *, VICMD *, VIKEYS const *);
+static gcret_t	v_cmd(SCR *, VICMD *, VICMD *, VICMD *, int *, int *);
+static int	v_count(SCR *, ARG_CHAR_T, u_long *);
+static void	v_dtoh(SCR *);
+static int	v_init(SCR *);
+static gcret_t	v_key(SCR *, int, EVENT *, u_int32_t);
+static int	v_motion(SCR *, VICMD *, VICMD *, int *);
 
 #if defined(DEBUG) && defined(COMLOG)
-static void	v_comlog __P((SCR *, VICMD *));
+static void	v_comlog(SCR *, VICMD *);
 #endif
 
 /*
@@ -58,27 +54,24 @@ static void	v_comlog __P((SCR *, VICMD *));
  * vi --
  * 	Main vi command loop.
  *
- * PUBLIC: int vi __P((SCR **));
+ * PUBLIC: int vi(SCR **);
  */
 int
 vi(SCR **spp)
 {
 	GS *gp;
-	WIN *wp;
 	MARK abs;
 	SCR *next, *sp;
-	VICMD cmd, *vp;
+	VICMD cmd = { 0 }, *vp;
 	VI_PRIVATE *vip;
 	int comcount, mapped, rval;
 
 	/* Get the first screen. */
 	sp = *spp;
-	wp = sp->wp;
 	gp = sp->gp;
 
-	/* Initialize the command structure. */
+	/* Point to the command structure. */
 	vp = &cmd;
-	memset(vp, 0, sizeof(VICMD));
 
 	/* Reset strange attraction. */
 	F_SET(vp, VM_RCM_SET);
@@ -131,7 +124,7 @@ vi(SCR **spp)
 		 * only because we exited a screen or file.  In this case,
 		 * we simply go back into the ex parser.
 		 */
-		if (EXCMD_RUNNING(wp)) {
+		if (EXCMD_RUNNING(gp)) {
 			vp->kp = &vikeys[':'];
 			goto ex_continue;
 		}
@@ -157,11 +150,12 @@ vi(SCR **spp)
 			goto err;
 		case GC_ERR_NOFLUSH:
 			goto gc_err_noflush;
+		case GC_EVENT:
+			goto gc_event;
 		case GC_FATAL:
 			goto ret;
 		case GC_INTERRUPT:
 			goto intr;
-		case GC_EVENT:
 		case GC_OK:
 			break;
 		}
@@ -231,10 +225,11 @@ vi(SCR **spp)
 		/* Call the function. */
 ex_continue:	if (vp->kp->func(sp, vp))
 			goto err;
+gc_event:
 #ifdef DEBUG
 		/* Make sure no function left the temporary space locked. */
-		if (F_ISSET(wp, W_TMP_INUSE)) {
-			F_CLR(wp, W_TMP_INUSE);
+		if (F_ISSET(gp, G_TMP_INUSE)) {
+			F_CLR(gp, G_TMP_INUSE);
 			msgq(sp, M_ERR,
 			    "232|vi: temporary buffer not released");
 		}
@@ -343,9 +338,9 @@ ex_continue:	if (vp->kp->func(sp, vp))
 		 * command, since the tag may be moving to the same file.
 		 */
 		if ((F_ISSET(vp, V_ABS) ||
-		    F_ISSET(vp, V_ABS_L) && sp->lno != abs.lno ||
-		    F_ISSET(vp, V_ABS_C) &&
-		    (sp->lno != abs.lno || sp->cno != abs.cno)) &&
+		    (F_ISSET(vp, V_ABS_L) && sp->lno != abs.lno) ||
+		    (F_ISSET(vp, V_ABS_C) &&
+		    (sp->lno != abs.lno || sp->cno != abs.cno))) &&
 		    mark_set(sp, ABSMARK1, &abs, 1))
 			goto err;
 
@@ -411,13 +406,13 @@ ret:		rval = 1;
 }
 
 #define	KEY(key, ec_flags) {						\
-	if ((gcret = v_key(sp, vp, 0, ec_flags)) != GC_OK)		\
+	if ((gcret = v_key(sp, 0, &ev, ec_flags)) != GC_OK)		\
 		return (gcret);						\
-	if (vp->ev.e_value == K_ESCAPE)					\
+	if (ev.e_value == K_ESCAPE)					\
 		goto esc;						\
-	if (FL_ISSET(vp->ev.e_flags, CH_MAPPED))			\
+	if (F_ISSET(&ev.e_ch, CH_MAPPED))				\
 		*mappedp = 1;						\
-	key = vp->ev.e_c;						\
+	key = ev.e_c;							\
 }
 
 /*
@@ -437,48 +432,36 @@ VIKEYS const tmotion = {
 
 /*
  * v_cmd --
- *	Get a vi command.
+ *
+ * The command structure for vi is less complex than ex (and don't think
+ * I'm not grateful!)  The command syntax is:
+ *
+ *	[count] [buffer] [count] key [[motion] | [buffer] [character]]
+ *
+ * and there are several special cases.  The motion value is itself a vi
+ * command, with the syntax:
+ *
+ *	[count] key [character]
  */
 static gcret_t
-v_cmd(SCR *sp, VICMD *dp, VICMD *vp, VICMD *ismotion, int *comcountp, int *mappedp)
-	        
-	               
-	                	/* Previous key if getting motion component. */
-	                         
+v_cmd(
+	SCR *sp,
+	VICMD *dp,
+	VICMD *vp,
+	VICMD *ismotion,	/* Previous key if getting motion component. */
+	int *comcountp,
+	int *mappedp)
 {
 	enum { COMMANDMODE, ISPARTIAL, NOTPARTIAL } cpart;
-	CHAR_T key;
+	EVENT ev;
 	VIKEYS const *kp;
 	gcret_t gcret;
 	u_int flags;
+	CHAR_T key;
 	char *s;
 
 	/*
-	 * Get an event command or a key.  Event commands are simple, and
-	 * don't have any additional information.
-	 */
-	cpart = ismotion == NULL ? COMMANDMODE : ISPARTIAL;
-	gcret = v_key(sp, vp, 1, EC_MAPCOMMAND);
-	if (gcret != GC_OK) {
-		if (gcret != GC_EVENT)
-			return (gcret);
-		if (v_event(sp, vp))
-			return (GC_ERR);
-		if (ismotion != NULL && !F_ISSET(vp->kp, V_MOVE))
-			v_event_err(sp, &vp->ev);
-		return (GC_EVENT);
-	}
-
-	/*
-	 * Keys are not simple.  (Although vi's command structure less complex
-	 * than ex (and don't think I'm not grateful!)  The command syntax is:
-	 *
-	 *	[count] [buffer] [count] key [[motion] | [buffer] [character]]
-	 *
-	 * and there are, of course, several special cases.  The motion value
-	 * is itself a vi command, with the syntax:
-	 *
-	 *	[count] key [character]
+	 * Get a key.
 	 *
 	 * <escape> cancels partial commands, i.e. a command where at least
 	 * one non-numeric character has been entered.  Otherwise, it beeps
@@ -489,16 +472,18 @@ v_cmd(SCR *sp, VICMD *dp, VICMD *vp, VICMD *ismotion, int *comcountp, int *mappe
 	 * all that's been entered is a number, requiring that the terminal
 	 * be alerted.
 	 */
-	if (vp->ev.e_value == K_ESCAPE)
+	cpart = ismotion == NULL ? COMMANDMODE : ISPARTIAL;
+	if ((gcret =
+	    v_key(sp, ismotion == NULL, &ev, EC_MAPCOMMAND)) != GC_OK) {
+		if (gcret == GC_EVENT)
+			vp->ev = ev;
+		return (gcret);
+	}
+	if (ev.e_value == K_ESCAPE)
 		goto esc;
-
-	/*
-	 * Commands that are mapped are treated differently (e.g., they
-	 * don't set the dot command.  Pass that information back.
-	 */
-	if (FL_ISSET(vp->ev.e_flags, CH_MAPPED))
+	if (F_ISSET(&ev.e_ch, CH_MAPPED))
 		*mappedp = 1;
-	key = vp->ev.e_c;
+	key = ev.e_c;
 
 	if (ismotion == NULL)
 		cpart = NOTPARTIAL;
@@ -517,14 +502,12 @@ v_cmd(SCR *sp, VICMD *dp, VICMD *vp, VICMD *ismotion, int *comcountp, int *mappe
 	}
 
 	/*
-	 * Pick up an optional count, where a leading 0 isn't a count, it's
-	 * a command.  When a count is specified, the dot command behaves
-	 * differently, pass the information back.
+	 * Pick up an optional count, where a leading 0 is not a count,
+	 * it's a command.
 	 */
 	if (ISDIGIT(key) && key != '0') {
-		if (v_count(sp, vp, key, &vp->count))
+		if (v_count(sp, key, &vp->count))
 			return (GC_ERR);
-
 		F_SET(vp, VC_C1SET);
 		*comcountp = 1;
 
@@ -567,11 +550,7 @@ v_cmd(SCR *sp, VICMD *dp, VICMD *vp, VICMD *ismotion, int *comcountp, int *mappe
 		F_CLR(vp, VC_C1SET);
 	}
 
-	/*
-	 * There are several commands that we implement as aliases, both
-	 * to match historic practice and to ensure consistency.  Check
-	 * for command aliases.
-	 */
+	/* Check for command aliases. */
 	if (kp->func == NULL && (kp = v_alias(sp, vp, kp)) == NULL)
 		return (GC_ERR);
 
@@ -591,8 +570,7 @@ v_cmd(SCR *sp, VICMD *dp, VICMD *vp, VICMD *ismotion, int *comcountp, int *mappe
 	if (kp->func == NULL) {
 		if (key != '.') {
 			v_emsg(sp, KEY_NAME(sp, key),
-			    vp->ev.e_value == K_ESCAPE ?
-			    VIM_NOCOM_B : VIM_NOCOM);
+			    ev.e_value == K_ESCAPE ? VIM_NOCOM_B : VIM_NOCOM);
 			return (GC_ERR);
 		}
 
@@ -686,7 +664,7 @@ usage:			if (ismotion == NULL)
 	if (vp->key == 'z') {
 		KEY(vp->character, 0);
 		if (ISDIGIT(vp->character)) {
-			if (v_count(sp, vp, vp->character, &vp->count2))
+			if (v_count(sp, vp->character, &vp->count2))
 				return (GC_ERR);
 			F_SET(vp, VC_C2SET);
 			KEY(vp->character, 0);
@@ -732,10 +710,13 @@ esc:	switch (cpart) {
  * Get resulting motion mark.
  */
 static int
-v_motion(SCR *sp, VICMD *dm, VICMD *vp, int *mappedp)
+v_motion(
+	SCR *sp,
+	VICMD *dm,
+	VICMD *vp,
+	int *mappedp)
 {
 	VICMD motion;
-	gcret_t gcret;
 	size_t len;
 	u_long cnt;
 	u_int flags;
@@ -750,11 +731,9 @@ v_motion(SCR *sp, VICMD *dm, VICMD *vp, int *mappedp)
 		motion = *dm;
 		F_SET(&motion, VC_ISDOT);
 		F_CLR(&motion, VM_COMMASK);
-		gcret = GC_OK;
 	} else {
 		memset(&motion, 0, sizeof(VICMD));
-		gcret = v_cmd(sp, NULL, &motion, vp, &notused, mappedp);
-		if (gcret != GC_OK && gcret != GC_EVENT)
+		if (v_cmd(sp, NULL, &motion, vp, &notused, mappedp) != GC_OK)
 			return (1);
 	}
 
@@ -784,7 +763,7 @@ v_motion(SCR *sp, VICMD *dm, VICMD *vp, int *mappedp)
 	 * appropriately.  If not a doubled command, run the function to get
 	 * the resulting mark.
  	 */
-	if (gcret != GC_EVENT && vp->key == motion.key) {
+	if (vp->key == motion.key) {
 		F_SET(vp, VM_LDOUBLE | VM_LMODE);
 
 		/* Set the origin of the command. */
@@ -801,7 +780,7 @@ v_motion(SCR *sp, VICMD *dm, VICMD *vp, int *mappedp)
 		vp->m_stop.lno = sp->lno + motion.count - 1;
 		if (db_get(sp, vp->m_stop.lno, 0, NULL, &len)) {
 			if (vp->m_stop.lno != 1 ||
-			   vp->key != 'c' && vp->key != '!') {
+			   (vp->key != 'c' && vp->key != '!')) {
 				v_emsg(sp, NULL, VIM_EMPTY);
 				return (1);
 			}
@@ -873,7 +852,7 @@ v_motion(SCR *sp, VICMD *dm, VICMD *vp, int *mappedp)
 		 */
 		if (!db_exist(sp, vp->m_stop.lno)) {
 			if (vp->m_stop.lno != 1 ||
-			   vp->key != 'c' && vp->key != '!') {
+			   (vp->key != 'c' && vp->key != '!')) {
 				v_emsg(sp, NULL, VIM_EMPTY);
 				return (1);
 			}
@@ -917,8 +896,8 @@ v_motion(SCR *sp, VICMD *dm, VICMD *vp, int *mappedp)
 		 * Motions are from the from MARK to the to MARK (inclusive).
 		 */
 		if (motion.m_start.lno > motion.m_stop.lno ||
-		    motion.m_start.lno == motion.m_stop.lno &&
-		    motion.m_start.cno > motion.m_stop.cno) {
+		    (motion.m_start.lno == motion.m_stop.lno &&
+		    motion.m_start.cno > motion.m_stop.cno)) {
 			vp->m_start = motion.m_stop;
 			vp->m_stop = motion.m_start;
 		} else {
@@ -979,7 +958,7 @@ v_init(SCR *sp)
 			sp->t_minrows = sp->t_rows = sp->rows - 1;
 			msgq(sp, M_INFO,
 			    "214|Windows option value is too large, max is %u",
-			    sp->t_rows);
+			    (u_int)sp->t_rows);
 		}
 		sp->t_maxrows = sp->rows - 1;
 	} else
@@ -987,7 +966,7 @@ v_init(SCR *sp)
 	sp->roff = sp->coff = 0;
 
 	/* Create a screen map. */
-	CALLOC_RET(sp, HMAP, SMAP *, SIZE_HMAP(sp), sizeof(SMAP));
+	CALLOC_RET(sp, HMAP, SIZE_HMAP(sp), sizeof(SMAP));
 	TMAP = HMAP + (sp->t_rows - 1);
 	HMAP->lno = sp->lno;
 	HMAP->coff = 0;
@@ -999,7 +978,6 @@ v_init(SCR *sp)
 	 * line in the middle, otherwise, it won't work and we'll end up with
 	 * the line at the top.
 	 */
-	F_CLR(sp, SC_SCR_TOP);
 	F_SET(sp, SC_SCR_REFORMAT | SC_SCR_CENTER);
 
 	/* Invalidate the cursor. */
@@ -1020,27 +998,22 @@ v_dtoh(SCR *sp)
 {
 	GS *gp;
 	SCR *tsp;
-	WIN *wp;
 	int hidden;
 
 	/* Move all screens to the hidden queue, tossing screen maps. */
-	for (hidden = 0, gp = sp->gp, wp = sp->wp;
-	    (tsp = wp->scrq.cqh_first) != (void *)&wp->scrq; ++hidden) {
-		if (_HMAP(tsp) != NULL) {
-			free(_HMAP(tsp));
-			_HMAP(tsp) = NULL;
-		}
-		CIRCLEQ_REMOVE(&wp->scrq, tsp, q);
-		CIRCLEQ_INSERT_TAIL(&gp->hq, tsp, q);
+	for (hidden = 0, gp = sp->gp;
+	    (tsp = TAILQ_FIRST(gp->dq)) != NULL; ++hidden) {
+		free(_HMAP(tsp));
+		_HMAP(tsp) = NULL;
+		TAILQ_REMOVE(gp->dq, tsp, q);
+		TAILQ_INSERT_TAIL(gp->hq, tsp, q);
 		/* XXXX Change if hidden screens per window */
-		tsp->wp = 0;
 		gp->scr_discard(tsp, NULL);
 	}
 
 	/* Move current screen back to the display queue. */
-	CIRCLEQ_REMOVE(&gp->hq, sp, q);
-	CIRCLEQ_INSERT_TAIL(&wp->scrq, sp, q);
-	sp->wp = wp;
+	TAILQ_REMOVE(gp->hq, sp, q);
+	TAILQ_INSERT_TAIL(gp->dq, sp, q);
 
 	if (hidden > 1)
 		msgq(sp, M_INFO,
@@ -1050,16 +1023,16 @@ v_dtoh(SCR *sp)
 
 /*
  * v_curword --
- *	Get the word (or non-word) the cursor is on.
+ *	Get the word (tagstring, actually) the cursor is on.
  *
- * PUBLIC: int v_curword __P((SCR *));
+ * PUBLIC: int v_curword(SCR *);
  */
 int
 v_curword(SCR *sp)
 {
 	VI_PRIVATE *vip;
 	size_t beg, end, len;
-	int moved, state;
+	int moved;
 	CHAR_T *p;
 
 	if (db_get(sp, sp->lno, DBG_FATAL, &p, &len))
@@ -1081,7 +1054,7 @@ v_curword(SCR *sp)
 	 * follow the same rule.
 	 */
 	for (moved = 0,
-	    beg = sp->cno; beg < len && isspace(p[beg]); moved = 1, ++beg);
+	    beg = sp->cno; beg < len && ISSPACE(p[beg]); moved = 1, ++beg);
 	if (beg >= len) {
 		msgq(sp, M_BERR, "212|Cursor not in a word");
 		return (1);
@@ -1091,14 +1064,19 @@ v_curword(SCR *sp)
 		(void)vs_refresh(sp, 0);
 	}
 
-	/* Find the end of the word. */
-	for (state = inword(p[beg]),
-	    end = beg; ++end < len && state == inword(p[end]););
+	/*
+	 * Find the end of the word.
+	 *
+	 * !!!
+	 * Historically, vi accepted any non-blank as initial character
+	 * when building up a tagstring.  Required by IEEE 1003.1-2001.
+	 */
+	for (end = beg; ++end < len && inword(p[end]););
 
 	vip = VIP(sp);
-	len = (end - beg);
-	BINC_RETW(sp, vip->keyw, vip->klen, len+1);
-	MEMMOVEW(vip->keyw, p + beg, len);
+	vip->klen = len = (end - beg);
+	BINC_RETW(sp, vip->keyw, vip->keywlen, len+1);
+	MEMMOVE(vip->keyw, p + beg, len);
 	vip->keyw[len] = '\0';				/* XXX */
 	return (0);
 }
@@ -1108,7 +1086,10 @@ v_curword(SCR *sp)
  *	Check for a command alias.
  */
 static VIKEYS const *
-v_alias(SCR *sp, VICMD *vp, const VIKEYS *kp)
+v_alias(
+	SCR *sp,
+	VICMD *vp,
+	VIKEYS const *kp)
 {
 	CHAR_T push;
 
@@ -1141,33 +1122,37 @@ v_alias(SCR *sp, VICMD *vp, const VIKEYS *kp)
  *	Return the next count.
  */
 static int
-v_count(SCR *sp, VICMD *vp, ARG_CHAR_T fkey, u_long *countp)
+v_count(
+	SCR *sp,
+	ARG_CHAR_T fkey,
+	u_long *countp)
 {
+	EVENT ev;
 	u_long count, tc;
 
-	vp->ev.e_c = fkey;
+	ev.e_c = fkey;
 	count = tc = 0;
 	do {
 		/*
 		 * XXX
 		 * Assume that overflow results in a smaller number.
 		 */
-		tc = count * 10 + vp->ev.e_c - '0';
+		tc = count * 10 + ev.e_c - '0';
 		if (count > tc) {
 			/* Toss to the next non-digit. */
 			do {
-				if (v_key(sp, vp, 0,
+				if (v_key(sp, 0, &ev,
 				    EC_MAPCOMMAND | EC_MAPNODIGIT) != GC_OK)
 					return (1);
-			} while (ISDIGIT(vp->ev.e_c));
+			} while (ISDIGIT(ev.e_c));
 			msgq(sp, M_ERR,
 			    "235|Number larger than %lu", ULONG_MAX);
 			return (1);
 		}
 		count = tc;
-		if (v_key(sp, vp, 0, EC_MAPCOMMAND | EC_MAPNODIGIT) != GC_OK)
+		if (v_key(sp, 0, &ev, EC_MAPCOMMAND | EC_MAPNODIGIT) != GC_OK)
 			return (1);
-	} while (ISDIGIT(vp->ev.e_c));
+	} while (ISDIGIT(ev.e_c));
 	*countp = count;
 	return (0);
 }
@@ -1177,12 +1162,15 @@ v_count(SCR *sp, VICMD *vp, ARG_CHAR_T fkey, u_long *countp)
  *	Return the next event.
  */
 static gcret_t
-v_key(SCR *sp, VICMD *vp, int events_ok, u_int32_t ec_flags)
+v_key(
+	SCR *sp,
+	int command_events,
+	EVENT *evp,
+	u_int32_t ec_flags)
 {
-	EVENT *evp;
 	u_int32_t quote;
 
-	for (evp = &vp->ev, quote = 0;;) {
+	for (quote = 0;;) {
 		if (v_event_get(sp, evp, 0, ec_flags | quote))
 			return (GC_FATAL);
 		quote = 0;
@@ -1211,28 +1199,17 @@ v_key(SCR *sp, VICMD *vp, int events_ok, u_int32_t ec_flags)
 			 *
 			 * Historically, vi exited to ex mode if no file was
 			 * named on the command line, and two interrupts were
-			 * generated in a row.  (I figured you might want to
-			 * know that, just in case there's a quiz later.)
+			 * generated in a row.  (Just figured you might want
+			 * to know that.)
 			 */
 			(void)sp->gp->scr_bell(sp);
 			return (GC_INTERRUPT);
 		case E_REPAINT:
-			if (v_erepaint(sp, evp))
+			if (vs_repaint(sp, evp))
 				return (GC_FATAL);
 			break;
 		case E_WRESIZE:
-			/*
-			 * !!!
-			 * We don't do anything here, just return an error.
-			 * The vi loop will return because of this, and then
-			 * the main loop will realize that we had to restart
-			 * the world and will call the vi loop again.
-			 */
 			return (GC_ERR);
-		case E_IPCOMMAND:
-			if (events_ok)
-				return (GC_EVENT);
-			/* FALLTHROUGH */
 		default:
 			v_event_err(sp, evp);
 			return (GC_ERR);
@@ -1247,17 +1224,17 @@ v_key(SCR *sp, VICMD *vp, int events_ok, u_int32_t ec_flags)
  *	Log the contents of the command structure.
  */
 static void
-v_comlog(sp, vp)
-	SCR *sp;
-	VICMD *vp;
+v_comlog(
+	SCR *sp,
+	VICMD *vp)
 {
-	vtrace(sp, "vcmd: %c", vp->key);
+	TRACE(sp, "vcmd: "WC, vp->key);
 	if (F_ISSET(vp, VC_BUFFER))
-		vtrace(sp, " buffer: %c", vp->buffer);
+		TRACE(sp, " buffer: "WC, vp->buffer);
 	if (F_ISSET(vp, VC_C1SET))
-		vtrace(sp, " c1: %lu", vp->count);
+		TRACE(sp, " c1: %lu", vp->count);
 	if (F_ISSET(vp, VC_C2SET))
-		vtrace(sp, " c2: %lu", vp->count2);
-	vtrace(sp, " flags: 0x%x\n", vp->flags);
+		TRACE(sp, " c2: %lu", vp->count2);
+	TRACE(sp, " flags: 0x%x\n", vp->flags);
 }
 #endif
