@@ -9,10 +9,6 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: vs_msg.c,v 10.85 2001/07/29 19:07:31 skimo Exp $ (Berkeley) $Date: 2001/07/29 19:07:31 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/time.h>
@@ -37,11 +33,11 @@ typedef enum {
 					 */
 } sw_t;
 
-static void	vs_divider __P((SCR *));
-static void	vs_msgsave __P((SCR *, mtype_t, char *, size_t));
-static void	vs_output __P((SCR *, mtype_t, const char *, int));
-static void	vs_scroll __P((SCR *, int *, sw_t));
-static void	vs_wait __P((SCR *, int *, sw_t));
+static void	vs_divider(SCR *);
+static void	vs_msgsave(SCR *, mtype_t, char *, size_t);
+static void	vs_output(SCR *, mtype_t, const char *, int);
+static void	vs_scroll(SCR *, int *, sw_t);
+static void	vs_wait(SCR *, int *, sw_t);
 
 /*
  * vs_busy --
@@ -53,7 +49,7 @@ static void	vs_wait __P((SCR *, int *, sw_t));
  * messages, e.g. X11 clock icons, should set their scr_busy function to the
  * correct function before calling the main editor routine.
  *
- * PUBLIC: void vs_busy __P((SCR *, const char *, busy_t));
+ * PUBLIC: void vs_busy(SCR *, const char *, busy_t);
  */
 void
 vs_busy(SCR *sp, const char *msg, busy_t btype)
@@ -61,7 +57,8 @@ vs_busy(SCR *sp, const char *msg, busy_t btype)
 	GS *gp;
 	VI_PRIVATE *vip;
 	static const char flagc[] = "|/-\\";
-	struct timeval tv;
+	struct timespec ts, ts_diff;
+	const struct timespec ts_min = { 0, 125000000 };
 	size_t len, notused;
 	const char *p;
 
@@ -86,7 +83,7 @@ vs_busy(SCR *sp, const char *msg, busy_t btype)
 
 		/* Initialize state for updates. */
 		vip->busy_ch = 0;
-		(void)gettimeofday(&vip->busy_tv, NULL);
+		timepoint_steady(&vip->busy_ts);
 
 		/* Save the current cursor. */
 		(void)gp->scr_cursor(sp, &vip->busy_oldy, &vip->busy_oldx);
@@ -119,11 +116,12 @@ vs_busy(SCR *sp, const char *msg, busy_t btype)
 			break;
 
 		/* Update no more than every 1/8 of a second. */
-		(void)gettimeofday(&tv, NULL);
-		if (((tv.tv_sec - vip->busy_tv.tv_sec) * 1000000 +
-		    (tv.tv_usec - vip->busy_tv.tv_usec)) < 125000)
+		timepoint_steady(&ts);
+		ts_diff = ts;
+		timespecsub(&ts_diff, &vip->busy_ts);
+		if (timespeccmp(&ts_diff, &ts_min, <))
 			return;
-		vip->busy_tv = tv;
+		vip->busy_ts = ts;
 
 		/* Display the update. */
 		if (vip->busy_ch == sizeof(flagc) - 1)
@@ -140,7 +138,7 @@ vs_busy(SCR *sp, const char *msg, busy_t btype)
  * vs_home --
  *	Home the cursor to the bottom row, left-most column.
  *
- * PUBLIC: void vs_home __P((SCR *));
+ * PUBLIC: void vs_home(SCR *);
  */
 void
 vs_home(SCR *sp)
@@ -153,7 +151,7 @@ vs_home(SCR *sp)
  * vs_update --
  *	Update a command.
  *
- * PUBLIC: void vs_update __P((SCR *, const char *, const CHAR_T *));
+ * PUBLIC: void vs_update(SCR *, const char *, const CHAR_T *);
  */
 void
 vs_update(SCR *sp, const char *m1, const CHAR_T *m2)
@@ -223,7 +221,7 @@ vs_update(SCR *sp, const char *m1, const CHAR_T *m2)
  * alternate method of displaying messages, e.g. dialog boxes, should set their
  * scr_msg function to the correct function before calling the editor.
  *
- * PUBLIC: void vs_msg __P((SCR *, mtype_t, char *, size_t));
+ * PUBLIC: void vs_msg(SCR *, mtype_t, char *, size_t);
  */
 void
 vs_msg(SCR *sp, mtype_t mtype, char *line, size_t len)
@@ -350,16 +348,16 @@ vs_msg(SCR *sp, mtype_t mtype, char *line, size_t len)
 		}
 	vip->mtype = mtype;
 	for (s = line;; s = t) {
-		for (; len > 0 && isblank(*s); --len, ++s);
+		for (; len > 0 && isblank((u_char)*s); --len, ++s);
 		if (len == 0)
 			break;
 		if (len + vip->lcontinue > maxcols) {
 			for (e = s + (maxcols - vip->lcontinue);
-			    e > s && !isblank(*e); --e);
+			    e > s && !isblank((u_char)*e); --e);
 			if (e == s)
 				 e = t = s + (maxcols - vip->lcontinue);
 			else
-				for (t = e; isblank(e[-1]); --e);
+				for (t = e; isblank((u_char)e[-1]); --e);
 		} else
 			e = t = s + len;
 
@@ -390,11 +388,10 @@ ret:	(void)gp->scr_move(sp, oldy, oldx);
 static void
 vs_output(SCR *sp, mtype_t mtype, const char *line, int llen)
 {
-	char *kp;
 	GS *gp;
 	VI_PRIVATE *vip;
-	size_t chlen, notused;
-	int ch, len, rlen, tlen;
+	size_t notused;
+	int len, rlen, tlen;
 	const char *p, *t;
 	char *cbp, *ecbp, cbuf[128];
 
@@ -462,7 +459,6 @@ vs_output(SCR *sp, mtype_t mtype, const char *line, int llen)
 }
 		ecbp = (cbp = cbuf) + sizeof(cbuf) - 1;
 		for (t = line, tlen = len; tlen--; ++t) {
-			ch = *t;
 			/*
 			 * Replace tabs with spaces, there are places in
 			 * ex that do column calculations without looking
@@ -470,13 +466,9 @@ vs_output(SCR *sp, mtype_t mtype, const char *line, int llen)
 			 * <tabs> do their own expansions.  This catches
 			 * <tabs> in things like tag search strings.
 			 */
-			if (ch == '\t')
-				ch = ' ';
-			chlen = KEY_LEN(sp, ch);
-			if (cbp + chlen >= ecbp)
+			if (cbp + 1 >= ecbp)
 				FLUSH;
-			for (kp = KEY_NAME(sp, ch); chlen--;)
-				*cbp++ = *kp++;
+			*cbp++ = *t == '\t' ? ' ' : *t;
 		}
 		if (cbp > cbuf)
 			FLUSH;
@@ -510,7 +502,7 @@ vs_output(SCR *sp, mtype_t mtype, const char *line, int llen)
  * This routine is called when exiting a colon command to resolve any ex
  * output that may have occurred.
  *
- * PUBLIC: int vs_ex_resolve __P((SCR *, int *));
+ * PUBLIC: int vs_ex_resolve(SCR *, int *);
  */
 int
 vs_ex_resolve(SCR *sp, int *continuep)
@@ -586,7 +578,7 @@ vs_ex_resolve(SCR *sp, int *continuep)
 	 * If we're not the bottom of the split screen stack, the screen
 	 * image itself is wrong, so redraw everything.
 	 */
-	if (sp->q.cqe_next != (void *)&sp->wp->scrq)
+	if (TAILQ_NEXT(sp, q) != NULL)
 		F_SET(sp, SC_SCR_REDRAW);
 
 	/* If ex changed the underlying file, the map itself is wrong. */
@@ -622,7 +614,7 @@ vs_ex_resolve(SCR *sp, int *continuep)
 		vip->linecount = vip->lcontinue = vip->totalcount = 0;
 
 		/* Redraw. */
-		(void)v_erepaint(sp, &ev);
+		(void)vs_repaint(sp, &ev);
 	} else
 		/* Reset the count of overwriting lines. */
 		vip->linecount = vip->lcontinue = vip->totalcount = 0;
@@ -634,14 +626,13 @@ vs_ex_resolve(SCR *sp, int *continuep)
  * vs_resolve --
  *	Deal with message output.
  *
- * PUBLIC: int vs_resolve __P((SCR *, SCR *, int));
+ * PUBLIC: int vs_resolve(SCR *, SCR *, int);
  */
 int
 vs_resolve(SCR *sp, SCR *csp, int forcewait)
 {
 	EVENT ev;
 	GS *gp;
-	WIN *wp;
 	MSGS *mp;
 	VI_PRIVATE *vip;
 	size_t oldy, oldx;
@@ -655,7 +646,6 @@ vs_resolve(SCR *sp, SCR *csp, int forcewait)
 	 * not, csp is the current screen, used for final cursor positioning.
 	 */
 	gp = sp->gp;
-	wp = sp->wp;
 	vip = VIP(sp);
 	if (csp == NULL)
 		csp = sp;
@@ -684,12 +674,12 @@ vs_resolve(SCR *sp, SCR *csp, int forcewait)
 	 * messages.)  Once this is done, don't trust the cursor.  That
 	 * extra refresh screwed the pooch.
 	 */
-	if (gp->msgq.lh_first != NULL) {
+	if (!SLIST_EMPTY(gp->msgq)) {
 		if (!F_ISSET(sp, SC_SCR_VI) && vs_refresh(sp, 1))
 			return (1);
-		while ((mp = gp->msgq.lh_first) != NULL) {
-			wp->scr_msg(sp, mp->mtype, mp->buf, mp->len);
-			LIST_REMOVE(mp, q);
+		while ((mp = SLIST_FIRST(gp->msgq)) != NULL) {
+			gp->scr_msg(sp, mp->mtype, mp->buf, mp->len);
+			SLIST_REMOVE_HEAD(gp->msgq, q);
 			free(mp->buf);
 			free(mp);
 		}
@@ -733,7 +723,7 @@ vs_resolve(SCR *sp, SCR *csp, int forcewait)
 
 	/* Redraw. */
 	if (redraw)
-		(void)v_erepaint(sp, &ev);
+		(void)vs_repaint(sp, &ev);
 
 	/* Restore the cursor position. */
 	(void)gp->scr_move(csp, oldy, oldx);
@@ -764,7 +754,7 @@ vs_scroll(SCR *sp, int *continuep, sw_t wtype)
 		(void)gp->scr_deleteln(sp);
 
 		/* If there are screens below us, push them back into place. */
-		if (sp->q.cqe_next != (void *)&sp->wp->scrq) {
+		if (TAILQ_NEXT(sp, q) != NULL) {
 			(void)gp->scr_move(sp, LASTLINE(sp), 0);
 			(void)gp->scr_insertln(sp);
 		}
@@ -881,24 +871,25 @@ vs_msgsave(SCR *sp, mtype_t mt, char *p, size_t len)
 	 * allocate memory here, we're genuinely screwed, dump the message
 	 * to stderr in the (probably) vain hope that someone will see it.
 	 */
-	CALLOC_GOTO(sp, mp_n, MSGS *, 1, sizeof(MSGS));
-	MALLOC_GOTO(sp, mp_n->buf, char *, len);
+	CALLOC_GOTO(sp, mp_n, 1, sizeof(MSGS));
+	MALLOC_GOTO(sp, mp_n->buf, len);
 
 	memmove(mp_n->buf, p, len);
 	mp_n->len = len;
 	mp_n->mtype = mt;
 
 	gp = sp->gp;
-	if ((mp_c = gp->msgq.lh_first) == NULL) {
-		LIST_INSERT_HEAD(&gp->msgq, mp_n, q);
+	if (SLIST_EMPTY(gp->msgq)) {
+		SLIST_INSERT_HEAD(gp->msgq, mp_n, q);
 	} else {
-		for (; mp_c->q.le_next != NULL; mp_c = mp_c->q.le_next);
-		LIST_INSERT_AFTER(mp_c, mp_n, q);
+		SLIST_FOREACH(mp_c, gp->msgq, q)
+			if (SLIST_NEXT(mp_c, q) == NULL)
+				break;
+		SLIST_INSERT_AFTER(mp_c, mp_n, q);
 	}
 	return;
 
 alloc_err:
-	if (mp_n != NULL)
-		free(mp_n);
+	free(mp_n);
 	(void)fprintf(stderr, "%.*s\n", (int)len, p);
 }
