@@ -9,16 +9,13 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: vs_refresh.c,v 10.50 2001/06/25 15:19:37 skimo Exp $ (Berkeley) $Date: 2001/06/25 15:19:37 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/time.h>
 
 #include <bitstring.h>
 #include <ctype.h>
+#include <libgen.h>
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -30,21 +27,45 @@ static const char sccsid[] = "$Id: vs_refresh.c,v 10.50 2001/06/25 15:19:37 skim
 #define	UPDATE_CURSOR	0x01			/* Update the cursor. */
 #define	UPDATE_SCREEN	0x02			/* Flush to screen. */
 
-static void	vs_modeline __P((SCR *));
-static int	vs_paint __P((SCR *, u_int));
+static void	vs_modeline(SCR *);
+static int	vs_paint(SCR *, u_int);
+
+/*
+ * v_repaint --
+ *	Repaint selected lines from the screen.
+ *
+ * PUBLIC: int vs_repaint(SCR *, EVENT *);
+ */
+int
+vs_repaint(
+	SCR *sp,
+	EVENT *evp)
+{
+	SMAP *smp;
+
+	for (; evp->e_flno <= evp->e_tlno; ++evp->e_flno) {
+		smp = HMAP + evp->e_flno - 1;
+		SMAP_FLUSH(smp);
+		if (vs_line(sp, smp, NULL, NULL))
+			return (1);
+	}
+	return (0);
+}
 
 /*
  * vs_refresh --
  *	Refresh all screens.
  *
- * PUBLIC: int vs_refresh __P((SCR *, int));
+ * PUBLIC: int vs_refresh(SCR *, int);
  */
 int
-vs_refresh(SCR *sp, int forcepaint)
+vs_refresh(
+	SCR *sp,
+	int forcepaint)
 {
 	GS *gp;
 	SCR *tsp;
-	int need_refresh;
+	int need_refresh = 0;
 	u_int priv_paint, pub_paint;
 
 	gp = sp->gp;
@@ -56,8 +77,7 @@ vs_refresh(SCR *sp, int forcepaint)
 	 * that we can find, including status lines.
 	 */
 	if (F_ISSET(sp, SC_SCR_REDRAW))
-		for (tsp = sp->wp->scrq.cqh_first;
-		    tsp != (void *)&sp->wp->scrq; tsp = tsp->q.cqe_next)
+		TAILQ_FOREACH(tsp, gp->dq, q)
 			if (tsp != sp)
 				F_SET(tsp, SC_SCR_REDRAW | SC_STATUS);
 
@@ -74,8 +94,7 @@ vs_refresh(SCR *sp, int forcepaint)
 	priv_paint = VIP_CUR_INVALID | VIP_N_REFRESH;
 	if (O_ISSET(sp, O_NUMBER))
 		priv_paint |= VIP_N_RENUMBER;
-	for (tsp = sp->wp->scrq.cqh_first;
-	    tsp != (void *)&sp->wp->scrq; tsp = tsp->q.cqe_next)
+	TAILQ_FOREACH(tsp, gp->dq, q)
 		if (tsp != sp && !F_ISSET(tsp, SC_EXIT | SC_EXIT_FORCE) &&
 		    (F_ISSET(tsp, pub_paint) ||
 		    F_ISSET(VIP(tsp), priv_paint))) {
@@ -110,8 +129,7 @@ vs_refresh(SCR *sp, int forcepaint)
 	 * And, finally, if we updated any status lines, make sure the cursor
 	 * gets back to where it belongs.
 	 */
-	for (need_refresh = 0, tsp = sp->wp->scrq.cqh_first;
-	    tsp != (void *)&sp->wp->scrq; tsp = tsp->q.cqe_next)
+	TAILQ_FOREACH(tsp, gp->dq, q)
 		if (F_ISSET(tsp, SC_STATUS)) {
 			need_refresh = 1;
 			vs_resolve(tsp, sp, 0);
@@ -138,14 +156,16 @@ vs_refresh(SCR *sp, int forcepaint)
  *	what you're doing.  It's subtle and quick to anger.
  */
 static int
-vs_paint(SCR *sp, u_int flags)
+vs_paint(
+	SCR *sp,
+	u_int flags)
 {
 	GS *gp;
 	SMAP *smp, tmp;
 	VI_PRIVATE *vip;
-	db_recno_t lastline, lcnt;
+	recno_t lastline, lcnt;
 	size_t cwtotal, cnt, len, notused, off, y;
-	int ch, didpaint, isempty, leftright_warp;
+	int ch = 0, didpaint, isempty, leftright_warp;
 	CHAR_T *p;
 
 #define	 LNO	sp->lno			/* Current file line. */
@@ -308,7 +328,8 @@ small_fill:			(void)gp->scr_move(sp, LASTLINE(sp), 0);
 				if (vs_sm_1down(sp))
 					return (1);
 			goto adjust;
-		}
+		} else
+			goto top;	/* XXX No such line. */
 
 		/*
 		 * If less than a half screen from the bottom of the file,
@@ -319,7 +340,7 @@ bottom:		if (db_last(sp, &lastline))
 		tmp.lno = LNO;
 		tmp.coff = HMAP->coff;
 		tmp.soff = 1;
-		lcnt = vs_sm_nlines(sp, &tmp, lastline+1, sp->t_rows);
+		lcnt = vs_sm_nlines(sp, &tmp, lastline, sp->t_rows);
 		if (lcnt < HALFTEXT(sp)) {
 			if (vs_sm_fill(sp, lastline, P_BOTTOM))
 				return (1);
@@ -444,8 +465,8 @@ adjust:	if (!O_ISSET(sp, O_LEFTRIGHT) &&
 #ifdef DEBUG
 	/* Sanity checking. */
 	if (CNO >= len && len != 0) {
-		msgq(sp, M_ERR, "Error: %s/%d: cno (%u) >= len (%u)",
-		     tail(__FILE__), __LINE__, CNO, len);
+		msgq(sp, M_ERR, "Error: %s/%d: cno (%zu) >= len (%zu)",
+		     basename(__FILE__), __LINE__, CNO, len);
 		return (1);
 	}
 #endif
@@ -479,7 +500,7 @@ adjust:	if (!O_ISSET(sp, O_LEFTRIGHT) &&
 
 		/*
 		 * Count up the widths of the characters.  If it's a tab
-		 * character, go do it the the slow way.
+		 * character, go do it the slow way.
 		 */
 		for (cwtotal = 0; cnt--; cwtotal += KEY_COL(sp, ch))
 			if ((ch = *(UCHAR_T *)p--) == '\t')
@@ -517,7 +538,7 @@ adjust:	if (!O_ISSET(sp, O_LEFTRIGHT) &&
 
 		/*
 		 * Count up the widths of the characters.  If it's a tab
-		 * character, go do it the the slow way.  If we cross a
+		 * character, go do it the slow way.  If we cross a
 		 * screen boundary, we can quit.
 		 */
 		for (cwtotal = SCNO; cnt--;) {
@@ -587,8 +608,8 @@ slow:	for (smp = HMAP; smp->lno != LNO; ++smp);
 		}
 
 		/* Adjust the window towards the end of the line. */
-		if (off == 0 && off + SCREEN_COLS(sp) < cnt ||
-		    off != 0 && off + sp->cols < cnt) {
+		if ((off == 0 && off + SCREEN_COLS(sp) < cnt) ||
+		    (off != 0 && off + sp->cols < cnt)) {
 			do {
 				off += O_VAL(sp, O_SIDESCROLL);
 			} while (off + sp->cols < cnt);
@@ -662,11 +683,8 @@ done_cursor:
 	 * NULL.  If we're debugging, die, otherwise restart from scratch.
 	 */
 #ifdef DEBUG
-	if (vip->sc_smap == NULL) {
-		fprintf(stderr, "smap error\n");
-		sleep(100);
+	if (vip->sc_smap == NULL)
 		abort();
-	}
 #else
 	if (vip->sc_smap == NULL) {
 		F_SET(sp, SC_SCR_REFORMAT);
@@ -751,9 +769,9 @@ vs_modeline(SCR *sp)
 	};
 	GS *gp;
 	size_t cols, curcol, curlen, endpoint, len, midpoint;
-	const char *t;
+	const char *t = NULL;
 	int ellipsis;
-	char *p, buf[20];
+	char buf[20];
 
 	gp = sp->gp;
 
@@ -775,19 +793,23 @@ vs_modeline(SCR *sp)
 	/* If more than one screen in the display, show the file name. */
 	curlen = 0;
 	if (IS_SPLIT(sp)) {
-		for (p = sp->frp->name; *p != '\0'; ++p);
-		for (ellipsis = 0, cols = sp->cols / 2; --p > sp->frp->name;) {
+		CHAR_T *wp, *p;
+		size_t l;
+
+		CHAR2INT(sp, sp->frp->name, strlen(sp->frp->name) + 1, wp, l);
+		p = wp + l;
+		for (ellipsis = 0, cols = sp->cols / 2; --p > wp;) {
 			if (*p == '/') {
 				++p;
 				break;
 			}
-			if ((curlen += KEY_LEN(sp, *p)) > cols) {
+			if ((curlen += KEY_COL(sp, *p)) > cols) {
 				ellipsis = 3;
 				curlen +=
 				    KEY_LEN(sp, '.') * 3 + KEY_LEN(sp, ' ');
 				while (curlen > cols) {
 					++p;
-					curlen -= KEY_LEN(sp, *p);
+					curlen -= KEY_COL(sp, *p);
 				}
 				break;
 			}
@@ -801,7 +823,7 @@ vs_modeline(SCR *sp)
 		}
 		for (; *p != '\0'; ++p)
 			(void)gp->scr_addstr(sp,
-			    KEY_NAME(sp, *p), KEY_LEN(sp, *p));
+			    KEY_NAME(sp, *p), KEY_COL(sp, *p));
 	}
 
 	/* Clear the rest of the line. */
@@ -821,8 +843,8 @@ vs_modeline(SCR *sp)
 	cols = sp->cols - 1;
 	if (O_ISSET(sp, O_RULER)) {
 		vs_column(sp, &curcol);
-		len =
-		    snprintf(buf, sizeof(buf), "%lu,%lu", sp->lno, curcol + 1);
+		len = snprintf(buf, sizeof(buf), "%lu,%lu",
+		    (u_long)sp->lno, (u_long)(curcol + 1));
 
 		midpoint = (cols - ((len + 1) / 2)) / 2;
 		if (curlen < midpoint) {
