@@ -9,10 +9,6 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: vs_smap.c,v 10.30 2002/01/19 21:59:07 skimo Exp $ (Berkeley) $Date: 2002/01/19 21:59:07 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/time.h>
@@ -26,23 +22,23 @@ static const char sccsid[] = "$Id: vs_smap.c,v 10.30 2002/01/19 21:59:07 skimo E
 #include "../common/common.h"
 #include "vi.h"
 
-static int	vs_deleteln __P((SCR *, int));
-static int	vs_insertln __P((SCR *, int));
-static int	vs_sm_delete __P((SCR *, db_recno_t));
-static int	vs_sm_down __P((SCR *, MARK *, db_recno_t, scroll_t, SMAP *));
-static int	vs_sm_erase __P((SCR *));
-static int	vs_sm_insert __P((SCR *, db_recno_t));
-static int	vs_sm_reset __P((SCR *, db_recno_t));
-static int	vs_sm_up __P((SCR *, MARK *, db_recno_t, scroll_t, SMAP *));
+static int	vs_deleteln(SCR *, int);
+static int	vs_insertln(SCR *, int);
+static int	vs_sm_delete(SCR *, recno_t);
+static int	vs_sm_down(SCR *, MARK *, recno_t, scroll_t, SMAP *);
+static int	vs_sm_erase(SCR *);
+static int	vs_sm_insert(SCR *, recno_t);
+static int	vs_sm_reset(SCR *, recno_t);
+static int	vs_sm_up(SCR *, MARK *, recno_t, scroll_t, SMAP *);
 
 /*
  * vs_change --
  *	Make a change to the screen.
  *
- * PUBLIC: int vs_change __P((SCR *, db_recno_t, lnop_t));
+ * PUBLIC: int vs_change(SCR *, recno_t, lnop_t);
  */
 int
-vs_change(SCR *sp, db_recno_t lno, lnop_t op)
+vs_change(SCR *sp, recno_t lno, lnop_t op)
 {
 	VI_PRIVATE *vip;
 	SMAP *p;
@@ -171,10 +167,10 @@ vs_change(SCR *sp, db_recno_t lno, lnop_t op)
  * slot is already filled in, P_BOTTOM means that the TMAP slot is
  * already filled in, and we just finish up the job.
  *
- * PUBLIC: int vs_sm_fill __P((SCR *, db_recno_t, pos_t));
+ * PUBLIC: int vs_sm_fill(SCR *, recno_t, pos_t);
  */
 int
-vs_sm_fill(SCR *sp, db_recno_t lno, pos_t pos)
+vs_sm_fill(SCR *sp, recno_t lno, pos_t pos)
 {
 	SMAP *p, tmp;
 	size_t cnt;
@@ -224,6 +220,16 @@ vs_sm_fill(SCR *sp, db_recno_t lno, pos_t pos)
 top:			HMAP->lno = lno;
 			HMAP->coff = 0;
 			HMAP->soff = 1;
+		} else {
+			/*
+			 * If number of lines HMAP->lno (top line) spans
+			 * changed due to, say reformatting, and now is
+			 * fewer than HMAP->soff, reset so the line is
+			 * redrawn at the top of the screen.
+			 */
+			cnt = vs_screens(sp, HMAP->lno, NULL);
+			if (cnt < HMAP->soff)
+				HMAP->soff = 1;
 		}
 		/* If we fail, just punt. */
 		for (p = HMAP, cnt = sp->t_rows; --cnt; ++p)
@@ -298,7 +304,7 @@ err:	HMAP->lno = 1;
  *	Delete a line out of the SMAP.
  */
 static int
-vs_sm_delete(SCR *sp, db_recno_t lno)
+vs_sm_delete(SCR *sp, recno_t lno)
 {
 	SMAP *p, *t;
 	size_t cnt_orig;
@@ -346,7 +352,7 @@ vs_sm_delete(SCR *sp, db_recno_t lno)
  *	Insert a line into the SMAP.
  */
 static int
-vs_sm_insert(SCR *sp, db_recno_t lno)
+vs_sm_insert(SCR *sp, recno_t lno)
 {
 	SMAP *p, *t;
 	size_t cnt_orig, cnt, coff;
@@ -400,7 +406,7 @@ vs_sm_insert(SCR *sp, db_recno_t lno)
  *	Reset a line in the SMAP.
  */
 static int
-vs_sm_reset(SCR *sp, db_recno_t lno)
+vs_sm_reset(SCR *sp, recno_t lno)
 {
 	SMAP *p, *t;
 	size_t cnt_orig, cnt_new, cnt, diff;
@@ -502,10 +508,10 @@ vs_sm_reset(SCR *sp, db_recno_t lno)
  *	Scroll the SMAP up/down count logical lines.  Different
  *	semantics based on the vi command, *sigh*.
  *
- * PUBLIC: int vs_sm_scroll __P((SCR *, MARK *, db_recno_t, scroll_t));
+ * PUBLIC: int vs_sm_scroll(SCR *, MARK *, recno_t, scroll_t);
  */
 int
-vs_sm_scroll(SCR *sp, MARK *rp, db_recno_t count, scroll_t scmd)
+vs_sm_scroll(SCR *sp, MARK *rp, recno_t count, scroll_t scmd)
 {
 	SMAP *smp;
 
@@ -562,7 +568,7 @@ vs_sm_scroll(SCR *sp, MARK *rp, db_recno_t count, scroll_t scmd)
  *	Scroll the SMAP up count logical lines.
  */
 static int
-vs_sm_up(SCR *sp, MARK *rp, db_recno_t count, scroll_t scmd, SMAP *smp)
+vs_sm_up(SCR *sp, MARK *rp, recno_t count, scroll_t scmd, SMAP *smp)
 {
 	int cursor_set, echanged, zset;
 	SMAP *ssmp, s1, s2;
@@ -735,7 +741,7 @@ vs_sm_up(SCR *sp, MARK *rp, db_recno_t count, scroll_t scmd, SMAP *smp)
  * vs_sm_1up --
  *	Scroll the SMAP up one.
  *
- * PUBLIC: int vs_sm_1up __P((SCR *));
+ * PUBLIC: int vs_sm_1up(SCR *);
  */
 int
 vs_sm_1up(SCR *sp)
@@ -799,7 +805,7 @@ vs_deleteln(SCR *sp, int cnt)
  *	Scroll the SMAP down count logical lines.
  */
 static int
-vs_sm_down(SCR *sp, MARK *rp, db_recno_t count, scroll_t scmd, SMAP *smp)
+vs_sm_down(SCR *sp, MARK *rp, recno_t count, scroll_t scmd, SMAP *smp)
 {
 	SMAP *ssmp, s1, s2;
 	int cursor_set, ychanged, zset;
@@ -968,7 +974,7 @@ vs_sm_erase(SCR *sp)
  * vs_sm_1down --
  *	Scroll the SMAP down one.
  *
- * PUBLIC: int vs_sm_1down __P((SCR *));
+ * PUBLIC: int vs_sm_1down(SCR *);
  */
 int
 vs_sm_1down(SCR *sp)
@@ -1012,7 +1018,7 @@ vs_insertln(SCR *sp, int cnt)
 		F_SET(sp, SC_SCR_REDRAW);
 		return (0);
 	}
-		
+
 	if (IS_ONELINE(sp)) {
 		(void)gp->scr_move(sp, LASTLINE(sp), 0);
 		(void)gp->scr_clrtoeol(sp);
@@ -1032,7 +1038,7 @@ vs_insertln(SCR *sp, int cnt)
  * vs_sm_next --
  *	Fill in the next entry in the SMAP.
  *
- * PUBLIC: int vs_sm_next __P((SCR *, SMAP *, SMAP *));
+ * PUBLIC: int vs_sm_next(SCR *, SMAP *, SMAP *);
  */
 int
 vs_sm_next(SCR *sp, SMAP *p, SMAP *t)
@@ -1060,7 +1066,7 @@ vs_sm_next(SCR *sp, SMAP *p, SMAP *t)
  * vs_sm_prev --
  *	Fill in the previous entry in the SMAP.
  *
- * PUBLIC: int vs_sm_prev __P((SCR *, SMAP *, SMAP *));
+ * PUBLIC: int vs_sm_prev(SCR *, SMAP *, SMAP *);
  */
 int
 vs_sm_prev(SCR *sp, SMAP *p, SMAP *t)
@@ -1085,7 +1091,7 @@ vs_sm_prev(SCR *sp, SMAP *p, SMAP *t)
  * vs_sm_cursor --
  *	Return the SMAP entry referenced by the cursor.
  *
- * PUBLIC: int vs_sm_cursor __P((SCR *, SMAP **));
+ * PUBLIC: int vs_sm_cursor(SCR *, SMAP **);
  */
 int
 vs_sm_cursor(SCR *sp, SMAP **smpp)
@@ -1124,13 +1130,13 @@ vs_sm_cursor(SCR *sp, SMAP **smpp)
  *	(The vi H, M and L commands.)  Here because only the screen routines
  *	know what's really out there.
  *
- * PUBLIC: int vs_sm_position __P((SCR *, MARK *, u_long, pos_t));
+ * PUBLIC: int vs_sm_position(SCR *, MARK *, u_long, pos_t);
  */
 int
 vs_sm_position(SCR *sp, MARK *rp, u_long cnt, pos_t pos)
 {
 	SMAP *smp;
-	db_recno_t last;
+	recno_t last;
 
 	switch (pos) {
 	case P_TOP:
@@ -1204,12 +1210,12 @@ eof:				msgq(sp, M_BERR,
  *	Return the number of screen lines from an SMAP entry to the
  *	start of some file line, less than a maximum value.
  *
- * PUBLIC: db_recno_t vs_sm_nlines __P((SCR *, SMAP *, db_recno_t, size_t));
+ * PUBLIC: recno_t vs_sm_nlines(SCR *, SMAP *, recno_t, size_t);
  */
-db_recno_t
-vs_sm_nlines(SCR *sp, SMAP *from_sp, db_recno_t to_lno, size_t max)
+recno_t
+vs_sm_nlines(SCR *sp, SMAP *from_sp, recno_t to_lno, size_t max)
 {
-	db_recno_t lno, lcnt;
+	recno_t lno, lcnt;
 
 	if (O_ISSET(sp, O_LEFTRIGHT))
 		return (from_sp->lno > to_lno ?
